@@ -5,52 +5,30 @@ from zope.interface import implements
 from twisted.internet import defer
 import simplejson
 from allmydata.mutable.common import NotMutableError
-from allmydata.mutable.node import MutableFileNode
+from allmydata.mutable.filenode import MutableFileNode
 from allmydata.interfaces import IMutableFileNode, IDirectoryNode,\
-     IURI, IFileNode, IMutableFileURI, IVerifierURI, IFilesystemNode, \
-     ExistingChildError, ICheckable
-from allmydata.immutable.checker import DeepCheckResults
-from allmydata.util import hashutil, mathutil
+     IURI, IFileNode, IMutableFileURI, IFilesystemNode, \
+     ExistingChildError, NoSuchChildError, ICheckable, IDeepCheckable
+from allmydata.check_results import DeepCheckResults, \
+     DeepCheckAndRepairResults
+from allmydata.monitor import Monitor
+from allmydata.util import hashutil, mathutil, base32, log
+from allmydata.util.assertutil import _assert, precondition
 from allmydata.util.hashutil import netstring
-from allmydata.util.limiter import ConcurrencyLimiter
-from allmydata.uri import NewDirectoryURI
+from allmydata.util.netstring import split_netstring
+from allmydata.uri import NewDirectoryURI, LiteralFileURI, from_string
 from pycryptopp.cipher.aes import AES
 
-def split_netstring(data, numstrings, allow_leftover=False):
-    """like string.split(), but extracts netstrings. If allow_leftover=False,
-    returns numstrings elements, and throws ValueError if there was leftover
-    data. If allow_leftover=True, returns numstrings+1 elements, in which the
-    last element is the leftover data (possibly an empty string)"""
-    elements = []
-    assert numstrings >= 0
-    while data:
-        colon = data.index(":")
-        length = int(data[:colon])
-        string = data[colon+1:colon+1+length]
-        assert len(string) == length
-        elements.append(string)
-        assert data[colon+1+length] == ","
-        data = data[colon+1+length+1:]
-        if len(elements) == numstrings:
-            break
-    if len(elements) < numstrings:
-        raise ValueError("ran out of netstrings")
-    if allow_leftover:
-        return tuple(elements + [data])
-    if data:
-        raise ValueError("leftover data in netstrings")
-    return tuple(elements)
-
 class Deleter:
     def __init__(self, node, name, must_exist=True):
         self.node = node
         self.name = name
         self.must_exist = True
-    def modify(self, old_contents):
+    def modify(self, old_contents, servermap, first_time):
         children = self.node._unpack_contents(old_contents)
         if self.name not in children:
-            if self.must_exist:
-                raise KeyError(self.name)
+            if first_time and self.must_exist:
+                raise NoSuchChildError(self.name)
             self.old_child = None
             return None
         self.old_child, metadata = children[self.name]
@@ -64,8 +42,10 @@ class MetadataSetter:
         self.name = name
         self.metadata = metadata
 
-    def modify(self, old_contents):
+    def modify(self, old_contents, servermap, first_time):
         children = self.node._unpack_contents(old_contents)
+        if self.name not in children:
+            raise NoSuchChildError(self.name)
         children[self.name] = (children[self.name][0], self.metadata)
         new_contents = self.node._pack_contents(children)
         return new_contents
@@ -80,9 +60,11 @@ class Adder:
         self.overwrite = overwrite
 
     def set_node(self, name, node, metadata):
+        precondition(isinstance(name, unicode), name)
+        precondition(IFilesystemNode.providedBy(node), node)
         self.entries.append( [name, node, metadata] )
 
-    def modify(self, old_contents):
+    def modify(self, old_contents, servermap, first_time):
         children = self.node._unpack_contents(old_contents)
         now = time.time()
         for e in self.entries:
@@ -92,6 +74,7 @@ class Adder:
             else:
                 assert len(e) == 3
                 name, child, new_metadata = e
+                assert _assert(IFilesystemNode.providedBy(child), child)
             assert isinstance(name, unicode)
             if name in children:
                 if not self.overwrite:
@@ -113,7 +96,7 @@ class Adder:
         return new_contents
 
 class NewDirectoryNode:
-    implements(IDirectoryNode, ICheckable)
+    implements(IDirectoryNode, ICheckable, IDeepCheckable)
     filenode_class = MutableFileNode
 
     def __init__(self, client):
@@ -168,14 +151,12 @@ class NewDirectoryNode:
         mac = hashutil.hmac(key, IV + crypttext)
         assert len(mac) == 32
         return IV + crypttext + mac
+        # The MAC is not checked by readers in Tahoe >= 1.3.0, but we still produce it for the sake of older readers.
 
     def _decrypt_rwcapdata(self, encwrcap):
         IV = encwrcap[:16]
         crypttext = encwrcap[16:-32]
-        mac = encwrcap[-32:]
         key = hashutil.mutable_rwcap_key_hash(IV, self._node.get_writekey())
-        if mac != hashutil.hmac(key, IV+crypttext):
-            raise hashutil.IntegrityCheckError("HMAC does not match, crypttext is corrupted")
         cryptor = AES(key)
         plaintext = cryptor.process(crypttext)
         return plaintext
@@ -221,7 +202,9 @@ class NewDirectoryNode:
                     or IDirectoryNode.providedBy(child)), (name,child)
             assert isinstance(metadata, dict)
             rwcap = child.get_uri() # might be RO if the child is not writeable
+            assert isinstance(rwcap, str), rwcap
             rocap = child.get_readonly_uri()
+            assert isinstance(rocap, str), rocap
             entry = "".join([netstring(name.encode("utf-8")),
                              netstring(rocap),
                              netstring(self._encrypt_rwcap(rwcap)),
@@ -240,12 +223,22 @@ class NewDirectoryNode:
     def get_readonly_uri(self):
         return self._uri.get_readonly().to_string()
 
-    def get_verifier(self):
-        return self._uri.get_verifier().to_string()
+    def get_verify_cap(self):
+        return self._uri.get_verify_cap()
 
-    def check(self, verify=False, repair=False):
+    def get_repair_cap(self):
+        if self._node.is_readonly():
+            return None
+        return self._uri
+
+    def get_storage_index(self):
+        return self._uri._filenode_uri.storage_index
+
+    def check(self, monitor, verify=False):
         """Perform a file check. See IChecker.check for details."""
-        return self._node.check(verify, repair)
+        return self._node.check(monitor, verify)
+    def check_and_repair(self, monitor, verify=False):
+        return self._node.check_and_repair(monitor, verify)
 
     def list(self):
         """I return a Deferred that fires with a dictionary mapping child
@@ -263,9 +256,15 @@ class NewDirectoryNode:
     def _get(self, children, name):
         child = children.get(name)
         if child is None:
-            raise KeyError(name)
+            raise NoSuchChildError(name)
         return child[0]
 
+    def _get_with_metadata(self, children, name):
+        child = children.get(name)
+        if child is None:
+            raise NoSuchChildError(name)
+        return child
+
     def get(self, name):
         """I return a Deferred that fires with the named child node,
         which is either an IFileNode or an IDirectoryNode."""
@@ -274,6 +273,15 @@ class NewDirectoryNode:
         d.addCallback(self._get, name)
         return d
 
+    def get_child_and_metadata(self, name):
+        """I return a Deferred that fires with the (node, metadata) pair for
+        the named child. The node is either an IFileNode or an
+        IDirectoryNode, and the metadata is a dictionary."""
+        assert isinstance(name, unicode)
+        d = self._read()
+        d.addCallback(self._get_with_metadata, name)
+        return d
+
     def get_metadata_for(self, name):
         assert isinstance(name, unicode)
         d = self._read()
@@ -300,9 +308,17 @@ class NewDirectoryNode:
         The path can be either a single string (slash-separated) or a list of
         path-name elements.
         """
+        d = self.get_child_and_metadata_at_path(path)
+        d.addCallback(lambda (node, metadata): node)
+        return d
+
+    def get_child_and_metadata_at_path(self, path):
+        """Transform a child path into an IDirectoryNode or IFileNode and
+        a metadata dictionary from the last edge that was traversed.
+        """
 
         if not path:
-            return defer.succeed(self)
+            return defer.succeed((self, {}))
         if isinstance(path, (list, tuple)):
             pass
         else:
@@ -311,11 +327,12 @@ class NewDirectoryNode:
             assert isinstance(p, unicode)
         childname = path[0]
         remaining_path = path[1:]
-        d = self.get(childname)
         if remaining_path:
-            def _got(node):
-                return node.get_child_at_path(remaining_path)
-            d.addCallback(_got)
+            d = self.get(childname)
+            d.addCallback(lambda node:
+                          node.get_child_and_metadata_at_path(remaining_path))
+            return d
+        d = self.get_child_and_metadata(childname)
         return d
 
     def set_uri(self, name, child_uri, metadata=None, overwrite=True):
@@ -328,7 +345,8 @@ class NewDirectoryNode:
 
         If this directory node is read-only, the Deferred will errback with a
         NotMutableError."""
-        assert isinstance(name, unicode)
+        precondition(isinstance(name, unicode), name)
+        precondition(isinstance(child_uri, str), child_uri)
         child_node = self._create_node(child_uri)
         d = self.set_node(name, child_node, metadata, overwrite)
         d.addCallback(lambda res: child_node)
@@ -358,6 +376,8 @@ class NewDirectoryNode:
         If this directory node is read-only, the Deferred will errback with a
         NotMutableError."""
 
+        precondition(IFilesystemNode.providedBy(child), child)
+
         if self.is_readonly():
             return defer.fail(NotMutableError())
         assert isinstance(name, unicode)
@@ -440,141 +460,117 @@ class NewDirectoryNode:
         d.addCallback(lambda child: self.delete(current_child_name))
         return d
 
-    def build_manifest(self):
-        """Return a frozenset of verifier-capability strings for all nodes
-        (directories and files) reachable from this one."""
+
+    def deep_traverse(self, walker):
+        """Perform a recursive walk, using this dirnode as a root, notifying
+        the 'walker' instance of everything I encounter.
+
+        I call walker.enter_directory(parent, children) once for each dirnode
+        I visit, immediately after retrieving the list of children. I pass in
+        the parent dirnode and the dict of childname->(childnode,metadata).
+        This function should *not* traverse the children: I will do that.
+        enter_directory() is most useful for the deep-stats number that
+        counts how large a directory is.
+
+        I call walker.add_node(node, path) for each node (both files and
+        directories) I can reach. Most work should be done here.
+
+        I avoid loops by keeping track of verifier-caps and refusing to call
+        walker.add_node() or traverse a node that I've seen before. This
+        means that any file or directory will only be given to the walker
+        once. If files or directories are referenced multiple times by a
+        directory structure, this may appear to under-count or miss some of
+        them.
+
+        I return a Monitor which can be used to wait for the operation to
+        finish, learn about its progress, or cancel the operation.
+        """
 
         # this is just a tree-walker, except that following each edge
-        # requires a Deferred. We use a ConcurrencyLimiter to make sure the
-        # fan-out doesn't cause problems.
+        # requires a Deferred. We used to use a ConcurrencyLimiter to limit
+        # fanout to 10 simultaneous operations, but the memory load of the
+        # queued operations was excessive (in one case, with 330k dirnodes,
+        # it caused the process to run into the 3.0GB-ish per-process 32bit
+        # linux memory limit, and crashed). So we use a single big Deferred
+        # chain, and do a strict depth-first traversal, one node at a time.
+        # This can be slower, because we aren't pipelining directory reads,
+        # but it brought the memory footprint down by roughly 50%.
 
-        manifest = set()
-        manifest.add(self.get_verifier())
-        limiter = ConcurrencyLimiter(10) # allow 10 in parallel
+        monitor = Monitor()
+        walker.set_monitor(monitor)
 
-        d = self._build_manifest_from_node(self, manifest, limiter)
-        def _done(res):
-            # LIT nodes have no verifier-capability: their data is stored
-            # inside the URI itself, so there is no need to refresh anything.
-            # They indicate this by returning None from their get_verifier
-            # method. We need to remove any such Nones from our set. We also
-            # want to convert all these caps into strings.
-            return frozenset([IVerifierURI(cap).to_string()
-                              for cap in manifest
-                              if cap is not None])
-        d.addCallback(_done)
+        found = set([self.get_verify_cap()])
+        d = self._deep_traverse_dirnode(self, [], walker, monitor, found)
+        d.addCallback(lambda ignored: walker.finish())
+        d.addBoth(monitor.finish)
+        d.addErrback(lambda f: None)
+
+        return monitor
+
+    def _deep_traverse_dirnode(self, node, path, walker, monitor, found):
+        # process this directory, then walk its children
+        monitor.raise_if_cancelled()
+        d = defer.maybeDeferred(walker.add_node, node, path)
+        d.addCallback(lambda ignored: node.list())
+        d.addCallback(self._deep_traverse_dirnode_children, node, path,
+                      walker, monitor, found)
         return d
 
-    def _build_manifest_from_node(self, node, manifest, limiter):
-        d = limiter.add(node.list)
-        def _got_list(res):
-            dl = []
-            for name, (child, metadata) in res.iteritems():
-                verifier = child.get_verifier()
-                if verifier not in manifest:
-                    manifest.add(verifier)
-                    if IDirectoryNode.providedBy(child):
-                        dl.append(self._build_manifest_from_node(child,
-                                                                 manifest,
-                                                                 limiter))
-            if dl:
-                return defer.DeferredList(dl)
-        d.addCallback(_got_list)
+    def _deep_traverse_dirnode_children(self, children, parent, path,
+                                        walker, monitor, found):
+        monitor.raise_if_cancelled()
+        d = defer.maybeDeferred(walker.enter_directory, parent, children)
+        # we process file-like children first, so we can drop their FileNode
+        # objects as quickly as possible. Tests suggest that a FileNode (held
+        # in the client's nodecache) consumes about 2440 bytes. dirnodes (not
+        # in the nodecache) seem to consume about 2000 bytes.
+        dirkids = []
+        filekids = []
+        for name, (child, metadata) in children.iteritems():
+            verifier = child.get_verify_cap()
+            # allow LIT files (for which verifier==None) to be processed
+            if (verifier is not None) and (verifier in found):
+                continue
+            found.add(verifier)
+            childpath = path + [name]
+            if IDirectoryNode.providedBy(child):
+                dirkids.append( (child, childpath) )
+            else:
+                filekids.append( (child, childpath) )
+        for (child, childpath) in filekids:
+            d.addCallback(lambda ignored, child=child, childpath=childpath:
+                          walker.add_node(child, childpath))
+        for (child, childpath) in dirkids:
+            d.addCallback(lambda ignored, child=child, childpath=childpath:
+                          self._deep_traverse_dirnode(child, childpath,
+                                                      walker, monitor,
+                                                      found))
         return d
 
-    def deep_stats(self):
-        stats = DeepStats()
-        # we track verifier caps, to avoid double-counting children for which
-        # we've got both a write-cap and a read-cap
-        found = set()
-        found.add(self.get_verifier())
 
-        limiter = ConcurrencyLimiter(10)
+    def build_manifest(self):
+        """Return a Monitor, with a ['status'] that will be a list of (path,
+        cap) tuples, for all nodes (directories and files) reachable from
+        this one."""
+        walker = ManifestWalker(self)
+        return self.deep_traverse(walker)
 
-        d = self._add_deepstats_from_node(self, found, stats, limiter)
-        d.addCallback(lambda res: stats.get_results())
-        return d
+    def start_deep_stats(self):
+        # Since deep_traverse tracks verifier caps, we avoid double-counting
+        # children for which we've got both a write-cap and a read-cap
+        return self.deep_traverse(DeepStats(self))
 
-    def _add_deepstats_from_node(self, node, found, stats, limiter):
-        d = limiter.add(node.list)
-        def _got_list(children):
-            dl = []
-            dirsize_bytes = node.get_size()
-            dirsize_children = len(children)
-            stats.add("count-directories")
-            stats.add("size-directories", dirsize_bytes)
-            stats.max("largest-directory", dirsize_bytes)
-            stats.max("largest-directory-children", dirsize_children)
-            for name, (child, metadata) in children.iteritems():
-                verifier = child.get_verifier()
-                if verifier in found:
-                    continue
-                found.add(verifier)
-                if IDirectoryNode.providedBy(child):
-                    dl.append(self._add_deepstats_from_node(child, found,
-                                                            stats, limiter))
-                elif IMutableFileNode.providedBy(child):
-                    stats.add("count-files")
-                    stats.add("count-mutable-files")
-                    # TODO: update the servermap, compute a size, add it to
-                    # size-mutable-files, max it into "largest-mutable-file"
-                elif IFileNode.providedBy(child): # CHK and LIT
-                    stats.add("count-files")
-                    size = child.get_size()
-                    stats.histogram("size-files-histogram", size)
-                    if child.get_uri().startswith("URI:LIT:"):
-                        stats.add("count-literal-files")
-                        stats.add("size-literal-files", size)
-                    else:
-                        stats.add("count-immutable-files")
-                        stats.add("size-immutable-files", size)
-                        stats.max("largest-immutable-file", size)
-            if dl:
-                return defer.DeferredList(dl)
-        d.addCallback(_got_list)
-        return d
+    def start_deep_check(self, verify=False):
+        return self.deep_traverse(DeepChecker(self, verify, repair=False))
 
-    def deep_check(self, verify=False, repair=False):
-        # shallow-check each object first, then traverse children
-        root_si = self._node.get_storage_index()
-        results = DeepCheckResults(root_si)
-        found = set()
-        limiter = ConcurrencyLimiter(10)
+    def start_deep_check_and_repair(self, verify=False):
+        return self.deep_traverse(DeepChecker(self, verify, repair=True))
 
-        d = self._add_deepcheck_from_node(self, results, found, limiter,
-                                          verify, repair)
-        d.addCallback(lambda res: results)
-        return d
-
-    def _add_deepcheck_from_node(self, node, results, found, limiter,
-                                 verify, repair):
-        verifier = node.get_verifier()
-        if verifier in found:
-            # avoid loops
-            return None
-        found.add(verifier)
-
-        d = limiter.add(node.check, verify, repair)
-        d.addCallback(results.add_check)
-
-        if IDirectoryNode.providedBy(node):
-            d.addCallback(lambda res: node.list())
-            def _got_children(children):
-                dl = []
-                for name, (child, metadata) in children.iteritems():
-                    d2 = self._add_deepcheck_from_node(child, results,
-                                                       found, limiter,
-                                                       verify, repair)
-                    if d2:
-                        dl.append(d2)
-                if dl:
-                    return defer.DeferredList(dl, fireOnOneErrback=True)
-            d.addCallback(_got_children)
-        return d
 
 
 class DeepStats:
-    def __init__(self):
+    def __init__(self, origin):
+        self.origin = origin
         self.stats = {}
         for k in ["count-immutable-files",
                   "count-mutable-files",
@@ -597,6 +593,39 @@ class DeepStats:
         self.buckets = [ (0,0), (1,3)]
         self.root = math.sqrt(10)
 
+    def set_monitor(self, monitor):
+        self.monitor = monitor
+        monitor.origin_si = self.origin.get_storage_index()
+        monitor.set_status(self.get_results())
+
+    def add_node(self, node, childpath):
+        if IDirectoryNode.providedBy(node):
+            self.add("count-directories")
+        elif IMutableFileNode.providedBy(node):
+            self.add("count-files")
+            self.add("count-mutable-files")
+            # TODO: update the servermap, compute a size, add it to
+            # size-mutable-files, max it into "largest-mutable-file"
+        elif IFileNode.providedBy(node): # CHK and LIT
+            self.add("count-files")
+            size = node.get_size()
+            self.histogram("size-files-histogram", size)
+            theuri = from_string(node.get_uri())
+            if isinstance(theuri, LiteralFileURI):
+                self.add("count-literal-files")
+                self.add("size-literal-files", size)
+            else:
+                self.add("count-immutable-files")
+                self.add("size-immutable-files", size)
+                self.max("largest-immutable-file", size)
+
+    def enter_directory(self, parent, children):
+        dirsize_bytes = parent.get_size()
+        dirsize_children = len(children)
+        self.add("size-directories", dirsize_bytes)
+        self.max("largest-directory", dirsize_bytes)
+        self.max("largest-directory-children", dirsize_children)
+
     def add(self, key, value=1):
         self.stats[key] += value
 
@@ -636,6 +665,71 @@ class DeepStats:
             stats[key] = out
         return stats
 
+    def finish(self):
+        return self.get_results()
+
+class ManifestWalker(DeepStats):
+    def __init__(self, origin):
+        DeepStats.__init__(self, origin)
+        self.manifest = []
+        self.storage_index_strings = set()
+        self.verifycaps = set()
+
+    def add_node(self, node, path):
+        self.manifest.append( (tuple(path), node.get_uri()) )
+        si = node.get_storage_index()
+        if si:
+            self.storage_index_strings.add(base32.b2a(si))
+        v = node.get_verify_cap()
+        if v:
+            self.verifycaps.add(v.to_string())
+        return DeepStats.add_node(self, node, path)
+
+    def get_results(self):
+        stats = DeepStats.get_results(self)
+        return {"manifest": self.manifest,
+                "verifycaps": self.verifycaps,
+                "storage-index": self.storage_index_strings,
+                "stats": stats,
+                }
+
+
+class DeepChecker:
+    def __init__(self, root, verify, repair):
+        root_si = root.get_storage_index()
+        self._lp = log.msg(format="deep-check starting (%(si)s),"
+                           " verify=%(verify)s, repair=%(repair)s",
+                           si=base32.b2a(root_si), verify=verify, repair=repair)
+        self._verify = verify
+        self._repair = repair
+        if repair:
+            self._results = DeepCheckAndRepairResults(root_si)
+        else:
+            self._results = DeepCheckResults(root_si)
+        self._stats = DeepStats(root)
+
+    def set_monitor(self, monitor):
+        self.monitor = monitor
+        monitor.set_status(self._results)
+
+    def add_node(self, node, childpath):
+        if self._repair:
+            d = node.check_and_repair(self.monitor, self._verify)
+            d.addCallback(self._results.add_check_and_repair, childpath)
+        else:
+            d = node.check(self.monitor, self._verify)
+            d.addCallback(self._results.add_check, childpath)
+        d.addCallback(lambda ignored: self._stats.add_node(node, childpath))
+        return d
+
+    def enter_directory(self, parent, children):
+        return self._stats.enter_directory(parent, children)
+
+    def finish(self):
+        log.msg("deep-check done", parent=self._lp)
+        self._results.update_stats(self._stats.get_results())
+        return self._results
+
 
 # use client.create_dirnode() to make one of these
 
