@@ -128,7 +128,7 @@ class CompleteBinaryTreeMixin:
         Return a list of node indices that are necessary for the hash chain.
         """
         if i < 0 or i >= len(self):
-            raise IndexError('index out of range: ' + repr(i))
+            raise IndexError('index out of range: 0 >= %s < %s' % (i, len(self)))
         needed = []
         here = i
         while here != 0:
@@ -384,7 +384,7 @@ class IncompleteHashTree(CompleteBinaryTreeMixin, list):
             for i in new_hashes:
                 if self[i]:
                     if self[i] != new_hashes[i]:
-                        msg = "new hash does not match existing hash at "
+                        msg = "new hash %s does not match existing hash %s at " % (base32.b2a(new_hashes[i]), base32.b2a(self[i]))
                         msg += self._name_hash(i)
                         raise BadHashError(msg)
                 else:
