@@ -10,7 +10,7 @@ from allmydata.util import mathutil, hashutil, base32, log
 from allmydata.util.assertutil import _assert, precondition
 from allmydata.codec import CRSEncoder
 from allmydata.interfaces import IEncoder, IStorageBucketWriter, \
-     IEncryptedUploadable, IUploadStatus
+     IEncryptedUploadable, IUploadStatus, NotEnoughSharesError
 
 """
 The goal of the encoder is to turn the original file into a series of
@@ -60,10 +60,6 @@ hash tree is put into the URI.
 
 """
 
-class NotEnoughSharesError(Exception):
-    servermap = None
-    pass
-
 class UploadAborted(Exception):
     pass
 
@@ -75,7 +71,6 @@ PiB=1024*TiB
 
 class Encoder(object):
     implements(IEncoder)
-    USE_PLAINTEXT_HASHES = False
 
     def __init__(self, log_parent=None, upload_status=None):
         object.__init__(self)
@@ -154,12 +149,12 @@ class Encoder(object):
         # it. If the tail is short, we use a different codec instance. In
         # addition, the tail codec must be fed data which has been padded out
         # to the right size.
-        self.tail_size = self.file_size % self.segment_size
-        if not self.tail_size:
-            self.tail_size = self.segment_size
+        tail_size = self.file_size % self.segment_size
+        if not tail_size:
+            tail_size = self.segment_size
 
         # the tail codec is responsible for encoding tail_size bytes
-        padded_tail_size = mathutil.next_multiple(self.tail_size,
+        padded_tail_size = mathutil.next_multiple(tail_size,
                                                   self.required_shares)
         self._tail_codec = CRSEncoder()
         self._tail_codec.set_params(padded_tail_size,
@@ -202,14 +197,16 @@ class Encoder(object):
         self.landlords = landlords.copy()
 
     def start(self):
+        """ Returns a Deferred that will fire with the verify cap (an instance of
+        uri.CHKFileVerifierURI)."""
         self.log("%s starting" % (self,))
         #paddedsize = self._size + mathutil.pad_size(self._size, self.needed_shares)
         assert self._codec
         self._crypttext_hasher = hashutil.crypttext_hasher()
         self._crypttext_hashes = []
         self.segment_num = 0
-        self.subshare_hashes = [[] for x in range(self.num_shares)]
-        # subshare_hashes[i] is a list that will be accumulated and then send
+        self.block_hashes = [[] for x in range(self.num_shares)]
+        # block_hashes[i] is a list that will be accumulated and then send
         # to landlord[i]. This list contains a hash of each segment_share
         # that we sent to that landlord.
         self.share_root_hashes = [None] * self.num_shares
@@ -241,12 +238,9 @@ class Encoder(object):
 
         d.addCallback(lambda res: self.finish_hashing())
 
-        if self.USE_PLAINTEXT_HASHES:
-            d.addCallback(lambda res:
-                          self.send_plaintext_hash_tree_to_all_shareholders())
         d.addCallback(lambda res:
                       self.send_crypttext_hash_tree_to_all_shareholders())
-        d.addCallback(lambda res: self.send_all_subshare_hash_trees())
+        d.addCallback(lambda res: self.send_all_block_hash_trees())
         d.addCallback(lambda res: self.send_all_share_hash_trees())
         d.addCallback(lambda res: self.send_uri_extension_to_all_shareholders())
 
@@ -291,7 +285,7 @@ class Encoder(object):
         self.set_status("Starting shareholders")
         dl = []
         for shareid in self.landlords:
-            d = self.landlords[shareid].start()
+            d = self.landlords[shareid].put_header()
             d.addErrback(self._remove_shareholder, shareid, "start")
             dl.append(d)
         return self._gather_responses(dl)
@@ -395,6 +389,9 @@ class Encoder(object):
                 self._crypttext_hasher.update(encrypted_piece)
                 encrypted_pieces.append(encrypted_piece)
 
+            precondition(length <= input_chunk_size,
+                         "length=%d > input_chunk_size=%d" %
+                         (length, input_chunk_size))
             if allow_short:
                 if length < input_chunk_size:
                     # padding
@@ -404,7 +401,8 @@ class Encoder(object):
                 # non-tail segments should be the full segment size
                 if length != input_chunk_size:
                     log.msg("non-tail segment should be full segment size: %d!=%d"
-                            % (length, input_chunk_size), level=log.BAD)
+                            % (length, input_chunk_size),
+                            level=log.BAD, umid="jNk5Yw")
                 precondition(length == input_chunk_size,
                              "length=%d != input_chunk_size=%d" %
                              (length, input_chunk_size))
@@ -434,17 +432,17 @@ class Encoder(object):
         self.set_encode_and_push_progress(segnum)
         lognum = self.log("send_segment(%d)" % segnum, level=log.NOISY)
         for i in range(len(shares)):
-            subshare = shares[i]
+            block = shares[i]
             shareid = shareids[i]
-            d = self.send_subshare(shareid, segnum, subshare, lognum)
+            d = self.send_block(shareid, segnum, block, lognum)
             dl.append(d)
-            subshare_hash = hashutil.block_hash(subshare)
+            block_hash = hashutil.block_hash(block)
             #from allmydata.util import base32
             #log.msg("creating block (shareid=%d, blocknum=%d) "
             #        "len=%d %r .. %r: %s" %
-            #        (shareid, segnum, len(subshare),
-            #         subshare[:50], subshare[-50:], base32.b2a(subshare_hash)))
-            self.subshare_hashes[shareid].append(subshare_hash)
+            #        (shareid, segnum, len(block),
+            #         block[:50], block[-50:], base32.b2a(block_hash)))
+            self.block_hashes[shareid].append(block_hash)
 
         dl = self._gather_responses(dl)
         def _logit(res):
@@ -461,13 +459,13 @@ class Encoder(object):
         dl.addCallback(_logit)
         return dl
 
-    def send_subshare(self, shareid, segment_num, subshare, lognum):
+    def send_block(self, shareid, segment_num, block, lognum):
         if shareid not in self.landlords:
             return defer.succeed(None)
         sh = self.landlords[shareid]
         lognum2 = self.log("put_block to %s" % self.landlords[shareid],
                            parent=lognum, level=log.NOISY)
-        d = sh.put_block(segment_num, subshare)
+        d = sh.put_block(segment_num, block)
         def _done(res):
             self.log("put_block done", parent=lognum2, level=log.NOISY)
             return res
@@ -486,7 +484,7 @@ class Encoder(object):
         else:
             # even more UNUSUAL
             self.log("they weren't in our list of landlords", parent=ln,
-                     level=log.WEIRD)
+                     level=log.WEIRD, umid="TQGFRw")
         if len(self.landlords) < self.shares_of_happiness:
             msg = "lost too many shareholders during upload: %s" % why
             raise NotEnoughSharesError(msg)
@@ -516,49 +514,7 @@ class Encoder(object):
         self.set_encode_and_push_progress(extra=0.0)
         crypttext_hash = self._crypttext_hasher.digest()
         self.uri_extension_data["crypttext_hash"] = crypttext_hash
-        d = self._uploadable.get_plaintext_hash()
-        def _got(plaintext_hash):
-            self.log(format="plaintext_hash=%(plaintext_hash)s, SI=%(SI)s, size=%(size)d",
-                     plaintext_hash=base32.b2a(plaintext_hash),
-                     SI=storage.si_b2a(self._storage_index),
-                     size=self.file_size)
-            return plaintext_hash
-        d.addCallback(_got)
-        if self.USE_PLAINTEXT_HASHES:
-            def _use_plaintext_hash(plaintext_hash):
-                self.uri_extension_data["plaintext_hash"] = plaintext_hash
-                return self._uploadable.get_plaintext_hashtree_leaves(0, self.num_segments, self.num_segments)
-            d.addCallback(_use_plaintext_hash)
-            def _got_hashtree_leaves(leaves):
-                self.log("Encoder: got plaintext_hashtree_leaves: %s" %
-                         (",".join([base32.b2a(h) for h in leaves]),),
-                         level=log.NOISY)
-                ht = list(HashTree(list(leaves)))
-                self.uri_extension_data["plaintext_root_hash"] = ht[0]
-                self._plaintext_hashtree_nodes = ht
-            d.addCallback(_got_hashtree_leaves)
-
-        d.addCallback(lambda res: self._uploadable.close())
-        return d
-
-    def send_plaintext_hash_tree_to_all_shareholders(self):
-        self.log("sending plaintext hash tree", level=log.NOISY)
-        self.set_status("Sending Plaintext Hash Tree")
-        self.set_encode_and_push_progress(extra=0.2)
-        dl = []
-        for shareid in self.landlords.keys():
-            d = self.send_plaintext_hash_tree(shareid,
-                                              self._plaintext_hashtree_nodes)
-            dl.append(d)
-        return self._gather_responses(dl)
-
-    def send_plaintext_hash_tree(self, shareid, all_hashes):
-        if shareid not in self.landlords:
-            return defer.succeed(None)
-        sh = self.landlords[shareid]
-        d = sh.put_plaintext_hashes(all_hashes)
-        d.addErrback(self._remove_shareholder, shareid, "put_plaintext_hashes")
-        return d
+        self._uploadable.close()
 
     def send_crypttext_hash_tree_to_all_shareholders(self):
         self.log("sending crypttext hash tree", level=log.NOISY)
@@ -580,19 +536,19 @@ class Encoder(object):
         d.addErrback(self._remove_shareholder, shareid, "put_crypttext_hashes")
         return d
 
-    def send_all_subshare_hash_trees(self):
-        self.log("sending subshare hash trees", level=log.NOISY)
+    def send_all_block_hash_trees(self):
+        self.log("sending block hash trees", level=log.NOISY)
         self.set_status("Sending Subshare Hash Trees")
         self.set_encode_and_push_progress(extra=0.4)
         dl = []
-        for shareid,hashes in enumerate(self.subshare_hashes):
-            # hashes is a list of the hashes of all subshares that were sent
+        for shareid,hashes in enumerate(self.block_hashes):
+            # hashes is a list of the hashes of all blocks that were sent
             # to shareholder[shareid].
-            dl.append(self.send_one_subshare_hash_tree(shareid, hashes))
+            dl.append(self.send_one_block_hash_tree(shareid, hashes))
         return self._gather_responses(dl)
 
-    def send_one_subshare_hash_tree(self, shareid, subshare_hashes):
-        t = HashTree(subshare_hashes)
+    def send_one_block_hash_tree(self, shareid, block_hashes):
+        t = HashTree(block_hashes)
         all_hashes = list(t)
         # all_hashes[0] is the root hash, == hash(ah[1]+ah[2])
         # all_hashes[1] is the left child, == hash(ah[3]+ah[4])
@@ -606,10 +562,9 @@ class Encoder(object):
         return d
 
     def send_all_share_hash_trees(self):
-        # each bucket gets a set of share hash tree nodes that are needed to
-        # validate their share. This includes the share hash itself, but does
-        # not include the top-level hash root (which is stored securely in
-        # the URI instead).
+        # Each bucket gets a set of share hash tree nodes that are needed to validate their
+        # share. This includes the share hash itself, but does not include the top-level hash
+        # root (which is stored securely in the URI instead).
         self.log("sending all share hash trees", level=log.NOISY)
         self.set_status("Sending Share Hash Trees")
         self.set_encode_and_push_progress(extra=0.6)
@@ -644,10 +599,6 @@ class Encoder(object):
         for k in ('crypttext_root_hash', 'crypttext_hash',
                   ):
             assert k in self.uri_extension_data
-        if self.USE_PLAINTEXT_HASHES:
-            for k in ('plaintext_root_hash', 'plaintext_hash',
-                      ):
-                assert k in self.uri_extension_data
         uri_extension = uri.pack_extension(self.uri_extension_data)
         ed = {}
         for k,v in self.uri_extension_data.items():
@@ -691,8 +642,8 @@ class Encoder(object):
 
         # update our sharemap
         self._shares_placed = set(self.landlords.keys())
-        return (self.uri_extension_hash, self.required_shares,
-                self.num_shares, self.file_size)
+        return uri.CHKFileVerifierURI(self._storage_index, self.uri_extension_hash,
+                                      self.required_shares, self.num_shares, self.file_size)
 
     def err(self, f):
         self.log("upload failed", failure=f, level=log.UNUSUAL)
