@@ -4,6 +4,7 @@ from zope.interface import implements
 from itertools import count
 from twisted.internet import defer
 from twisted.python import failure
+from foolscap import DeadReferenceError
 from foolscap.eventual import eventually
 from allmydata.util import base32, hashutil, idlib, log
 from allmydata import storage
@@ -101,24 +102,38 @@ class ServerMap:
 
     @ivar connections: maps peerid to a RemoteReference
 
-    @ivar bad_shares: a sequence of (peerid, shnum) tuples, describing
+    @ivar bad_shares: dict with keys of (peerid, shnum) tuples, describing
                       shares that I should ignore (because a previous user of
                       the servermap determined that they were invalid). The
                       updater only locates a certain number of shares: if
                       some of these turn out to have integrity problems and
                       are unusable, the caller will need to mark those shares
                       as bad, then re-update the servermap, then try again.
+                      The dict maps (peerid, shnum) tuple to old checkstring.
     """
 
     def __init__(self):
         self.servermap = {}
         self.connections = {}
         self.unreachable_peers = set() # peerids that didn't respond to queries
+        self.reachable_peers = set() # peerids that did respond to queries
         self.problems = [] # mostly for debugging
         self.bad_shares = {} # maps (peerid,shnum) to old checkstring
         self.last_update_mode = None
         self.last_update_time = 0
 
+    def copy(self):
+        s = ServerMap()
+        s.servermap = self.servermap.copy() # tuple->tuple
+        s.connections = self.connections.copy() # str->RemoteReference
+        s.unreachable_peers = set(self.unreachable_peers)
+        s.reachable_peers = set(self.reachable_peers)
+        s.problems = self.problems[:]
+        s.bad_shares = self.bad_shares.copy() # tuple->str
+        s.last_update_mode = self.last_update_mode
+        s.last_update_time = self.last_update_time
+        return s
+
     def mark_bad_share(self, peerid, shnum, checkstring):
         """This share was found to be bad, either in the checkstring or
         signature (detected during mapupdate), or deeper in the share
@@ -160,6 +175,13 @@ class ServerMap:
                     for (peerid, shnum)
                     in self.servermap])
 
+    def all_peers_for_version(self, verinfo):
+        """Return a set of peerids that hold shares for the given version."""
+        return set([peerid
+                    for ( (peerid, shnum), (verinfo2, timestamp) )
+                    in self.servermap.items()
+                    if verinfo == verinfo2])
+
     def make_sharemap(self):
         """Return a dict that maps shnum to a set of peerds that hold it."""
         sharemap = DictOfSets()
@@ -209,16 +231,20 @@ class ServerMap:
         seqnums.append(0)
         return max(seqnums)
 
+    def summarize_version(self, verinfo):
+        """Take a versionid, return a string that describes it."""
+        (seqnum, root_hash, IV, segsize, datalength, k, N, prefix,
+         offsets_tuple) = verinfo
+        return "seq%d-%s" % (seqnum, base32.b2a(root_hash)[:4])
+
     def summarize_versions(self):
         """Return a string describing which versions we know about."""
         versionmap = self.make_versionmap()
         bits = []
         for (verinfo, shares) in versionmap.items():
-            (seqnum, root_hash, IV, segsize, datalength, k, N, prefix,
-             offsets_tuple) = verinfo
+            vstr = self.summarize_version(verinfo)
             shnums = set([shnum for (shnum, peerid, timestamp) in shares])
-            bits.append("%d*seq%d-%s" %
-                        (len(shnums), seqnum, base32.b2a(root_hash)[:4]))
+            bits.append("%d*%s" % (len(shnums), vstr))
         return "/".join(bits)
 
     def recoverable_versions(self):
@@ -303,17 +329,23 @@ class ServerMap:
         # same seqnum, meaning that MutableFileNode.read_best_version is not
         # giving you the whole story, and that using its data to do a
         # subsequent publish will lose information.
-        return bool(len(self.recoverable_versions()) > 1)
+        recoverable_seqnums = [verinfo[0]
+                               for verinfo in self.recoverable_versions()]
+        for seqnum in recoverable_seqnums:
+            if recoverable_seqnums.count(seqnum) > 1:
+                return True
+        return False
 
 
 class ServermapUpdater:
-    def __init__(self, filenode, servermap, mode=MODE_READ):
+    def __init__(self, filenode, monitor, servermap, mode=MODE_READ):
         """I update a servermap, locating a sufficient number of useful
         shares and remembering where they are located.
 
         """
 
         self._node = filenode
+        self._monitor = monitor
         self._servermap = servermap
         self.mode = mode
         self._running = True
@@ -326,6 +358,8 @@ class ServermapUpdater:
         self._status.set_progress(0.0)
         self._status.set_mode(mode)
 
+        self._servers_responded = set()
+
         # how much data should we read?
         #  * if we only need the checkstring, then [0:75]
         #  * if we need to validate the checkstring sig, then [543ish:799ish]
@@ -345,6 +379,9 @@ class ServermapUpdater:
         self._need_privkey = False
         if mode == MODE_WRITE and not self._node._privkey:
             self._need_privkey = True
+        # check+repair: repair requires the privkey, so if we didn't happen
+        # to ask for it during the check, we'll have problems doing the
+        # publish.
 
         prefix = storage.si_b2a(self._storage_index)[:5]
         self._log_number = log.msg(format="SharemapUpdater(%(si)s): starting (%(mode)s)",
@@ -356,6 +393,8 @@ class ServermapUpdater:
     def log(self, *args, **kwargs):
         if "parent" not in kwargs:
             kwargs["parent"] = self._log_number
+        if "facility" not in kwargs:
+            kwargs["facility"] = "tahoe.mutable.mapupdate"
         return log.msg(*args, **kwargs)
 
     def update(self):
@@ -507,10 +546,12 @@ class ServermapUpdater:
         now = time.time()
         elapsed = now - started
         self._queries_outstanding.discard(peerid)
+        self._servermap.reachable_peers.add(peerid)
         self._must_query.discard(peerid)
         self._queries_completed += 1
         if not self._running:
-            self.log("but we're not running, so we'll ignore it", parent=lp)
+            self.log("but we're not running, so we'll ignore it", parent=lp,
+                     level=log.NOISY)
             self._status.add_per_server_time(peerid, "late", started, elapsed)
             return
         self._status.add_per_server_time(peerid, "query", started, elapsed)
@@ -532,8 +573,9 @@ class ServermapUpdater:
             except CorruptShareError, e:
                 # log it and give the other shares a chance to be processed
                 f = failure.Failure()
-                self.log("bad share: %s %s" % (f, f.value),
-                         parent=lp, level=log.WEIRD)
+                self.log(format="bad share: %(f_value)s", f_value=str(f.value),
+                         failure=f, parent=lp, level=log.WEIRD, umid="h5llHg")
+                self.notify_server_corruption(peerid, shnum, str(e))
                 self._bad_peers.add(peerid)
                 self._last_failure = f
                 checkstring = data[:SIGNED_PREFIX_LENGTH]
@@ -566,12 +608,18 @@ class ServermapUpdater:
             d.addErrback(self._fatal_error)
 
         # all done!
-        self.log("_got_results done", parent=lp)
+        self.log("_got_results done", parent=lp, level=log.NOISY)
+
+    def notify_server_corruption(self, peerid, shnum, reason):
+        ss = self._servermap.connections[peerid]
+        ss.callRemoteOnly("advise_corrupt_share",
+                          "mutable", self._storage_index, shnum, reason)
 
     def _got_results_one_share(self, shnum, data, peerid, lp):
         self.log(format="_got_results: got shnum #%(shnum)d from peerid %(peerid)s",
                  shnum=shnum,
                  peerid=idlib.shortnodeid_b2a(peerid),
+                 level=log.NOISY,
                  parent=lp)
 
         # this might raise NeedMoreDataError, if the pubkey and signature
@@ -668,7 +716,7 @@ class ServermapUpdater:
         if alleged_writekey != self._node.get_writekey():
             self.log("invalid privkey from %s shnum %d" %
                      (idlib.nodeid_b2a(peerid)[:8], shnum),
-                     parent=lp, level=log.WEIRD)
+                     parent=lp, level=log.WEIRD, umid="aJVccw")
             return
 
         # it's good
@@ -683,14 +731,22 @@ class ServermapUpdater:
 
 
     def _query_failed(self, f, peerid):
-        self.log("error during query: %s %s" % (f, f.value), level=log.WEIRD)
         if not self._running:
             return
+        level = log.WEIRD
+        if f.check(DeadReferenceError):
+            level = log.UNUSUAL
+        self.log(format="error during query: %(f_value)s",
+                 f_value=str(f.value), failure=f,
+                 level=level, umid="IHXuQg")
         self._must_query.discard(peerid)
         self._queries_outstanding.discard(peerid)
         self._bad_peers.add(peerid)
         self._servermap.problems.append(f)
-        self._servermap.unreachable_peers.add(peerid) # TODO: overkill?
+        # a peerid could be in both ServerMap.reachable_peers and
+        # .unreachable_peers if they responded to our query, but then an
+        # exception was raised in _got_results.
+        self._servermap.unreachable_peers.add(peerid)
         self._queries_completed += 1
         self._last_failure = f
 
@@ -702,7 +758,8 @@ class ServermapUpdater:
         if not self._need_privkey:
             return
         if shnum not in datavs:
-            self.log("privkey wasn't there when we asked it", level=log.WEIRD)
+            self.log("privkey wasn't there when we asked it",
+                     level=log.WEIRD, umid="VA9uDQ")
             return
         datav = datavs[shnum]
         enc_privkey = datav[0]
@@ -710,11 +767,14 @@ class ServermapUpdater:
 
     def _privkey_query_failed(self, f, peerid, shnum, lp):
         self._queries_outstanding.discard(peerid)
-        self.log("error during privkey query: %s %s" % (f, f.value),
-                 parent=lp, level=log.WEIRD)
         if not self._running:
             return
-        self._queries_outstanding.discard(peerid)
+        level = log.WEIRD
+        if f.check(DeadReferenceError):
+            level = log.UNUSUAL
+        self.log(format="error during privkey query: %(f_value)s",
+                 f_value=str(f.value), failure=f,
+                 parent=lp, level=level, umid="McoJ5w")
         self._servermap.problems.append(f)
         self._last_failure = f
 
@@ -747,7 +807,7 @@ class ServermapUpdater:
             # a share, so we must continue to wait. No additional queries are
             # required at this time.
             self.log("%d 'must query' peers left" % len(self._must_query),
-                     parent=lp)
+                     level=log.NOISY, parent=lp)
             return
 
         if (not self._queries_outstanding and not self.extra_peers):
@@ -785,11 +845,11 @@ class ServermapUpdater:
                 self.log(format="%(completed)d completed, %(query)d to query: need more",
                          completed=self._queries_completed,
                          query=self.num_peers_to_query,
-                         parent=lp)
+                         level=log.NOISY, parent=lp)
                 return self._send_more_queries(MAX_IN_FLIGHT)
             if not recoverable_versions:
                 self.log("no recoverable versions: need more",
-                         parent=lp)
+                         level=log.NOISY, parent=lp)
                 return self._send_more_queries(MAX_IN_FLIGHT)
             highest_recoverable = max(recoverable_versions)
             highest_recoverable_seqnum = highest_recoverable[0]
@@ -799,7 +859,8 @@ class ServermapUpdater:
                     # don't yet see enough shares to recover it. Try harder.
                     # TODO: consider sending more queries.
                     # TODO: consider limiting the search distance
-                    self.log("evidence of higher seqnum: need more")
+                    self.log("evidence of higher seqnum: need more",
+                             level=log.UNUSUAL, parent=lp)
                     return self._send_more_queries(MAX_IN_FLIGHT)
             # all the unrecoverable versions were old or concurrent with a
             # recoverable version. Good enough.
@@ -813,7 +874,8 @@ class ServermapUpdater:
             # every server in the world.
 
             if not recoverable_versions:
-                self.log("no recoverable versions: need more", parent=lp)
+                self.log("no recoverable versions: need more", parent=lp,
+                         level=log.NOISY)
                 return self._send_more_queries(MAX_IN_FLIGHT)
 
             last_found = -1
@@ -837,7 +899,7 @@ class ServermapUpdater:
                         if num_not_found >= self.EPSILON:
                             self.log("found our boundary, %s" %
                                      "".join(states),
-                                     parent=lp)
+                                     parent=lp, level=log.NOISY)
                             found_boundary = True
                             break
 
@@ -860,11 +922,11 @@ class ServermapUpdater:
                 if last_not_responded == -1:
                     # we're done
                     self.log("have all our answers",
-                             parent=lp)
+                             parent=lp, level=log.NOISY)
                     # .. unless we're still waiting on the privkey
                     if self._need_privkey:
                         self.log("but we're still waiting for the privkey",
-                                 parent=lp)
+                                 parent=lp, level=log.NOISY)
                         # if we found the boundary but we haven't yet found
                         # the privkey, we may need to look further. If
                         # somehow all the privkeys were corrupted (but the
@@ -877,13 +939,14 @@ class ServermapUpdater:
 
             # if we hit here, we didn't find our boundary, so we're still
             # waiting for peers
-            self.log("no boundary yet, %s" % "".join(states), parent=lp)
+            self.log("no boundary yet, %s" % "".join(states), parent=lp,
+                     level=log.NOISY)
             return self._send_more_queries(MAX_IN_FLIGHT)
 
         # otherwise, keep up to 5 queries in flight. TODO: this is pretty
         # arbitrary, really I want this to be something like k -
         # max(known_version_sharecounts) + some extra
-        self.log("catchall: need more", parent=lp)
+        self.log("catchall: need more", parent=lp, level=log.NOISY)
         return self._send_more_queries(MAX_IN_FLIGHT)
 
     def _send_more_queries(self, num_outstanding):
@@ -929,7 +992,7 @@ class ServermapUpdater:
         eventually(self._done_deferred.callback, self._servermap)
 
     def _fatal_error(self, f):
-        self.log("fatal error", failure=f, level=log.WEIRD)
+        self.log("fatal error", failure=f, level=log.WEIRD, umid="1cNvlw")
         self._done_deferred.errback(f)
 
 
