@@ -15,7 +15,7 @@ class VDriveOptions(BaseOptions, usage.Options):
          "the mapping from alias name to root dirnode URI."
          ],
         ["node-url", "u", None,
-         "URL of the tahoe node to use, a URL like \"http://127.0.0.1:8123\". "
+         "URL of the tahoe node to use, a URL like \"http://127.0.0.1:3456\". "
          "This overrides the URL found in the --node-directory ."],
         ["dir-cap", None, None,
          "Which dirnode URI should be used as the 'tahoe' alias."]
@@ -39,6 +39,8 @@ class VDriveOptions(BaseOptions, usage.Options):
         else:
             node_url_file = os.path.join(self['node-directory'], "node.url")
             self['node-url'] = open(node_url_file, "r").read().strip()
+        if self['node-url'][-1] != "/":
+            self['node-url'] += "/"
 
         aliases = get_aliases(self['node-directory'])
         if self['dir-cap']:
@@ -56,6 +58,10 @@ class AddAliasOptions(VDriveOptions):
         self.alias = alias
         self.cap = cap
 
+class CreateAliasOptions(VDriveOptions):
+    def parseArgs(self, alias):
+        self.alias = alias
+
 class ListAliasOptions(VDriveOptions):
     pass
 
@@ -108,9 +114,10 @@ class PutOptions(VDriveOptions):
         ]
 
     def parseArgs(self, arg1=None, arg2=None):
-        # cat FILE > tahoe put           # create unlinked file from stdin
-        # cat FILE > tahoe put FOO       # create tahoe:FOO from stdin
-        # cat FILE > tahoe put tahoe:FOO # same
+        # cat FILE | tahoe put           # create unlinked file from stdin
+        # cat FILE | tahoe put -         # same
+        # tahoe put bar                  # create unlinked file from local 'bar'
+        # cat FILE | tahoe put - FOO     # create tahoe:FOO from stdin
         # tahoe put bar FOO              # copy local 'bar' to tahoe:FOO
         # tahoe put bar tahoe:FOO        # same
 
@@ -118,11 +125,11 @@ class PutOptions(VDriveOptions):
             self.from_file = arg1
             self.to_file = arg2
         elif arg1 is not None and arg2 is None:
-            self.from_file = None
-            self.to_file = arg1
+            self.from_file = arg1 # might be "-"
+            self.to_file = None
         else:
-            self.from_file = arg1
-            self.to_file = arg2
+            self.from_file = None
+            self.to_file = None
         if self.from_file == "-":
             self.from_file = None
 
@@ -130,19 +137,22 @@ class PutOptions(VDriveOptions):
         return "%s put LOCAL_FILE VDRIVE_FILE" % (os.path.basename(sys.argv[0]),)
 
     longdesc = """Put a file into the virtual drive (copying the file's
-    contents from the local filesystem). If LOCAL_FILE is missing or '-',
-    data will be copied from stdin. VDRIVE_FILE is assumed to start with
-    tahoe: unless otherwise specified."""
+    contents from the local filesystem). If VDRIVE_FILE is missing, upload
+    the file but do not link it into a directory: prints the new filecap to
+    stdout. If LOCAL_FILE is missing or '-', data will be copied from stdin.
+    VDRIVE_FILE is assumed to start with tahoe: unless otherwise specified."""
 
     def getUsage(self, width=None):
         t = VDriveOptions.getUsage(self, width)
         t += """
 Examples:
- % cat FILE > tahoe put           # create unlinked file from stdin
- % cat FILE > tahoe put FOO       # create tahoe:FOO from stdin
- % cat FILE > tahoe put tahoe:FOO # same
- % tahoe put bar FOO              # copy local 'bar' to tahoe:FOO
- % tahoe put bar tahoe:FOO        # same
+ % cat FILE | tahoe put                # create unlinked file from stdin
+ % cat FILE | tahoe -                  # same
+ % tahoe put bar                       # create unlinked file from local 'bar'
+ % cat FILE | tahoe put - FOO          # create tahoe:FOO from stdin
+ % tahoe put bar FOO                   # copy local 'bar' to tahoe:FOO
+ % tahoe put bar tahoe:FOO             # same
+ % tahoe put bar MUTABLE-FILE-WRITECAP # modify the mutable file in-place
 """
         return t
 
@@ -180,18 +190,90 @@ class LnOptions(VDriveOptions):
     def getSynopsis(self):
         return "%s ln FROM TO" % (os.path.basename(sys.argv[0]),)
 
+class BackupOptions(VDriveOptions):
+    optFlags = [
+        ("verbose", "v", "Be noisy about what is happening."),
+        ("no-backupdb", None, "Do not use the SQLite-based backup-database (always upload all files)."),
+        ("ignore-timestamps", None, "Do not use backupdb timestamps to decide if a local file is unchanged."),
+        ]
+
+    def parseArgs(self, localdir, topath):
+        self.from_dir = localdir
+        self.to_dir = topath
+
+    def getSynopsis(Self):
+        return "%s backup FROM ALIAS:TO" % os.path.basename(sys.argv[0])
+
+    longdesc = """Add a versioned backup of the local FROM directory to a timestamped subdir of the (tahoe) TO/Archives directory, sharing as many files and directories as possible with the previous backup. Creates TO/Latest as a reference to the latest backup. Behaves somewhat like 'rsync -a --link-dest=TO/Archives/(previous) FROM TO/Archives/(new); ln -sf TO/Archives/(new) TO/Latest'."""
+
 class WebopenOptions(VDriveOptions):
-    def parseArgs(self, vdrive_pathname=""):
-        self['vdrive_pathname'] = vdrive_pathname
+    def parseArgs(self, where=''):
+        self.where = where
+
+    def getSynopsis(self):
+        return "%s webopen [ALIAS:PATH]" % (os.path.basename(sys.argv[0]),)
 
     longdesc = """Opens a webbrowser to the contents of some portion of the virtual drive."""
 
-class ReplOptions(usage.Options):
-    pass
+class ManifestOptions(VDriveOptions):
+    optFlags = [
+        ("storage-index", "s", "Only print storage index strings, not pathname+cap"),
+        ("verify-cap", None, "Only print verifycap, not pathname+cap"),
+        ("repair-cap", None, "Only print repaircap, not pathname+cap"),
+        ("raw", "r", "Display raw JSON data instead of parsed"),
+        ]
+    def parseArgs(self, where=''):
+        self.where = where
+
+    def getSynopsis(self):
+        return "%s manifest [ALIAS:PATH]" % (os.path.basename(sys.argv[0]),)
+
+    longdesc = """Print a list of all files/directories reachable from the given starting point."""
+
+class StatsOptions(VDriveOptions):
+    optFlags = [
+        ("raw", "r", "Display raw JSON data instead of parsed"),
+        ]
+    def parseArgs(self, where=''):
+        self.where = where
+
+    def getSynopsis(self):
+        return "%s stats [ALIAS:PATH]" % (os.path.basename(sys.argv[0]),)
+
+    longdesc = """Print statistics about of all files/directories reachable from the given starting point."""
+
+class CheckOptions(VDriveOptions):
+    optFlags = [
+        ("raw", "r", "Display raw JSON data instead of parsed"),
+        ("verify", "v", "Verify all hashes, instead of merely querying share presence"),
+        ("repair", "r", "Automatically repair any problems found"),
+        ]
+    def parseArgs(self, where=''):
+        self.where = where
+
+    def getSynopsis(self):
+        return "%s check [ALIAS:PATH]" % (os.path.basename(sys.argv[0]),)
+
+    longdesc = """Check a single file or directory: count how many shares are available, verify their hashes. Optionally repair the file if any problems were found."""
+
+class DeepCheckOptions(VDriveOptions):
+    optFlags = [
+        ("raw", "r", "Display raw JSON data instead of parsed"),
+        ("verify", "v", "Verify all hashes, instead of merely querying share presence"),
+        ("repair", "r", "Automatically repair any problems found"),
+        ]
+    def parseArgs(self, where=''):
+        self.where = where
+
+    def getSynopsis(self):
+        return "%s deep-check [ALIAS:PATH]" % (os.path.basename(sys.argv[0]),)
+
+    longdesc = """Check all files/directories reachable from the given starting point (which must be a directory), like 'tahoe check' but for multiple files. Optionally repair any problems found."""
 
 subCommands = [
     ["mkdir", None, MakeDirectoryOptions, "Create a new directory"],
     ["add-alias", None, AddAliasOptions, "Add a new alias cap"],
+    ["create-alias", None, CreateAliasOptions, "Create a new alias cap"],
     ["list-aliases", None, ListAliasOptions, "List all alias caps"],
     ["ls", None, ListOptions, "List a directory"],
     ["get", None, GetOptions, "Retrieve a file from the virtual drive."],
@@ -200,141 +282,112 @@ subCommands = [
     ["rm", None, RmOptions, "Unlink a file or directory in the virtual drive."],
     ["mv", None, MvOptions, "Move a file within the virtual drive."],
     ["ln", None, LnOptions, "Make an additional link to an existing file."],
+    ["backup", None, BackupOptions, "Make target dir look like local dir."],
     ["webopen", None, WebopenOptions, "Open a webbrowser to the root_dir"],
-    ["repl", None, ReplOptions, "Open a python interpreter"],
+    ["manifest", None, ManifestOptions, "List all files/dirs in a subtree"],
+    ["stats", None, StatsOptions, "Print statistics about all files/dirs in a subtree"],
+    ["check", None, CheckOptions, "Check a single file or directory"],
+    ["deep-check", None, DeepCheckOptions, "Check all files/directories reachable from a starting point"],
     ]
 
-def mkdir(config, stdout, stderr):
+def mkdir(options):
     from allmydata.scripts import tahoe_mkdir
-    rc = tahoe_mkdir.mkdir(config['node-url'],
-                           config.aliases,
-                           config.where,
-                           stdout, stderr)
+    rc = tahoe_mkdir.mkdir(options)
     return rc
 
-def add_alias(config, stdout, stderr):
+def add_alias(options):
     from allmydata.scripts import tahoe_add_alias
-    rc = tahoe_add_alias.add_alias(config['node-directory'],
-                                   config.alias,
-                                   config.cap,
-                                   stdout, stderr)
+    rc = tahoe_add_alias.add_alias(options)
     return rc
 
-def list_aliases(config, stdout, stderr):
+def create_alias(options):
     from allmydata.scripts import tahoe_add_alias
-    rc = tahoe_add_alias.list_aliases(config['node-directory'],
-                                      stdout, stderr)
+    rc = tahoe_add_alias.create_alias(options)
     return rc
 
-def list(config, stdout, stderr):
+def list_aliases(options):
+    from allmydata.scripts import tahoe_add_alias
+    rc = tahoe_add_alias.list_aliases(options)
+    return rc
+
+def list(options):
     from allmydata.scripts import tahoe_ls
-    rc = tahoe_ls.list(config['node-url'],
-                       config.aliases,
-                       config.where,
-                       config,
-                       stdout, stderr)
+    rc = tahoe_ls.list(options)
     return rc
 
-def get(config, stdout, stderr):
+def get(options):
     from allmydata.scripts import tahoe_get
-    rc = tahoe_get.get(config['node-url'],
-                       config.aliases,
-                       config.from_file,
-                       config.to_file,
-                       stdout, stderr)
+    rc = tahoe_get.get(options)
     if rc == 0:
-        if config.to_file is None:
+        if options.to_file is None:
             # be quiet, since the file being written to stdout should be
             # proof enough that it worked, unless the user is unlucky
             # enough to have picked an empty file
             pass
         else:
-            print >>stderr, "%s retrieved and written to %s" % \
-                  (config.from_file, config.to_file)
+            print >>options.stderr, "%s retrieved and written to %s" % \
+                  (options.from_file, options.to_file)
     return rc
 
-def put(config, stdout, stderr, stdin=sys.stdin):
+def put(options):
     from allmydata.scripts import tahoe_put
-    if config['quiet']:
-        verbosity = 0
-    else:
-        verbosity = 2
-    rc = tahoe_put.put(config['node-url'],
-                       config.aliases,
-                       config.from_file,
-                       config.to_file,
-                       config['mutable'],
-                       verbosity,
-                       stdin, stdout, stderr)
+    rc = tahoe_put.put(options)
     return rc
 
-def cp(config, stdout, stderr):
+def cp(options):
     from allmydata.scripts import tahoe_cp
-    if config['quiet']:
-        verbosity = 0
-    else:
-        verbosity = 2
-    rc = tahoe_cp.copy(config['node-url'],
-                       config,
-                       config.aliases,
-                       config.sources,
-                       config.destination,
-                       verbosity,
-                       stdout, stderr)
+    rc = tahoe_cp.copy(options)
     return rc
 
-def rm(config, stdout, stderr):
+def rm(options):
     from allmydata.scripts import tahoe_rm
-    if config['quiet']:
-        verbosity = 0
-    else:
-        verbosity = 2
-    rc = tahoe_rm.rm(config['node-url'],
-                     config.aliases,
-                     config.where,
-                     verbosity,
-                     stdout, stderr)
+    rc = tahoe_rm.rm(options)
     return rc
 
-def mv(config, stdout, stderr):
+def mv(options):
     from allmydata.scripts import tahoe_mv
-    rc = tahoe_mv.mv(config['node-url'],
-                     config.aliases,
-                     config.from_file,
-                     config.to_file,
-                     stdout, stderr,
-                     mode="move")
+    rc = tahoe_mv.mv(options, mode="move")
     return rc
 
-def ln(config, stdout, stderr):
+def ln(options):
     from allmydata.scripts import tahoe_mv
-    rc = tahoe_mv.mv(config['node-url'],
-                     config.aliases,
-                     config.from_file,
-                     config.to_file,
-                     stdout, stderr,
-                     mode="link")
+    rc = tahoe_mv.mv(options, mode="link")
     return rc
 
-def webopen(config, stdout, stderr):
-    import urllib, webbrowser
-    nodeurl = config['node-url']
-    if nodeurl[-1] != "/":
-        nodeurl += "/"
-    root_cap = config.aliases["tahoe"]
-    url = nodeurl + "uri/%s/" % urllib.quote(root_cap)
-    if config['vdrive_pathname']:
-        url += urllib.quote(config['vdrive_pathname'])
-    webbrowser.open(url)
-    return 0
+def backup(options):
+    from allmydata.scripts import tahoe_backup
+    rc = tahoe_backup.backup(options)
+    return rc
 
-def repl(config, stdout, stderr):
-    import code
-    return code.interact()
+def webopen(options, opener=None):
+    from allmydata.scripts import tahoe_webopen
+    rc = tahoe_webopen.webopen(options, opener=opener)
+    return rc
+
+def manifest(options):
+    from allmydata.scripts import tahoe_manifest
+    rc = tahoe_manifest.manifest(options)
+    return rc
+
+def stats(options):
+    from allmydata.scripts import tahoe_manifest
+    rc = tahoe_manifest.stats(options)
+    return rc
+
+def check(options):
+    from allmydata.scripts import tahoe_check
+    rc = tahoe_check.check(options)
+    return rc
+
+def deepcheck(options):
+    from allmydata.scripts import tahoe_check
+    rc = tahoe_check.deepcheck(options)
+    return rc
 
 dispatch = {
     "mkdir": mkdir,
     "add-alias": add_alias,
+    "create-alias": create_alias,
     "list-aliases": list_aliases,
     "ls": list,
     "get": get,
@@ -343,7 +396,11 @@ dispatch = {
     "rm": rm,
     "mv": mv,
     "ln": ln,
+    "backup": backup,
     "webopen": webopen,
-    "repl": repl,
+    "manifest": manifest,
+    "stats": stats,
+    "check": check,
+    "deep-check": deepcheck,
     }
 
