@@ -4,9 +4,15 @@ from twisted.python import usage
 
 
 class BaseOptions:
+    # unit tests can override these to point at StringIO instances
+    stdin = sys.stdin
+    stdout = sys.stdout
+    stderr = sys.stderr
+
     optFlags = [
         ["quiet", "q", "Operate silently."],
         ["version", "V", "Display version numbers and exit."],
+        ["version-and-path", None, "Display version numbers and paths to their locations and exit."],
         ]
 
     def opt_version(self):
@@ -14,6 +20,11 @@ class BaseOptions:
         print allmydata.get_package_versions_string()
         sys.exit(0)
 
+    def opt_version_and_path(self):
+        import allmydata
+        print allmydata.get_package_versions_string(show_paths=True)
+        sys.exit(0)
+
 
 class BasedirMixin:
     optFlags = [
@@ -25,37 +36,48 @@ class BasedirMixin:
             raise usage.UsageError("<basedir> parameter is required")
         if self['basedir']:
             del self['basedir']
-        self['basedirs'] = [os.path.abspath(os.path.expanduser(b))
-                            for b in self.basedirs]
+        self['basedirs'] = [os.path.abspath(os.path.expanduser(b)) for b in self.basedirs]
 
     def parseArgs(self, *args):
+        from allmydata.util.assertutil import precondition
         self.basedirs = []
         if self['basedir']:
+            precondition(isinstance(self['basedir'], (str, unicode)), self['basedir'])
             self.basedirs.append(self['basedir'])
         if self['multiple']:
+            precondition(not [x for x in args if not isinstance(x, (str, unicode))], args)
             self.basedirs.extend(args)
         else:
             if len(args) == 0 and not self.basedirs:
                 if sys.platform == 'win32':
                     from allmydata.windows import registry
-                    self.basedirs.append(registry.get_base_dir_path())
+                    rbdp = registry.get_base_dir_path()
+                    if rbdp:
+                        precondition(isinstance(registry.get_base_dir_path(), (str, unicode)), registry.get_base_dir_path())
+                        self.basedirs.append(rbdp)
                 else:
+                    precondition(isinstance(os.path.expanduser("~/.tahoe"), (str, unicode)), os.path.expanduser("~/.tahoe"))
                     self.basedirs.append(os.path.expanduser("~/.tahoe"))
             if len(args) > 0:
+                precondition(isinstance(args[0], (str, unicode)), args[0])
                 self.basedirs.append(args[0])
             if len(args) > 1:
                 raise usage.UsageError("I wasn't expecting so many arguments")
 
 class NoDefaultBasedirMixin(BasedirMixin):
     def parseArgs(self, *args):
+        from allmydata.util.assertutil import precondition
         # create-client won't default to --basedir=~/.tahoe
         self.basedirs = []
         if self['basedir']:
+            precondition(isinstance(self['basedir'], (str, unicode)), self['basedir'])
             self.basedirs.append(self['basedir'])
         if self['multiple']:
+            precondition(not [x for x in args if not isinstance(x, (str, unicode))], args)
             self.basedirs.extend(args)
         else:
             if len(args) > 0:
+                precondition(isinstance(args[0], (str, unicode)), args[0])
                 self.basedirs.append(args[0])
             if len(args) > 1:
                 raise usage.UsageError("I wasn't expecting so many arguments")
@@ -81,7 +103,7 @@ def get_aliases(nodedir):
         f = open(aliasfile, "r")
         for line in f.readlines():
             line = line.strip()
-            if line.startswith("#"):
+            if line.startswith("#") or not line:
                 continue
             name, cap = line.split(":", 1)
             # normalize it: remove http: prefix, urldecode
@@ -99,6 +121,7 @@ def get_alias(aliases, path, default):
     # If default=None, then an empty alias is indicated by returning
     # DefaultAliasMarker. We special-case "URI:" to make it easy to access
     # specific files/directories by their read-cap.
+    path = path.strip()
     if path.startswith("URI:"):
         # The only way to get a sub-path is to use URI:blah:./foo, and we
         # strip out the :./ sequence.
