@@ -10,6 +10,7 @@ class StartOptions(BasedirMixin, usage.Options):
         ]
     optFlags = [
         ["profile", "p", "whether to run under the Python profiler, putting results in \"profiling_results.prof\""],
+        ["syslog", None, "tell the node to log to syslog, not a file"],
         ]
 
 class StopOptions(BasedirMixin, usage.Options):
@@ -26,6 +27,7 @@ class RestartOptions(BasedirMixin, usage.Options):
          "instead of complaining that you should have used 'start' instead "
          "of 'restart'"],
         ["profile", "p", "whether to run under the Python profiler, putting results in \"profiling_results.prof\""],
+        ["syslog", None, "tell the node to log to syslog, not a file"],
         ]
 
 class RunOptions(usage.Options):
@@ -33,7 +35,7 @@ class RunOptions(usage.Options):
         ["basedir", "C", None, "which directory to run the node in, CWD by default"],
         ]
 
-def do_start(basedir, profile=False, out=sys.stdout, err=sys.stderr):
+def do_start(basedir, opts, out=sys.stdout, err=sys.stderr):
     print >>out, "STARTING", basedir
     if not os.path.isdir(basedir):
         print >>err, "%s does not look like a directory at all" % basedir
@@ -53,15 +55,46 @@ def do_start(basedir, profile=False, out=sys.stdout, err=sys.stderr):
         nodetype = "unknown (%s)" % tac
 
     cmd = find_exe.find_exe('twistd')
+    if not cmd:
+        # If 'twistd' wasn't on $PATH, maybe we're running from source and
+        # Twisted was built as one of our dependencies. If so, we're at
+        # BASEDIR/src/allmydata/scripts/startstop_node.py, and it's at
+        # BASEDIR/support/$BINDIR/twistd
+        up = os.path.dirname
+        TAHOEDIR = up(up(up(up(os.path.abspath(__file__)))))
+        if sys.platform == "win32":
+            bin_dir = "Scripts"
+        else:
+            bin_dir = "bin"
+        bindir = os.path.join(TAHOEDIR, "support", bin_dir)
+
+        maybe = os.path.join(bindir, "twistd")
+        if os.path.exists(maybe):
+            cmd = [maybe]
+            oldpath = os.environ.get("PATH", "").split(os.pathsep)
+            os.environ["PATH"] = os.pathsep.join(oldpath + [bindir])
+            # sys.path and $PYTHONPATH are taken care of by the extra code in
+            # 'setup.py trial'
+        else:
+            maybe = maybe+'.py'
+            if os.path.exists(maybe):
+                cmd = [sys.executable, maybe]
+                oldpath = os.environ.get("PATH", "").split(os.pathsep)
+                os.environ["PATH"] = os.pathsep.join(oldpath + [bindir])
+                # sys.path and $PYTHONPATH are taken care of by the extra code in
+                # 'setup.py trial'
+
     if not cmd:
         print "Can't find twistd (it comes with Twisted).  Aborting."
         sys.exit(1)
 
     cmd.extend(["-y", tac])
-    if nodetype in ("client", "introducer"):
+    if opts["syslog"]:
+        cmd.append("--syslog")
+    elif nodetype in ("client", "introducer"):
         fileutil.make_dirs(os.path.join(basedir, "logs"))
         cmd.extend(["--logfile", os.path.join("logs", "twistd.log")])
-    if profile:
+    if opts["profile"]:
         cmd.extend(["--profile=profiling_results.prof", "--savestats",])
     curdir = os.getcwd()
     try:
@@ -86,7 +119,7 @@ def do_stop(basedir, out=sys.stdout, err=sys.stderr):
     pid = int(pid)
 
     # kill it hard (SIGKILL), delete the twistd.pid file, then wait for the
-    # process itself to go away. If it hasn't gone away after 5 seconds, warn
+    # process itself to go away. If it hasn't gone away after 20 seconds, warn
     # the user but keep waiting until they give up.
     try:
         os.kill(pid, signal.SIGKILL)
@@ -104,7 +137,7 @@ def do_stop(basedir, out=sys.stdout, err=sys.stderr):
         pass
     start = time.time()
     time.sleep(0.1)
-    wait = 5
+    wait = 40
     first_time = True
     while True:
         # poll once per second until we see the process is no longer running
@@ -132,7 +165,7 @@ def do_stop(basedir, out=sys.stdout, err=sys.stderr):
 def start(config, stdout, stderr):
     rc = 0
     for basedir in config['basedirs']:
-        rc = do_start(basedir, config['profile'], stdout, stderr) or rc
+        rc = do_start(basedir, config, stdout, stderr) or rc
     return rc
 
 def stop(config, stdout, stderr):
@@ -152,7 +185,7 @@ def restart(config, stdout, stderr):
         print >>stderr, "not restarting"
         return rc
     for basedir in config['basedirs']:
-        rc = do_start(basedir, config['profile'], stdout, stderr) or rc
+        rc = do_start(basedir, config, stdout, stderr) or rc
     return rc
 
 def run(config, stdout, stderr):
