@@ -117,9 +117,11 @@ class LocalDirectorySource:
                 self.children[n] = child
                 if recurse:
                     child.populate(True)
-            else:
-                assert os.path.isfile(pn)
+            elif os.path.isfile(pn):
                 self.children[n] = LocalFileSource(pn)
+            else:
+                # Could be dangling symlink; probably not copy-able.
+                pass
 
 class LocalDirectoryTarget:
     def __init__(self, progressfunc, pathname):
@@ -222,7 +224,9 @@ class TahoeDirectorySource:
         nodetype, d = parsed
         assert nodetype == "dirnode"
         self.mutable = d.get("mutable", False) # older nodes don't provide it
-        self.children_d = d["children"]
+        self.children_d = dict( [(unicode(name),value)
+                                 for (name,value)
+                                 in d["children"].iteritems()] )
         self.children = None
 
     def init_from_parsed(self, parsed):
@@ -230,7 +234,9 @@ class TahoeDirectorySource:
         self.writecap = ascii_or_none(d.get("rw_uri"))
         self.readcap = ascii_or_none(d.get("ro_uri"))
         self.mutable = d.get("mutable", False) # older nodes don't provide it
-        self.children_d = d["children"]
+        self.children_d = dict( [(unicode(name),value)
+                                 for (name,value)
+                                 in d["children"].iteritems()] )
         self.children = None
 
     def populate(self, recurse):
@@ -294,7 +300,9 @@ class TahoeDirectoryTarget:
         self.writecap = ascii_or_none(d.get("rw_uri"))
         self.readcap = ascii_or_none(d.get("ro_uri"))
         self.mutable = d.get("mutable", False) # older nodes don't provide it
-        self.children_d = d["children"]
+        self.children_d = dict( [(unicode(name),value)
+                                 for (name,value)
+                                 in d["children"].iteritems()] )
         self.children = None
 
     def init_from_grid(self, writecap, readcap):
@@ -308,7 +316,9 @@ class TahoeDirectoryTarget:
         nodetype, d = parsed
         assert nodetype == "dirnode"
         self.mutable = d.get("mutable", False) # older nodes don't provide it
-        self.children_d = d["children"]
+        self.children_d = dict( [(unicode(name),value)
+                                 for (name,value)
+                                 in d["children"].iteritems()] )
         self.children = None
 
     def just_created(self, writecap):
@@ -332,7 +342,7 @@ class TahoeDirectoryTarget:
                 if self.writecap:
                     url = self.nodeurl + "/".join(["uri",
                                                    urllib.quote(self.writecap),
-                                                   urllib.quote(name)])
+                                                   urllib.quote(name.encode('utf-8'))])
                 self.children[name] = TahoeFileTarget(self.nodeurl, mutable,
                                                       writecap, readcap, url)
             else:
@@ -397,29 +407,31 @@ class TahoeDirectoryTarget:
         POST(url, body)
 
 class Copier:
-    def __init__(self, nodeurl, config, aliases,
-                 verbosity, stdout, stderr,
-                 progressfunc=None):
+
+    def do_copy(self, options, progressfunc=None):
+        if options['quiet']:
+            verbosity = 0
+        else:
+            verbosity = 2
+
+        nodeurl = options['node-url']
         if nodeurl[-1] != "/":
             nodeurl += "/"
         self.nodeurl = nodeurl
         self.progressfunc = progressfunc
-        self.config = config
-        self.aliases = aliases
+        self.options = options
+        self.aliases = options.aliases
         self.verbosity = verbosity
-        if config["verbose"] and not self.progressfunc:
+        self.stdout = options.stdout
+        self.stderr = options.stderr
+        if options["verbose"] and not self.progressfunc:
             def progress(message):
-                print >>stderr, message
+                print >>self.stderr, message
             self.progressfunc = progress
-        self.stdout = stdout
-        self.stderr = stderr
         self.cache = {}
-
-    def to_stderr(self, text):
-        print >>self.stderr, text
-
-    def do_copy(self, source_specs, destination_spec):
-        recursive = self.config["recursive"]
+        source_specs = options.sources
+        destination_spec = options.destination
+        recursive = self.options["recursive"]
 
         target = self.get_target_info(destination_spec)
 
@@ -470,6 +482,9 @@ class Copier:
         self.to_stderr("unknown target")
         return 1
 
+    def to_stderr(self, text):
+        print >>self.stderr, text
+
     def get_target_info(self, destination_spec):
         rootcap, path = get_alias(self.aliases, destination_spec, None)
         if rootcap == DefaultAliasMarker:
@@ -705,7 +720,5 @@ class Copier:
         return graphs
 
 
-def copy(nodeurl, config, aliases, sources, destination,
-         verbosity, stdout, stderr):
-    c = Copier(nodeurl, config, aliases, verbosity, stdout, stderr)
-    return c.do_copy(sources, destination)
+def copy(options):
+    return Copier().do_copy(options)
