@@ -1,17 +1,14 @@
 import os, re, weakref, stat, struct, time
-from distutils.version import LooseVersion
 
 from foolscap import Referenceable
 from twisted.application import service
-from twisted.internet import defer
 
 from zope.interface import implements
 from allmydata.interfaces import RIStorageServer, RIBucketWriter, \
-     RIBucketReader, IStorageBucketWriter, IStorageBucketReader, HASH_SIZE, \
-     BadWriteEnablerError, IStatsProducer, FileTooLargeError
-from allmydata.util import base32, fileutil, idlib, mathutil, log
-from allmydata.util.assertutil import precondition, _assert
-import allmydata # for __version__
+     RIBucketReader, BadWriteEnablerError, IStatsProducer
+from allmydata.util import base32, fileutil, idlib, log, time_format
+from allmydata.util.assertutil import precondition
+import allmydata # for __full_version__
 
 class DataTooLargeError(Exception):
     pass
@@ -36,9 +33,9 @@ NUM_RE=re.compile("^[0-9]+$")
 
 # The share file has the following layout:
 #  0x00: share file version number, four bytes, current version is 1
-#  0x04: share data length, four bytes big-endian = A
+#  0x04: share data length, four bytes big-endian = A # See Footnote 1 below.
 #  0x08: number of leases, four bytes big-endian
-#  0x0c: beginning of share data (described below, at WriteBucketProxy)
+#  0x0c: beginning of share data (see immutable.layout.WriteBucketProxy)
 #  A+0x0c = B: first lease. Lease format is:
 #   B+0x00: owner number, 4 bytes big-endian, 0 is reserved for no-owner
 #   B+0x04: renew secret, 32 bytes (SHA256)
@@ -46,6 +43,12 @@ NUM_RE=re.compile("^[0-9]+$")
 #   B+0x44: expiration time, 4 bytes big-endian seconds-since-epoch
 #   B+0x48: next lease, or end of record
 
+# Footnote 1: as of Tahoe v1.3.0 this field is not used by storage servers, but it is still
+# filled in by storage servers in case the storage server software gets downgraded from >= Tahoe
+# v1.3.0 to < Tahoe v1.3.0, or the share file is moved from one storage server to another.  The
+# value stored in this field is truncated, so If the actual share data length is >= 2**32, then
+# the value stored in this field will be the actual share data length modulo 2**32.
+
 def si_b2a(storageindex):
     return base32.b2a(storageindex)
 
@@ -98,30 +101,62 @@ class LeaseInfo:
 class ShareFile:
     LEASE_SIZE = struct.calcsize(">L32s32sL")
 
-    def __init__(self, filename):
+    def __init__(self, filename, max_size=None, create=False):
+        """ If max_size is not None then I won't allow more than max_size to be written to me. If create=True and max_size must not be None. """
+        precondition((max_size is not None) or (not create), max_size, create)
         self.home = filename
-        f = open(self.home, 'rb')
-        (version, size, num_leases) = struct.unpack(">LLL", f.read(0xc))
-        assert version == 1
-        self._size = size
-        self._num_leases = num_leases
+        self._max_size = max_size
+        if create:
+            # touch the file, so later callers will see that we're working on it.
+            # Also construct the metadata.
+            assert not os.path.exists(self.home)
+            fileutil.make_dirs(os.path.dirname(self.home))
+            f = open(self.home, 'wb')
+            # The second field -- the four-byte share data length -- is no
+            # longer used as of Tahoe v1.3.0, but we continue to write it in
+            # there in case someone downgrades a storage server from >=
+            # Tahoe-1.3.0 to < Tahoe-1.3.0, or moves a share file from one
+            # server to another, etc. We do saturation -- a share data length
+            # larger than 2**32-1 (what can fit into the field) is marked as
+            # the largest length that can fit into the field. That way, even
+            # if this does happen, the old < v1.3.0 server will still allow
+            # clients to read the first part of the share.
+            f.write(struct.pack(">LLL", 1, min(2**32-1, max_size), 0))
+            f.close()
+            self._lease_offset = max_size + 0x0c
+            self._num_leases = 0
+        else:
+            f = open(self.home, 'rb')
+            filesize = os.path.getsize(self.home)
+            (version, unused, num_leases) = struct.unpack(">LLL", f.read(0xc))
+            f.close()
+            assert version == 1, version
+            self._num_leases = num_leases
+            self._lease_offset = filesize - (num_leases * self.LEASE_SIZE)
         self._data_offset = 0xc
-        self._lease_offset = 0xc + self._size
 
     def unlink(self):
         os.unlink(self.home)
 
     def read_share_data(self, offset, length):
         precondition(offset >= 0)
-        precondition(offset+length <= self._size)
+        # reads beyond the end of the data are truncated. Reads that start beyond the end of the
+        # data return an empty string.
+        # I wonder why Python doesn't do the following computation for me?
+        seekpos = self._data_offset+offset
+        fsize = os.path.getsize(self.home)
+        actuallength = max(0, min(length, fsize-seekpos))
+        if actuallength == 0:
+            return ""
         f = open(self.home, 'rb')
-        f.seek(self._data_offset+offset)
-        return f.read(length)
+        f.seek(seekpos)
+        return f.read(actuallength)
 
     def write_share_data(self, offset, data):
         length = len(data)
-        precondition(offset >= 0)
-        precondition(offset+length <= self._size)
+        precondition(offset >= 0, offset)
+        if self._max_size is not None and offset+length > self._max_size:
+            raise DataTooLargeError(self._max_size, offset, length)
         f = open(self.home, 'rb+')
         real_offset = self._data_offset+offset
         f.seek(real_offset)
@@ -151,7 +186,7 @@ class ShareFile:
         """Yields (ownernum, renew_secret, cancel_secret, expiration_time)
         for all leases."""
         f = open(self.home, 'rb')
-        (version, size, num_leases) = struct.unpack(">LLL", f.read(0xc))
+        (version, unused, num_leases) = struct.unpack(">LLL", f.read(0xc))
         f.seek(self._lease_offset)
         for i in range(num_leases):
             data = f.read(self.LEASE_SIZE)
@@ -224,29 +259,22 @@ class ShareFile:
 class BucketWriter(Referenceable):
     implements(RIBucketWriter)
 
-    def __init__(self, ss, incominghome, finalhome, size, lease_info, canary):
+    def __init__(self, ss, incominghome, finalhome, max_size, lease_info, canary):
         self.ss = ss
         self.incominghome = incominghome
         self.finalhome = finalhome
-        self._size = size
+        self._max_size = max_size # don't allow the client to write more than this
         self._canary = canary
         self._disconnect_marker = canary.notifyOnDisconnect(self._disconnected)
         self.closed = False
         self.throw_out_all_data = False
-        # touch the file, so later callers will see that we're working on it.
-        assert not os.path.exists(incominghome)
-        fileutil.make_dirs(os.path.dirname(incominghome))
-        # Also construct the metadata.
-        f = open(incominghome, 'wb')
-        f.write(struct.pack(">LLL", 1, size, 0))
-        f.close()
-        self._sharefile = ShareFile(incominghome)
+        self._sharefile = ShareFile(incominghome, create=True, max_size=max_size)
         # also, add our lease to the file now, so that other ones can be
         # added by simultaneous uploaders
         self._sharefile.add_lease(lease_info)
 
     def allocated_size(self):
-        return self._size
+        return self._max_size
 
     def remote_write(self, offset, data):
         start = time.time()
@@ -323,9 +351,14 @@ class BucketWriter(Referenceable):
 class BucketReader(Referenceable):
     implements(RIBucketReader)
 
-    def __init__(self, ss, sharefname):
+    def __init__(self, ss, sharefname, storage_index=None, shnum=None):
         self.ss = ss
         self._share_file = ShareFile(sharefname)
+        self.storage_index = storage_index
+        self.shnum = shnum
+
+    def __repr__(self):
+        return "<%s %s %s>" % (self.__class__.__name__, base32.b2a_l(self.storage_index[:8], 60), self.shnum)
 
     def remote_read(self, offset, length):
         start = time.time()
@@ -334,6 +367,11 @@ class BucketReader(Referenceable):
         self.ss.count("read")
         return data
 
+    def remote_advise_corrupt_share(self, reason):
+        return self.ss.remote_advise_corrupt_share("immutable",
+                                                   self.storage_index,
+                                                   self.shnum,
+                                                   reason)
 
 # the MutableShareFile is like the ShareFile, but used for mutable data. It
 # has a different layout. See docs/mutable.txt for more details.
@@ -355,8 +393,8 @@ class BucketReader(Referenceable):
 # 9   ??        n*92    extra leases
 
 
-assert struct.calcsize("L"), 4
-assert struct.calcsize("Q"), 8
+assert struct.calcsize("L"), 4 # The struct module doc says that L's are 4 bytes in size.
+assert struct.calcsize("Q"), 8 # The struct module doc says that Q's are 8 bytes in size (at least with big-endian ordering).
 
 class MutableShareFile:
 
@@ -577,6 +615,7 @@ class MutableShareFile:
         return leases
 
     def add_lease(self, lease_info):
+        precondition(lease_info.owner_num != 0) # 0 means "no lease here"
         f = open(self.home, 'rb+')
         num_lease_slots = self._get_num_lease_slots(f)
         empty_slot = self._get_first_empty_lease_slot(f)
@@ -611,6 +650,7 @@ class MutableShareFile:
         raise IndexError(msg)
 
     def add_or_renew_lease(self, lease_info):
+        precondition(lease_info.owner_num != 0) # 0 means "no lease here"
         try:
             self.renew_lease(lease_info.renew_secret,
                              lease_info.expiration_time)
@@ -694,7 +734,7 @@ class MutableShareFile:
             self.log(format="bad write enabler on SI %(si)s,"
                      " recorded by nodeid %(nodeid)s",
                      facility="tahoe.storage",
-                     level=log.WEIRD,
+                     level=log.WEIRD, umid="cE1eBQ",
                      si=si_s, nodeid=idlib.nodeid_b2a(write_enabler_nodeid))
             msg = "The write enabler was recorded by nodeid '%s'." % \
                   (idlib.nodeid_b2a(write_enabler_nodeid),)
@@ -759,9 +799,7 @@ class StorageServer(service.MultiService, Referenceable):
     implements(RIStorageServer, IStatsProducer)
     name = 'storage'
 
-    OLDEST_SUPPORTED_VERSION = LooseVersion("0.8.0")
-
-    def __init__(self, storedir, sizelimit=None,
+    def __init__(self, storedir, reserved_space=0,
                  discard_storage=False, readonly_storage=False,
                  stats_provider=None):
         service.MultiService.__init__(self)
@@ -769,7 +807,10 @@ class StorageServer(service.MultiService, Referenceable):
         sharedir = os.path.join(storedir, "shares")
         fileutil.make_dirs(sharedir)
         self.sharedir = sharedir
-        self.sizelimit = sizelimit
+        # we don't actually create the corruption-advisory dir until necessary
+        self.corruption_advisory_dir = os.path.join(storedir,
+                                                    "corruption-advisories")
+        self.reserved_space = int(reserved_space)
         self.no_storage = discard_storage
         self.readonly_storage = readonly_storage
         self.stats_provider = stats_provider
@@ -779,23 +820,23 @@ class StorageServer(service.MultiService, Referenceable):
         self._clean_incomplete()
         fileutil.make_dirs(self.incomingdir)
         self._active_writers = weakref.WeakKeyDictionary()
-        lp = log.msg("StorageServer created, now measuring space..",
-                     facility="tahoe.storage")
-        self.consumed = None
-        if self.sizelimit:
-            self.consumed = fileutil.du(self.sharedir)
-            log.msg(format="space measurement done, consumed=%(consumed)d bytes",
-                    consumed=self.consumed,
-                    parent=lp, facility="tahoe.storage")
+        lp = log.msg("StorageServer created", facility="tahoe.storage")
+
+        if reserved_space:
+            if self.get_available_space() is None:
+                log.msg("warning: [storage]reserved_space= is set, but this platform does not support statvfs(2), so this reservation cannot be honored",
+                        umin="0wZ27w", level=log.UNUSUAL)
+
         self.latencies = {"allocate": [], # immutable
                           "write": [],
                           "close": [],
                           "read": [],
-                          "renew": [],
-                          "cancel": [],
                           "get": [],
                           "writev": [], # mutable
                           "readv": [],
+                          "add-lease": [], # both
+                          "renew": [],
+                          "cancel": [],
                           }
 
     def count(self, name, delta=1):
@@ -857,22 +898,79 @@ class StorageServer(service.MultiService, Referenceable):
         fileutil.rm_dir(self.incomingdir)
 
     def get_stats(self):
+        # remember: RIStatsProvider requires that our return dict
+        # contains numeric values.
         stats = { 'storage_server.allocated': self.allocated_size(), }
-        if self.consumed is not None:
-            stats['storage_server.consumed'] = self.consumed
         for category,ld in self.get_latencies().items():
             for name,v in ld.items():
                 stats['storage_server.latencies.%s.%s' % (category, name)] = v
+        writeable = True
+        if self.readonly_storage:
+            writeable = False
+        try:
+            s = os.statvfs(self.storedir)
+            disk_total = s.f_bsize * s.f_blocks
+            disk_used = s.f_bsize * (s.f_blocks - s.f_bfree)
+            # spacetime predictors should look at the slope of disk_used.
+            disk_avail = s.f_bsize * s.f_bavail # available to non-root users
+            # include our local policy here: if we stop accepting shares when
+            # the available space drops below 1GB, then include that fact in
+            # disk_avail.
+            disk_avail -= self.reserved_space
+            disk_avail = max(disk_avail, 0)
+            if self.readonly_storage:
+                disk_avail = 0
+            if disk_avail == 0:
+                writeable = False
+
+            # spacetime predictors should use disk_avail / (d(disk_used)/dt)
+            stats["storage_server.disk_total"] = disk_total
+            stats["storage_server.disk_used"] = disk_used
+            stats["storage_server.disk_avail"] = disk_avail
+        except AttributeError:
+            # os.statvfs is available only on unix
+            pass
+        stats["storage_server.accepting_immutable_shares"] = int(writeable)
         return stats
 
+
+    def stat_disk(self, d):
+        s = os.statvfs(d)
+        # s.f_bavail: available to non-root users
+        disk_avail = s.f_bsize * s.f_bavail
+        return disk_avail
+
+    def get_available_space(self):
+        # returns None if it cannot be measured (windows)
+        try:
+            disk_avail = self.stat_disk(self.storedir)
+            disk_avail -= self.reserved_space
+        except AttributeError:
+            disk_avail = None
+        if self.readonly_storage:
+            disk_avail = 0
+        return disk_avail
+
     def allocated_size(self):
-        space = self.consumed or 0
+        space = 0
         for bw in self._active_writers:
             space += bw.allocated_size()
         return space
 
-    def remote_get_versions(self):
-        return (str(allmydata.__version__), str(self.OLDEST_SUPPORTED_VERSION))
+    def remote_get_version(self):
+        remaining_space = self.get_available_space()
+        if remaining_space is None:
+            # we're on a platform that doesn't have 'df', so make a vague
+            # guess.
+            remaining_space = 2**64
+        version = { "http://allmydata.org/tahoe/protocols/storage/v1" :
+                    { "maximum-immutable-share-size": remaining_space,
+                      "tolerates-immutable-read-overrun": True,
+                      "delete-mutable-shares-with-zero-length-writev": True,
+                      },
+                    "application-version": str(allmydata.__full_version__),
+                    }
+        return version
 
     def remote_allocate_buckets(self, storage_index,
                                 renew_secret, cancel_secret,
@@ -893,17 +991,21 @@ class StorageServer(service.MultiService, Referenceable):
         # in this implementation, the lease information (including secrets)
         # goes into the share files themselves. It could also be put into a
         # separate database. Note that the lease should not be added until
-        # the BucketWrite has been closed.
+        # the BucketWriter has been closed.
         expire_time = time.time() + 31*24*60*60
         lease_info = LeaseInfo(owner_num,
                                renew_secret, cancel_secret,
                                expire_time, self.my_nodeid)
 
-        space_per_bucket = allocated_size
-        no_limits = self.sizelimit is None
-        yes_limits = not no_limits
-        if yes_limits:
-            remaining_space = self.sizelimit - self.allocated_size()
+        max_space_per_bucket = allocated_size
+
+        remaining_space = self.get_available_space()
+        limited = remaining_space is not None
+        if limited:
+            # this is a bit conservative, since some of this allocated_size()
+            # has already been written to disk, where it will show up in
+            # get_available_space.
+            remaining_space -= self.allocated_size()
 
         # fill alreadygot with all shares that we have, not just the ones
         # they asked about: this will save them a lot of work. Add or update
@@ -914,10 +1016,7 @@ class StorageServer(service.MultiService, Referenceable):
             sf = ShareFile(fn)
             sf.add_or_renew_lease(lease_info)
 
-        if self.readonly_storage:
-            # we won't accept new shares
-            self.add_latency("allocate", time.time() - start)
-            return alreadygot, bucketwriters
+        # self.readonly_storage causes remaining_space=0
 
         for shnum in sharenums:
             incominghome = os.path.join(self.incomingdir, si_dir, "%d" % shnum)
@@ -931,16 +1030,16 @@ class StorageServer(service.MultiService, Referenceable):
                 # occurs while the first is still in progress, the second
                 # uploader will use different storage servers.
                 pass
-            elif no_limits or remaining_space >= space_per_bucket:
+            elif (not limited) or (remaining_space >= max_space_per_bucket):
                 # ok! we need to create the new share file.
                 bw = BucketWriter(self, incominghome, finalhome,
-                                  space_per_bucket, lease_info, canary)
+                                  max_space_per_bucket, lease_info, canary)
                 if self.no_storage:
                     bw.throw_out_all_data = True
                 bucketwriters[shnum] = bw
                 self._active_writers[bw] = 1
-                if yes_limits:
-                    remaining_space -= space_per_bucket
+                if limited:
+                    remaining_space -= max_space_per_bucket
             else:
                 # bummer! not enough space to accept this bucket
                 pass
@@ -968,7 +1067,7 @@ class StorageServer(service.MultiService, Referenceable):
             yield sf
 
     def remote_add_lease(self, storage_index, renew_secret, cancel_secret,
-                         owner_num=0):
+                         owner_num=1):
         start = time.time()
         self.count("add-lease")
         new_expire_time = time.time() + 31*24*60*60
@@ -1020,8 +1119,6 @@ class StorageServer(service.MultiService, Referenceable):
             if not os.listdir(storagedir):
                 os.rmdir(storagedir)
 
-        if self.consumed is not None:
-            self.consumed -= total_space_freed
         if self.stats_provider:
             self.stats_provider.count('storage_server.bytes_freed',
                                       total_space_freed)
@@ -1030,8 +1127,6 @@ class StorageServer(service.MultiService, Referenceable):
             raise IndexError("no such storage index")
 
     def bucket_writer_closed(self, bw, consumed_size):
-        if self.consumed is not None:
-            self.consumed += consumed_size
         if self.stats_provider:
             self.stats_provider.count('storage_server.bytes_added', consumed_size)
         del self._active_writers[bw]
@@ -1057,7 +1152,8 @@ class StorageServer(service.MultiService, Referenceable):
         log.msg("storage: get_buckets %s" % si_s)
         bucketreaders = {} # k: sharenum, v: BucketReader
         for shnum, filename in self._get_bucket_shares(storage_index):
-            bucketreaders[shnum] = BucketReader(self, filename)
+            bucketreaders[shnum] = BucketReader(self, filename,
+                                                storage_index, shnum)
         self.add_latency("get", time.time() - start)
         return bucketreaders
 
@@ -1126,27 +1222,37 @@ class StorageServer(service.MultiService, Referenceable):
         for sharenum, share in shares.items():
             read_data[sharenum] = share.readv(read_vector)
 
+        ownerid = 1 # TODO
+        expire_time = time.time() + 31*24*60*60   # one month
+        lease_info = LeaseInfo(ownerid,
+                               renew_secret, cancel_secret,
+                               expire_time, self.my_nodeid)
+
         if testv_is_good:
             # now apply the write vectors
             for sharenum in test_and_write_vectors:
                 (testv, datav, new_length) = test_and_write_vectors[sharenum]
-                if sharenum not in shares:
-                    # allocate a new share
-                    allocated_size = 2000 # arbitrary, really
-                    share = self._allocate_slot_share(bucketdir, secrets,
-                                                      sharenum,
-                                                      allocated_size,
-                                                      owner_num=0)
-                    shares[sharenum] = share
-                shares[sharenum].writev(datav, new_length)
-            # and update the leases on all shares
-            ownerid = 1 # TODO
-            expire_time = time.time() + 31*24*60*60   # one month
-            lease_info = LeaseInfo(ownerid,
-                                   renew_secret, cancel_secret,
-                                   expire_time, self.my_nodeid)
-            for share in shares.values():
-                share.add_or_renew_lease(lease_info)
+                if new_length == 0:
+                    if sharenum in shares:
+                        shares[sharenum].unlink()
+                else:
+                    if sharenum not in shares:
+                        # allocate a new share
+                        allocated_size = 2000 # arbitrary, really
+                        share = self._allocate_slot_share(bucketdir, secrets,
+                                                          sharenum,
+                                                          allocated_size,
+                                                          owner_num=0)
+                        shares[sharenum] = share
+                    shares[sharenum].writev(datav, new_length)
+                    # and update the lease
+                    shares[sharenum].add_or_renew_lease(lease_info)
+
+            if new_length == 0:
+                # delete empty bucket directories
+                if not os.listdir(bucketdir):
+                    os.rmdir(bucketdir)
+
 
         # all done
         self.add_latency("writev", time.time() - start)
@@ -1189,297 +1295,25 @@ class StorageServer(service.MultiService, Referenceable):
         self.add_latency("readv", time.time() - start)
         return datavs
 
-
-# the code before here runs on the storage server, not the client
-# the code beyond here runs on the client, not the storage server
-
-"""
-Share data is written into a single file. At the start of the file, there is
-a series of four-byte big-endian offset values, which indicate where each
-section starts. Each offset is measured from the beginning of the file.
-
-0x00: version number (=00 00 00 01)
-0x04: segment size
-0x08: data size
-0x0c: offset of data (=00 00 00 24)
-0x10: offset of plaintext_hash_tree
-0x14: offset of crypttext_hash_tree
-0x18: offset of block_hashes
-0x1c: offset of share_hashes
-0x20: offset of uri_extension_length + uri_extension
-0x24: start of data
-?   : start of plaintext_hash_tree
-?   : start of crypttext_hash_tree
-?   : start of block_hashes
-?   : start of share_hashes
-       each share_hash is written as a two-byte (big-endian) hashnum
-       followed by the 32-byte SHA-256 hash. We only store the hashes
-       necessary to validate the share hash root
-?   : start of uri_extension_length (four-byte big-endian value)
-?   : start of uri_extension
-"""
-
-def allocated_size(data_size, num_segments, num_share_hashes,
-                   uri_extension_size):
-    wbp = WriteBucketProxy(None, data_size, 0, num_segments, num_share_hashes,
-                           uri_extension_size, None)
-    uri_extension_starts_at = wbp._offsets['uri_extension']
-    return uri_extension_starts_at + 4 + uri_extension_size
-
-class WriteBucketProxy:
-    implements(IStorageBucketWriter)
-    def __init__(self, rref, data_size, segment_size, num_segments,
-                 num_share_hashes, uri_extension_size, nodeid):
-        self._rref = rref
-        self._data_size = data_size
-        self._segment_size = segment_size
-        self._num_segments = num_segments
-        self._nodeid = nodeid
-
-        if segment_size >= 2**32 or data_size >= 2**32:
-            raise FileTooLargeError("This file is too large to be uploaded (data_size).")
-
-        effective_segments = mathutil.next_power_of_k(num_segments,2)
-        self._segment_hash_size = (2*effective_segments - 1) * HASH_SIZE
-        # how many share hashes are included in each share? This will be
-        # about ln2(num_shares).
-        self._share_hash_size = num_share_hashes * (2+HASH_SIZE)
-        # we commit to not sending a uri extension larger than this
-        self._uri_extension_size = uri_extension_size
-
-        offsets = self._offsets = {}
-        x = 0x24
-        offsets['data'] = x
-        x += data_size
-        offsets['plaintext_hash_tree'] = x
-        x += self._segment_hash_size
-        offsets['crypttext_hash_tree'] = x
-        x += self._segment_hash_size
-        offsets['block_hashes'] = x
-        x += self._segment_hash_size
-        offsets['share_hashes'] = x
-        x += self._share_hash_size
-        offsets['uri_extension'] = x
-
-        if x >= 2**32:
-            raise FileTooLargeError("This file is too large to be uploaded (offsets).")
-
-        offset_data = struct.pack(">LLLLLLLLL",
-                                  1, # version number
-                                  segment_size,
-                                  data_size,
-                                  offsets['data'],
-                                  offsets['plaintext_hash_tree'],
-                                  offsets['crypttext_hash_tree'],
-                                  offsets['block_hashes'],
-                                  offsets['share_hashes'],
-                                  offsets['uri_extension'],
-                                  )
-        assert len(offset_data) == 0x24
-        self._offset_data = offset_data
-
-    def __repr__(self):
-        if self._nodeid:
-            nodeid_s = idlib.nodeid_b2a(self._nodeid)
-        else:
-            nodeid_s = "[None]"
-        return "<allmydata.storage.WriteBucketProxy for node %s>" % nodeid_s
-
-    def start(self):
-        return self._write(0, self._offset_data)
-
-    def put_block(self, segmentnum, data):
-        offset = self._offsets['data'] + segmentnum * self._segment_size
-        assert offset + len(data) <= self._offsets['uri_extension']
-        assert isinstance(data, str)
-        if segmentnum < self._num_segments-1:
-            precondition(len(data) == self._segment_size,
-                         len(data), self._segment_size)
-        else:
-            precondition(len(data) == (self._data_size -
-                                       (self._segment_size *
-                                        (self._num_segments - 1))),
-                         len(data), self._segment_size)
-        return self._write(offset, data)
-
-    def put_plaintext_hashes(self, hashes):
-        offset = self._offsets['plaintext_hash_tree']
-        assert isinstance(hashes, list)
-        data = "".join(hashes)
-        precondition(len(data) == self._segment_hash_size,
-                     len(data), self._segment_hash_size)
-        precondition(offset+len(data) <= self._offsets['crypttext_hash_tree'],
-                     offset, len(data), offset+len(data),
-                     self._offsets['crypttext_hash_tree'])
-        return self._write(offset, data)
-
-    def put_crypttext_hashes(self, hashes):
-        offset = self._offsets['crypttext_hash_tree']
-        assert isinstance(hashes, list)
-        data = "".join(hashes)
-        precondition(len(data) == self._segment_hash_size,
-                     len(data), self._segment_hash_size)
-        precondition(offset + len(data) <= self._offsets['block_hashes'],
-                     offset, len(data), offset+len(data),
-                     self._offsets['block_hashes'])
-        return self._write(offset, data)
-
-    def put_block_hashes(self, blockhashes):
-        offset = self._offsets['block_hashes']
-        assert isinstance(blockhashes, list)
-        data = "".join(blockhashes)
-        precondition(len(data) == self._segment_hash_size,
-                     len(data), self._segment_hash_size)
-        precondition(offset + len(data) <= self._offsets['share_hashes'],
-                     offset, len(data), offset+len(data),
-                     self._offsets['share_hashes'])
-        return self._write(offset, data)
-
-    def put_share_hashes(self, sharehashes):
-        # sharehashes is a list of (index, hash) tuples, so they get stored
-        # as 2+32=34 bytes each
-        offset = self._offsets['share_hashes']
-        assert isinstance(sharehashes, list)
-        data = "".join([struct.pack(">H", hashnum) + hashvalue
-                        for hashnum,hashvalue in sharehashes])
-        precondition(len(data) == self._share_hash_size,
-                     len(data), self._share_hash_size)
-        precondition(offset + len(data) <= self._offsets['uri_extension'],
-                     offset, len(data), offset+len(data),
-                     self._offsets['uri_extension'])
-        return self._write(offset, data)
-
-    def put_uri_extension(self, data):
-        offset = self._offsets['uri_extension']
-        assert isinstance(data, str)
-        precondition(len(data) <= self._uri_extension_size,
-                     len(data), self._uri_extension_size)
-        length = struct.pack(">L", len(data))
-        return self._write(offset, length+data)
-
-    def _write(self, offset, data):
-        # TODO: for small shares, buffer the writes and do just a single call
-        return self._rref.callRemote("write", offset, data)
-
-    def close(self):
-        return self._rref.callRemote("close")
-
-    def abort(self):
-        return self._rref.callRemoteOnly("abort")
-
-class ReadBucketProxy:
-    implements(IStorageBucketReader)
-    def __init__(self, rref, peerid=None, storage_index_s=None):
-        self._rref = rref
-        self._peerid = peerid
-        self._si_s = storage_index_s
-        self._started = False
-
-    def get_peerid(self):
-        return self._peerid
-
-    def __repr__(self):
-        peerid_s = idlib.shortnodeid_b2a(self._peerid)
-        return "<ReadBucketProxy to peer [%s] SI %s>" % (peerid_s,
-                                                         self._si_s)
-
-    def startIfNecessary(self):
-        if self._started:
-            return defer.succeed(self)
-        d = self.start()
-        d.addCallback(lambda res: self)
-        return d
-
-    def start(self):
-        # TODO: for small shares, read the whole bucket in start()
-        d = self._read(0, 0x24)
-        d.addCallback(self._parse_offsets)
-        def _started(res):
-            self._started = True
-            return res
-        d.addCallback(_started)
-        return d
-
-    def _parse_offsets(self, data):
-        precondition(len(data) == 0x24)
-        self._offsets = {}
-        (version, self._segment_size, self._data_size) = \
-                  struct.unpack(">LLL", data[0:0xc])
-        _assert(version == 1)
-        x = 0x0c
-        for field in ( 'data',
-                       'plaintext_hash_tree',
-                       'crypttext_hash_tree',
-                       'block_hashes',
-                       'share_hashes',
-                       'uri_extension',
-                       ):
-            offset = struct.unpack(">L", data[x:x+4])[0]
-            x += 4
-            self._offsets[field] = offset
-        return self._offsets
-
-    def get_block(self, blocknum):
-        num_segments = mathutil.div_ceil(self._data_size, self._segment_size)
-        if blocknum < num_segments-1:
-            size = self._segment_size
-        else:
-            size = self._data_size % self._segment_size
-            if size == 0:
-                size = self._segment_size
-        offset = self._offsets['data'] + blocknum * self._segment_size
-        return self._read(offset, size)
-
-    def _str2l(self, s):
-        """ split string (pulled from storage) into a list of blockids """
-        return [ s[i:i+HASH_SIZE]
-                 for i in range(0, len(s), HASH_SIZE) ]
-
-    def get_plaintext_hashes(self):
-        offset = self._offsets['plaintext_hash_tree']
-        size = self._offsets['crypttext_hash_tree'] - offset
-        d = self._read(offset, size)
-        d.addCallback(self._str2l)
-        return d
-
-    def get_crypttext_hashes(self):
-        offset = self._offsets['crypttext_hash_tree']
-        size = self._offsets['block_hashes'] - offset
-        d = self._read(offset, size)
-        d.addCallback(self._str2l)
-        return d
-
-    def get_block_hashes(self):
-        offset = self._offsets['block_hashes']
-        size = self._offsets['share_hashes'] - offset
-        d = self._read(offset, size)
-        d.addCallback(self._str2l)
-        return d
-
-    def get_share_hashes(self):
-        offset = self._offsets['share_hashes']
-        size = self._offsets['uri_extension'] - offset
-        assert size % (2+HASH_SIZE) == 0
-        d = self._read(offset, size)
-        def _unpack_share_hashes(data):
-            assert len(data) == size
-            hashes = []
-            for i in range(0, size, 2+HASH_SIZE):
-                hashnum = struct.unpack(">H", data[i:i+2])[0]
-                hashvalue = data[i+2:i+2+HASH_SIZE]
-                hashes.append( (hashnum, hashvalue) )
-            return hashes
-        d.addCallback(_unpack_share_hashes)
-        return d
-
-    def get_uri_extension(self):
-        offset = self._offsets['uri_extension']
-        d = self._read(offset, 4)
-        def _got_length(data):
-            length = struct.unpack(">L", data)[0]
-            return self._read(offset+4, length)
-        d.addCallback(_got_length)
-        return d
-
-    def _read(self, offset, length):
-        return self._rref.callRemote("read", offset, length)
+    def remote_advise_corrupt_share(self, share_type, storage_index, shnum,
+                                    reason):
+        fileutil.make_dirs(self.corruption_advisory_dir)
+        now = time_format.iso_utc(sep="T")
+        si_s = base32.b2a(storage_index)
+        # windows can't handle colons in the filename
+        fn = os.path.join(self.corruption_advisory_dir,
+                          "%s--%s-%d" % (now, si_s, shnum)).replace(":","")
+        f = open(fn, "w")
+        f.write("report: Share Corruption\n")
+        f.write("type: %s\n" % share_type)
+        f.write("storage_index: %s\n" % si_s)
+        f.write("share_number: %d\n" % shnum)
+        f.write("\n")
+        f.write(reason)
+        f.write("\n")
+        f.close()
+        log.msg(format=("client claims corruption in (%(share_type)s) " +
+                        "%(si)s-%(shnum)d: %(reason)s"),
+                share_type=share_type, si=si_s, shnum=shnum, reason=reason,
+                level=log.SCARY, umid="SGx2fA")
+        return None
