@@ -8,7 +8,7 @@ from twisted.web import client as tw_client
 from allmydata import client, introducer
 from allmydata.immutable import upload
 from allmydata.scripts import create_node
-from allmydata.util import testutil, fileutil
+from allmydata.util import fileutil, pollmixin
 import foolscap
 from foolscap import eventual
 from twisted.python import log
@@ -57,7 +57,7 @@ def discardPage(url, stall=False, *args, **kwargs):
     reactor.connectTCP(host, port, factory)
     return factory.deferred
 
-class SystemFramework(testutil.PollMixin):
+class SystemFramework(pollmixin.PollMixin):
     numnodes = 5
 
     def __init__(self, basedir, mode):
@@ -198,8 +198,8 @@ class SystemFramework(testutil.PollMixin):
             if self.mode in ("receive",):
                 # for this mode, the client-under-test gets all the shares,
                 # so our internal nodes can refuse requests
-                f = open(os.path.join(nodedir, "sizelimit"), "w")
-                f.write("0\n")
+                f = open(os.path.join(nodedir, "readonly_storage"), "w")
+                f.write("\n")
                 f.close()
             c = self.add_service(client.Client(basedir=nodedir))
             self.nodes.append(c)
