@@ -4,12 +4,19 @@ from zope.interface import implements
 from twisted.trial import unittest
 from twisted.internet import defer
 from allmydata import uri, dirnode
-from allmydata.immutable import upload, checker
+from allmydata.immutable import upload
 from allmydata.interfaces import IURI, IClient, IMutableFileNode, \
-     INewDirectoryURI, IReadonlyNewDirectoryURI, IFileNode, ExistingChildError
-from allmydata.util import hashutil, testutil
+     INewDirectoryURI, IReadonlyNewDirectoryURI, IFileNode, \
+     ExistingChildError, NoSuchChildError, \
+     IDeepCheckResults, IDeepCheckAndRepairResults
+from allmydata.mutable.filenode import MutableFileNode
+from allmydata.mutable.common import UncoordinatedWriteError
+from allmydata.util import hashutil, base32
+from allmydata.monitor import Monitor
 from allmydata.test.common import make_chk_file_uri, make_mutable_file_uri, \
-     FakeDirectoryNode, create_chk_filenode
+     FakeDirectoryNode, create_chk_filenode, ErrorMixin, SystemTestMixin
+from allmydata.check_results import CheckResults, CheckAndRepairResults
+import common_util as testutil
 
 # to test dirnode.py, we want to construct a tree of real DirectoryNodes that
 # contain pointers to fake files. We start with a fake MutableFileNode that
@@ -22,22 +29,33 @@ class Marker:
             nodeuri = nodeuri.to_string()
         self.nodeuri = nodeuri
         si = hashutil.tagged_hash("tag1", nodeuri)[:16]
+        self.storage_index = si
         fp = hashutil.tagged_hash("tag2", nodeuri)
-        self.verifieruri = uri.SSKVerifierURI(storage_index=si,
-                                              fingerprint=fp).to_string()
+        self.verifieruri = uri.SSKVerifierURI(storage_index=si, fingerprint=fp)
     def get_uri(self):
         return self.nodeuri
     def get_readonly_uri(self):
         return self.nodeuri
-    def get_verifier(self):
+    def get_verify_cap(self):
         return self.verifieruri
+    def get_storage_index(self):
+        return self.storage_index
 
-    def check(self, verify=False, repair=False):
-        r = checker.Results(None)
-        r.healthy = True
-        r.problems = []
+    def check(self, monitor, verify=False):
+        r = CheckResults(uri.from_string(self.nodeuri), None)
+        r.set_healthy(True)
+        r.set_recoverable(True)
         return defer.succeed(r)
 
+    def check_and_repair(self, monitor, verify=False):
+        d = self.check(verify)
+        def _got(cr):
+            r = CheckAndRepairResults(None)
+            r.pre_repair_results = r.post_repair_results = cr
+            return r
+        d.addCallback(_got)
+        return d
+
 # dirnode requires three methods from the client: upload(),
 # create_node_from_uri(), and create_empty_dirnode(). Of these, upload() is
 # only used by the convenience composite method add_file().
@@ -71,7 +89,8 @@ class FakeClient:
         return d
 
 
-class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
+class Dirnode(unittest.TestCase,
+              testutil.ShouldFailMixin, testutil.StallMixin, ErrorMixin):
     def setUp(self):
         self.client = FakeClient()
 
@@ -88,29 +107,29 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
         d = self.client.create_empty_dirnode()
         def _created(dn):
             u = make_mutable_file_uri()
-            d = dn.set_uri(u"child", u, {})
+            d = dn.set_uri(u"child", u.to_string(), {})
             d.addCallback(lambda res: dn.list())
             def _check1(children):
                 self.failUnless(u"child" in children)
             d.addCallback(_check1)
             d.addCallback(lambda res:
-                          self.shouldFail(KeyError, "get bogus", None,
+                          self.shouldFail(NoSuchChildError, "get bogus", None,
                                           dn.get, u"bogus"))
             def _corrupt(res):
                 filenode = dn._node
                 si = IURI(filenode.get_uri()).storage_index
                 old_contents = filenode.all_contents[si]
-                # we happen to know that the writecap is encrypted near the
-                # end of the string. Flip one of its bits and make sure we
-                # detect the corruption.
+                # We happen to know that the writecap MAC is near the end of the string. Flip
+                # one of its bits and make sure we ignore the corruption.
                 new_contents = testutil.flip_bit(old_contents, -10)
                 # TODO: also test flipping bits in the other portions
                 filenode.all_contents[si] = new_contents
             d.addCallback(_corrupt)
             def _check2(res):
-                self.shouldFail(hashutil.IntegrityCheckError, "corrupt",
-                                "HMAC does not match, crypttext is corrupted",
-                                dn.list)
+                d = dn.list()
+                def _c3(res):
+                    self.failUnless(res.has_key('child'))
+                d.addCallback(_c3)
             d.addCallback(_check2)
             return d
         d.addCallback(_created)
@@ -118,7 +137,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
     def test_check(self):
         d = self.client.create_empty_dirnode()
-        d.addCallback(lambda dn: dn.check())
+        d.addCallback(lambda dn: dn.check(Monitor()))
         def _done(res):
             self.failUnless(res.is_healthy())
         d.addCallback(_done)
@@ -130,6 +149,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
         #  root/subdir/
         #  root/subdir/file1
         #  root/subdir/link -> root
+        #  root/rodir
         d = self.client.create_empty_dirnode()
         def _created_root(rootnode):
             self._rootnode = rootnode
@@ -139,6 +159,10 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             self._subdir = subdir
             d = subdir.add_file(u"file1", upload.Data("data", None))
             d.addCallback(lambda res: subdir.set_node(u"link", self._rootnode))
+            d.addCallback(lambda res: self.client.create_empty_dirnode())
+            d.addCallback(lambda dn:
+                          self._rootnode.set_uri(u"rodir",
+                                                 dn.get_readonly_uri()))
             return d
         d.addCallback(_created_subdir)
         def _done(res):
@@ -148,14 +172,68 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
     def test_deepcheck(self):
         d = self._test_deepcheck_create()
-        d.addCallback(lambda rootnode: rootnode.deep_check())
+        d.addCallback(lambda rootnode: rootnode.start_deep_check().when_done())
         def _check_results(r):
-            self.failUnlessEqual(r.count_objects_checked(), 3)
-            self.failUnlessEqual(r.count_objects_healthy(), 3)
-            self.failUnlessEqual(r.count_repairs_attempted(), 0)
-            self.failUnlessEqual(r.count_repairs_successful(), 0)
-            self.failUnlessEqual(len(r.get_server_problems()), 0)
-            self.failUnlessEqual(len(r.get_problems()), 0)
+            self.failUnless(IDeepCheckResults.providedBy(r))
+            c = r.get_counters()
+            self.failUnlessEqual(c,
+                                 {"count-objects-checked": 4,
+                                  "count-objects-healthy": 4,
+                                  "count-objects-unhealthy": 0,
+                                  "count-objects-unrecoverable": 0,
+                                  "count-corrupt-shares": 0,
+                                  })
+            self.failIf(r.get_corrupt_shares())
+            self.failUnlessEqual(len(r.get_all_results()), 4)
+        d.addCallback(_check_results)
+        return d
+
+    def test_deepcheck_and_repair(self):
+        d = self._test_deepcheck_create()
+        d.addCallback(lambda rootnode:
+                      rootnode.start_deep_check_and_repair().when_done())
+        def _check_results(r):
+            self.failUnless(IDeepCheckAndRepairResults.providedBy(r))
+            c = r.get_counters()
+            self.failUnlessEqual(c,
+                                 {"count-objects-checked": 4,
+                                  "count-objects-healthy-pre-repair": 4,
+                                  "count-objects-unhealthy-pre-repair": 0,
+                                  "count-objects-unrecoverable-pre-repair": 0,
+                                  "count-corrupt-shares-pre-repair": 0,
+                                  "count-objects-healthy-post-repair": 4,
+                                  "count-objects-unhealthy-post-repair": 0,
+                                  "count-objects-unrecoverable-post-repair": 0,
+                                  "count-corrupt-shares-post-repair": 0,
+                                  "count-repairs-attempted": 0,
+                                  "count-repairs-successful": 0,
+                                  "count-repairs-unsuccessful": 0,
+                                  })
+            self.failIf(r.get_corrupt_shares())
+            self.failIf(r.get_remaining_corrupt_shares())
+            self.failUnlessEqual(len(r.get_all_results()), 4)
+        d.addCallback(_check_results)
+        return d
+
+    def _mark_file_bad(self, rootnode):
+        si = IURI(rootnode.get_uri())._filenode_uri.storage_index
+        rootnode._node.bad_shares[si] = "unhealthy"
+        return rootnode
+
+    def test_deepcheck_problems(self):
+        d = self._test_deepcheck_create()
+        d.addCallback(lambda rootnode: self._mark_file_bad(rootnode))
+        d.addCallback(lambda rootnode: rootnode.start_deep_check().when_done())
+        def _check_results(r):
+            c = r.get_counters()
+            self.failUnlessEqual(c,
+                                 {"count-objects-checked": 4,
+                                  "count-objects-healthy": 3,
+                                  "count-objects-unhealthy": 1,
+                                  "count-objects-unrecoverable": 0,
+                                  "count-corrupt-shares": 0,
+                                  })
+            #self.failUnlessEqual(len(r.get_problems()), 1) # TODO
         d.addCallback(_check_results)
         return d
 
@@ -166,7 +244,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
         d = self.client.create_empty_dirnode()
         def _created(rw_dn):
-            d2 = rw_dn.set_uri(u"child", fileuri)
+            d2 = rw_dn.set_uri(u"child", fileuri.to_string())
             d2.addCallback(lambda res: rw_dn)
             return d2
         d.addCallback(_created)
@@ -178,7 +256,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             self.failUnless(ro_dn.is_mutable())
 
             self.shouldFail(dirnode.NotMutableError, "set_uri ro", None,
-                            ro_dn.set_uri, u"newchild", fileuri)
+                            ro_dn.set_uri, u"newchild", fileuri.to_string())
             self.shouldFail(dirnode.NotMutableError, "set_uri ro", None,
                             ro_dn.set_node, u"newchild", filenode)
             self.shouldFail(dirnode.NotMutableError, "set_nodes ro", None,
@@ -210,6 +288,8 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
     def test_create(self):
         self.expected_manifest = []
+        self.expected_verifycaps = set()
+        self.expected_storage_indexes = set()
 
         d = self.client.create_empty_dirnode()
         def _then(n):
@@ -220,9 +300,16 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             self.failUnless(u.startswith("URI:DIR2:"), u)
             u_ro = n.get_readonly_uri()
             self.failUnless(u_ro.startswith("URI:DIR2-RO:"), u_ro)
-            u_v = n.get_verifier()
+            u_v = n.get_verify_cap().to_string()
             self.failUnless(u_v.startswith("URI:DIR2-Verifier:"), u_v)
-            self.expected_manifest.append(u_v)
+            u_r = n.get_repair_cap().to_string()
+            self.failUnlessEqual(u_r, u)
+            self.expected_manifest.append( ((), u) )
+            self.expected_verifycaps.add(u_v)
+            si = n.get_storage_index()
+            self.expected_storage_indexes.add(base32.b2a(si))
+            expected_si = n._uri._filenode_uri.storage_index
+            self.failUnlessEqual(si, expected_si)
 
             d = n.list()
             d.addCallback(lambda res: self.failUnlessEqual(res, {}))
@@ -231,14 +318,15 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             fake_file_uri = make_mutable_file_uri()
             other_file_uri = make_mutable_file_uri()
             m = Marker(fake_file_uri)
-            ffu_v = m.get_verifier()
-            assert isinstance(ffu_v, str)
-            self.expected_manifest.append(ffu_v)
-            d.addCallback(lambda res: n.set_uri(u"child", fake_file_uri))
+            ffu_v = m.get_verify_cap().to_string()
+            self.expected_manifest.append( ((u"child",) , m.get_uri()) )
+            self.expected_verifycaps.add(ffu_v)
+            self.expected_storage_indexes.add(base32.b2a(m.get_storage_index()))
+            d.addCallback(lambda res: n.set_uri(u"child", fake_file_uri.to_string()))
             d.addCallback(lambda res:
                           self.shouldFail(ExistingChildError, "set_uri-no",
                                           "child 'child' already exists",
-                                          n.set_uri, u"child", other_file_uri,
+                                          n.set_uri, u"child", other_file_uri.to_string(),
                                           overwrite=False))
             # /
             # /child = mutable
@@ -251,9 +339,12 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             def _created(subdir):
                 self.failUnless(isinstance(subdir, FakeDirectoryNode))
                 self.subdir = subdir
-                new_v = subdir.get_verifier()
+                new_v = subdir.get_verify_cap().to_string()
                 assert isinstance(new_v, str)
-                self.expected_manifest.append(new_v)
+                self.expected_manifest.append( ((u"subdir",), subdir.get_uri()) )
+                self.expected_verifycaps.add(new_v)
+                si = subdir.get_storage_index()
+                self.expected_storage_indexes.add(base32.b2a(si))
             d.addCallback(_created)
 
             d.addCallback(lambda res:
@@ -267,13 +358,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
                           self.failUnlessEqual(sorted(children.keys()),
                                                sorted([u"child", u"subdir"])))
 
-            d.addCallback(lambda res: n.build_manifest())
-            def _check_manifest(manifest):
-                self.failUnlessEqual(sorted(manifest),
-                                     sorted(self.expected_manifest))
-            d.addCallback(_check_manifest)
-
-            d.addCallback(lambda res: n.deep_stats())
+            d.addCallback(lambda res: n.start_deep_stats().when_done())
             def _check_deepstats(stats):
                 self.failUnless(isinstance(stats, dict))
                 expected = {"count-immutable-files": 0,
@@ -299,9 +384,26 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
                 self.failUnlessEqual(stats["size-files-histogram"], [])
             d.addCallback(_check_deepstats)
 
+            d.addCallback(lambda res: n.build_manifest().when_done())
+            def _check_manifest(res):
+                manifest = res["manifest"]
+                self.failUnlessEqual(sorted(manifest),
+                                     sorted(self.expected_manifest))
+                stats = res["stats"]
+                _check_deepstats(stats)
+                self.failUnlessEqual(self.expected_verifycaps,
+                                     res["verifycaps"])
+                self.failUnlessEqual(self.expected_storage_indexes,
+                                     res["storage-index"])
+            d.addCallback(_check_manifest)
+
             def _add_subsubdir(res):
                 return self.subdir.create_empty_directory(u"subsubdir")
             d.addCallback(_add_subsubdir)
+            # /
+            # /child = mutable
+            # /subdir = directory
+            # /subdir/subsubdir = directory
             d.addCallback(lambda res: n.get_child_at_path(u"subdir/subsubdir"))
             d.addCallback(lambda subsubdir:
                           self.failUnless(isinstance(subsubdir,
@@ -315,14 +417,47 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
                           self.failUnlessEqual(sorted(metadata.keys()),
                                                ["ctime", "mtime"]))
 
+            d.addCallback(lambda res:
+                          self.shouldFail(NoSuchChildError, "gcamap-no",
+                                          "nope",
+                                          n.get_child_and_metadata_at_path,
+                                          u"subdir/nope"))
+            d.addCallback(lambda res:
+                          n.get_child_and_metadata_at_path(u""))
+            def _check_child_and_metadata1(res):
+                child, metadata = res
+                self.failUnless(isinstance(child, FakeDirectoryNode))
+                # edge-metadata needs at least one path segment
+                self.failUnlessEqual(sorted(metadata.keys()), [])
+            d.addCallback(_check_child_and_metadata1)
+            d.addCallback(lambda res:
+                          n.get_child_and_metadata_at_path(u"child"))
+
+            def _check_child_and_metadata2(res):
+                child, metadata = res
+                self.failUnlessEqual(child.get_uri(),
+                                     fake_file_uri.to_string())
+                self.failUnlessEqual(sorted(metadata.keys()),
+                                     ["ctime", "mtime"])
+            d.addCallback(_check_child_and_metadata2)
+
+            d.addCallback(lambda res:
+                          n.get_child_and_metadata_at_path(u"subdir/subsubdir"))
+            def _check_child_and_metadata3(res):
+                child, metadata = res
+                self.failUnless(isinstance(child, FakeDirectoryNode))
+                self.failUnlessEqual(sorted(metadata.keys()),
+                                     ["ctime", "mtime"])
+            d.addCallback(_check_child_and_metadata3)
+
             # set_uri + metadata
             # it should be possible to add a child without any metadata
-            d.addCallback(lambda res: n.set_uri(u"c2", fake_file_uri, {}))
+            d.addCallback(lambda res: n.set_uri(u"c2", fake_file_uri.to_string(), {}))
             d.addCallback(lambda res: n.get_metadata_for(u"c2"))
             d.addCallback(lambda metadata: self.failUnlessEqual(metadata, {}))
 
             # if we don't set any defaults, the child should get timestamps
-            d.addCallback(lambda res: n.set_uri(u"c3", fake_file_uri))
+            d.addCallback(lambda res: n.set_uri(u"c3", fake_file_uri.to_string()))
             d.addCallback(lambda res: n.get_metadata_for(u"c3"))
             d.addCallback(lambda metadata:
                           self.failUnlessEqual(sorted(metadata.keys()),
@@ -330,7 +465,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
             # or we can add specific metadata at set_uri() time, which
             # overrides the timestamps
-            d.addCallback(lambda res: n.set_uri(u"c4", fake_file_uri,
+            d.addCallback(lambda res: n.set_uri(u"c4", fake_file_uri.to_string(),
                                                 {"key": "value"}))
             d.addCallback(lambda res: n.get_metadata_for(u"c4"))
             d.addCallback(lambda metadata:
@@ -372,9 +507,9 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
             d.addCallback(lambda res: n.delete(u"d4"))
 
             # metadata through set_children()
-            d.addCallback(lambda res: n.set_children([ (u"e1", fake_file_uri),
-                                                   (u"e2", fake_file_uri, {}),
-                                                   (u"e3", fake_file_uri,
+            d.addCallback(lambda res: n.set_children([ (u"e1", fake_file_uri.to_string()),
+                                                   (u"e2", fake_file_uri.to_string(), {}),
+                                                   (u"e3", fake_file_uri.to_string(),
                                                     {"key": "value"}),
                                                    ]))
             d.addCallback(lambda res:
@@ -572,7 +707,7 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
             # now make sure that we honor overwrite=False
             d.addCallback(lambda res:
-                          self.subdir2.set_uri(u"newchild", other_file_uri))
+                          self.subdir2.set_uri(u"newchild", other_file_uri.to_string()))
 
             d.addCallback(lambda res:
                           self.shouldFail(ExistingChildError, "move_child_to-no",
@@ -589,11 +724,12 @@ class Dirnode(unittest.TestCase, testutil.ShouldFailMixin, testutil.StallMixin):
 
         d.addCallback(_then)
 
+        d.addErrback(self.explain_error)
         return d
 
 class DeepStats(unittest.TestCase):
     def test_stats(self):
-        ds = dirnode.DeepStats()
+        ds = dirnode.DeepStats(None)
         ds.add("count-files")
         ds.add("size-immutable-files", 123)
         ds.histogram("size-files-histogram", 123)
@@ -618,7 +754,7 @@ class DeepStats(unittest.TestCase):
         self.failUnlessEqual(s["size-files-histogram"],
                              [ (101, 316, 1), (317, 1000, 1) ])
 
-        ds = dirnode.DeepStats()
+        ds = dirnode.DeepStats(None)
         for i in range(1, 1100):
             ds.histogram("size-files-histogram", i)
         ds.histogram("size-files-histogram", 4*1000*1000*1000*1000) # 4TB
@@ -634,39 +770,50 @@ class DeepStats(unittest.TestCase):
                                (3162277660169L, 10000000000000L, 1),
                                ])
 
+class UCWEingMutableFileNode(MutableFileNode):
+    please_ucwe_after_next_upload = False
 
-netstring = hashutil.netstring
-split_netstring = dirnode.split_netstring
+    def _upload(self, new_contents, servermap):
+        d = MutableFileNode._upload(self, new_contents, servermap)
+        def _ucwe(res):
+            if self.please_ucwe_after_next_upload:
+                self.please_ucwe_after_next_upload = False
+                raise UncoordinatedWriteError()
+            return res
+        d.addCallback(_ucwe)
+        return d
+class UCWEingNewDirectoryNode(dirnode.NewDirectoryNode):
+    filenode_class = UCWEingMutableFileNode
 
-class Netstring(unittest.TestCase):
-    def test_split(self):
-        a = netstring("hello") + netstring("world")
-        self.failUnlessEqual(split_netstring(a, 2), ("hello", "world"))
-        self.failUnlessEqual(split_netstring(a, 2, False), ("hello", "world"))
-        self.failUnlessEqual(split_netstring(a, 2, True),
-                             ("hello", "world", ""))
-        self.failUnlessRaises(ValueError, split_netstring, a, 3)
-        self.failUnlessRaises(ValueError, split_netstring, a+" extra", 2)
-        self.failUnlessRaises(ValueError, split_netstring, a+" extra", 2, False)
 
-    def test_extra(self):
-        a = netstring("hello")
-        self.failUnlessEqual(split_netstring(a, 1, True), ("hello", ""))
-        b = netstring("hello") + "extra stuff"
-        self.failUnlessEqual(split_netstring(b, 1, True),
-                             ("hello", "extra stuff"))
+class Deleter(SystemTestMixin, unittest.TestCase):
+    def test_retry(self):
+        # ticket #550, a dirnode.delete which experiences an
+        # UncoordinatedWriteError will fail with an incorrect "you're
+        # deleting something which isn't there" NoSuchChildError exception.
 
-    def test_nested(self):
-        a = netstring("hello") + netstring("world") + "extra stuff"
-        b = netstring("a") + netstring("is") + netstring(a) + netstring(".")
-        top = split_netstring(b, 4)
-        self.failUnlessEqual(len(top), 4)
-        self.failUnlessEqual(top[0], "a")
-        self.failUnlessEqual(top[1], "is")
-        self.failUnlessEqual(top[2], a)
-        self.failUnlessEqual(top[3], ".")
-        self.failUnlessRaises(ValueError, split_netstring, a, 2)
-        self.failUnlessRaises(ValueError, split_netstring, a, 2, False)
-        bottom = split_netstring(a, 2, True)
-        self.failUnlessEqual(bottom, ("hello", "world", "extra stuff"))
+        # to trigger this, we start by creating a directory with a single
+        # file in it. Then we create a special dirnode that uses a modified
+        # MutableFileNode which will raise UncoordinatedWriteError once on
+        # demand. We then call dirnode.delete, which ought to retry and
+        # succeed.
+
+        self.basedir = self.mktemp()
+        d = self.set_up_nodes()
+        d.addCallback(lambda ignored: self.clients[0].create_empty_dirnode())
+        small = upload.Data("Small enough for a LIT", None)
+        def _created_dir(dn):
+            self.root = dn
+            self.root_uri = dn.get_uri()
+            return dn.add_file(u"file", small)
+        d.addCallback(_created_dir)
+        def _do_delete(ignored):
+            n = UCWEingNewDirectoryNode(self.clients[0]).init_from_uri(self.root_uri)
+            assert n._node.please_ucwe_after_next_upload == False
+            n._node.please_ucwe_after_next_upload = True
+            # This should succeed, not raise an exception
+            return n.delete(u"file")
+        d.addCallback(_do_delete)
+
+        return d
 
