@@ -1,4 +1,3 @@
-
 import os
 from twisted.trial import unittest
 from twisted.application import service
@@ -6,8 +5,9 @@ from twisted.application import service
 from foolscap import Tub, eventual
 from foolscap.logging import log
 
-from allmydata import offloaded, storage
-from allmydata.immutable import upload
+from allmydata import storage
+from allmydata.immutable import offloaded, upload
+from allmydata import uri
 from allmydata.util import hashutil, fileutil, mathutil
 from pycryptopp.cipher.aes import AES
 
@@ -28,8 +28,10 @@ class CHKUploadHelper_fake(offloaded.CHKUploadHelper):
                             "size": size,
                             }
                 self._results.uri_extension_data = ueb_data
-                return (hashutil.uri_extension_hash(""),
-                        needed_shares, total_shares, size)
+                self._results.verifycapstr = uri.CHKFileVerifierURI(self._storage_index, "x"*32,
+                                                                 needed_shares, total_shares,
+                                                                 size).to_string()
+                return self._results
             d2.addCallback(_got_parms)
             return d2
         d.addCallback(_got_size)
@@ -74,6 +76,15 @@ def flush_but_dont_ignore(res):
     d.addCallback(_done)
     return d
 
+def wait_a_few_turns(ignored=None):
+    d = eventual.fireEventually()
+    d.addCallback(eventual.fireEventually)
+    d.addCallback(eventual.fireEventually)
+    d.addCallback(eventual.fireEventually)
+    d.addCallback(eventual.fireEventually)
+    d.addCallback(eventual.fireEventually)
+    return d
+
 def upload_data(uploader, data, convergence):
     u = upload.Data(data, convergence=convergence)
     return uploader.upload(u)
@@ -110,10 +121,7 @@ class AssistedUpload(unittest.TestCase):
         u = upload.Uploader(self.helper_furl)
         u.setServiceParent(self.s)
 
-        # wait a few turns
-        d = eventual.fireEventually()
-        d.addCallback(eventual.fireEventually)
-        d.addCallback(eventual.fireEventually)
+        d = wait_a_few_turns()
 
         def _ready(res):
             assert u._helper
@@ -121,8 +129,8 @@ class AssistedUpload(unittest.TestCase):
             return upload_data(u, DATA, convergence="some convergence string")
         d.addCallback(_ready)
         def _uploaded(results):
-            uri = results.uri
-            assert "CHK" in uri
+            the_uri = results.uri
+            assert "CHK" in the_uri
         d.addCallback(_uploaded)
 
         def _check_empty(res):
@@ -164,18 +172,15 @@ class AssistedUpload(unittest.TestCase):
         u = upload.Uploader(self.helper_furl)
         u.setServiceParent(self.s)
 
-        # wait a few turns
-        d = eventual.fireEventually()
-        d.addCallback(eventual.fireEventually)
-        d.addCallback(eventual.fireEventually)
+        d = wait_a_few_turns()
 
         def _ready(res):
             assert u._helper
             return upload_data(u, DATA, convergence="test convergence string")
         d.addCallback(_ready)
         def _uploaded(results):
-            uri = results.uri
-            assert "CHK" in uri
+            the_uri = results.uri
+            assert "CHK" in the_uri
         d.addCallback(_uploaded)
 
         def _check_empty(res):
@@ -194,10 +199,7 @@ class AssistedUpload(unittest.TestCase):
         u = upload.Uploader(self.helper_furl)
         u.setServiceParent(self.s)
 
-        # wait a few turns
-        d = eventual.fireEventually()
-        d.addCallback(eventual.fireEventually)
-        d.addCallback(eventual.fireEventually)
+        d = wait_a_few_turns()
 
         def _ready(res):
             assert u._helper
@@ -205,8 +207,8 @@ class AssistedUpload(unittest.TestCase):
             return upload_data(u, DATA, convergence="some convergence string")
         d.addCallback(_ready)
         def _uploaded(results):
-            uri = results.uri
-            assert "CHK" in uri
+            the_uri = results.uri
+            assert "CHK" in the_uri
         d.addCallback(_uploaded)
 
         def _check_empty(res):
