@@ -1,14 +1,15 @@
-
 from twisted.trial import unittest
 
 from twisted.internet import defer
 import time, os.path, stat
 import itertools
 from allmydata import interfaces
-from allmydata.util import fileutil, hashutil
+from allmydata.util import fileutil, hashutil, base32
 from allmydata.storage import BucketWriter, BucketReader, \
-     WriteBucketProxy, ReadBucketProxy, StorageServer, MutableShareFile, \
+     StorageServer, MutableShareFile, \
      storage_index_to_dir, DataTooLargeError, LeaseInfo
+from allmydata.immutable.layout import WriteBucketProxy, WriteBucketProxy_v2, \
+     ReadBucketProxy
 from allmydata.interfaces import BadWriteEnablerError
 from allmydata.test.common import LoggingServiceParent
 
@@ -124,25 +125,26 @@ class BucketProxy(unittest.TestCase):
         bw, rb, sharefname = self.make_bucket("test_create", 500)
         bp = WriteBucketProxy(rb,
                               data_size=300,
-                              segment_size=10,
+                              block_size=10,
                               num_segments=5,
                               num_share_hashes=3,
-                              uri_extension_size=500, nodeid=None)
+                              uri_extension_size_max=500, nodeid=None)
         self.failUnless(interfaces.IStorageBucketWriter.providedBy(bp))
 
-    def test_readwrite(self):
+    def _do_test_readwrite(self, name, header_size, wbp_class, rbp_class):
         # Let's pretend each share has 100 bytes of data, and that there are
-        # 4 segments (25 bytes each), and 8 shares total. So the three
-        # per-segment merkle trees (plaintext_hash_tree, crypttext_hash_tree,
+        # 4 segments (25 bytes each), and 8 shares total. So the two
+        # per-segment merkle trees (crypttext_hash_tree,
         # block_hashes) will have 4 leaves and 7 nodes each. The per-share
         # merkle tree (share_hashes) has 8 leaves and 15 nodes, and we need 3
         # nodes. Furthermore, let's assume the uri_extension is 500 bytes
         # long. That should make the whole share:
         #
         # 0x24 + 100 + 7*32 + 7*32 + 7*32 + 3*(2+32) + 4+500 = 1414 bytes long
+        # 0x44 + 100 + 7*32 + 7*32 + 7*32 + 3*(2+32) + 4+500 = 1446 bytes long
+
+        sharesize = header_size + 100 + 7*32 + 7*32 + 7*32 + 3*(2+32) + 4+500
 
-        plaintext_hashes = [hashutil.tagged_hash("plain", "bar%d" % i)
-                            for i in range(7)]
         crypttext_hashes = [hashutil.tagged_hash("crypt", "bar%d" % i)
                             for i in range(7)]
         block_hashes = [hashutil.tagged_hash("block", "bar%d" % i)
@@ -151,21 +153,20 @@ class BucketProxy(unittest.TestCase):
                         for i in (1,9,13)]
         uri_extension = "s" + "E"*498 + "e"
 
-        bw, rb, sharefname = self.make_bucket("test_readwrite", 1414)
-        bp = WriteBucketProxy(rb,
-                              data_size=95,
-                              segment_size=25,
-                              num_segments=4,
-                              num_share_hashes=3,
-                              uri_extension_size=len(uri_extension),
-                              nodeid=None)
+        bw, rb, sharefname = self.make_bucket(name, sharesize)
+        bp = wbp_class(rb,
+                       data_size=95,
+                       block_size=25,
+                       num_segments=4,
+                       num_share_hashes=3,
+                       uri_extension_size_max=len(uri_extension),
+                       nodeid=None)
 
-        d = bp.start()
+        d = bp.put_header()
         d.addCallback(lambda res: bp.put_block(0, "a"*25))
         d.addCallback(lambda res: bp.put_block(1, "b"*25))
         d.addCallback(lambda res: bp.put_block(2, "c"*25))
         d.addCallback(lambda res: bp.put_block(3, "d"*20))
-        d.addCallback(lambda res: bp.put_plaintext_hashes(plaintext_hashes))
         d.addCallback(lambda res: bp.put_crypttext_hashes(crypttext_hashes))
         d.addCallback(lambda res: bp.put_block_hashes(block_hashes))
         d.addCallback(lambda res: bp.put_share_hashes(share_hashes))
@@ -177,28 +178,23 @@ class BucketProxy(unittest.TestCase):
             br = BucketReader(self, sharefname)
             rb = RemoteBucket()
             rb.target = br
-            rbp = ReadBucketProxy(rb, peerid="abc")
+            rbp = rbp_class(rb, peerid="abc", storage_index="")
             self.failUnless("to peer" in repr(rbp))
             self.failUnless(interfaces.IStorageBucketReader.providedBy(rbp))
 
-            d1 = rbp.startIfNecessary()
-            d1.addCallback(lambda res: rbp.startIfNecessary()) # idempotent
-            d1.addCallback(lambda res: rbp.get_block(0))
+            d1 = rbp.get_block_data(0, 25, 25)
             d1.addCallback(lambda res: self.failUnlessEqual(res, "a"*25))
-            d1.addCallback(lambda res: rbp.get_block(1))
+            d1.addCallback(lambda res: rbp.get_block_data(1, 25, 25))
             d1.addCallback(lambda res: self.failUnlessEqual(res, "b"*25))
-            d1.addCallback(lambda res: rbp.get_block(2))
+            d1.addCallback(lambda res: rbp.get_block_data(2, 25, 25))
             d1.addCallback(lambda res: self.failUnlessEqual(res, "c"*25))
-            d1.addCallback(lambda res: rbp.get_block(3))
+            d1.addCallback(lambda res: rbp.get_block_data(3, 25, 20))
             d1.addCallback(lambda res: self.failUnlessEqual(res, "d"*20))
 
-            d1.addCallback(lambda res: rbp.get_plaintext_hashes())
-            d1.addCallback(lambda res:
-                           self.failUnlessEqual(res, plaintext_hashes))
             d1.addCallback(lambda res: rbp.get_crypttext_hashes())
             d1.addCallback(lambda res:
                            self.failUnlessEqual(res, crypttext_hashes))
-            d1.addCallback(lambda res: rbp.get_block_hashes())
+            d1.addCallback(lambda res: rbp.get_block_hashes(set(range(4))))
             d1.addCallback(lambda res: self.failUnlessEqual(res, block_hashes))
             d1.addCallback(lambda res: rbp.get_share_hashes())
             d1.addCallback(lambda res: self.failUnlessEqual(res, share_hashes))
@@ -212,7 +208,17 @@ class BucketProxy(unittest.TestCase):
 
         return d
 
+    def test_readwrite_v1(self):
+        return self._do_test_readwrite("test_readwrite_v1",
+                                       0x24, WriteBucketProxy, ReadBucketProxy)
 
+    def test_readwrite_v2(self):
+        return self._do_test_readwrite("test_readwrite_v2",
+                                       0x44, WriteBucketProxy_v2, ReadBucketProxy)
+
+class FakeDiskStorageServer(StorageServer):
+    def stat_disk(self, d):
+        return self.DISKAVAIL
 
 class Server(unittest.TestCase):
 
@@ -226,10 +232,10 @@ class Server(unittest.TestCase):
         basedir = os.path.join("storage", "Server", name)
         return basedir
 
-    def create(self, name, sizelimit=None):
+    def create(self, name, reserved_space=0, klass=StorageServer):
         workdir = self.workdir(name)
-        ss = StorageServer(workdir, sizelimit,
-                           stats_provider=FakeStatsProvider())
+        ss = klass(workdir, reserved_space=reserved_space,
+                   stats_provider=FakeStatsProvider())
         ss.setNodeID("\x00" * 20)
         ss.setServiceParent(self.sparent)
         return ss
@@ -246,6 +252,23 @@ class Server(unittest.TestCase):
                                           renew_secret, cancel_secret,
                                           sharenums, size, canary)
 
+    def test_large_share(self):
+        ss = self.create("test_large_share")
+
+        already,writers = self.allocate(ss, "allocate", [0], 2**32+2)
+        self.failUnlessEqual(already, set())
+        self.failUnlessEqual(set(writers.keys()), set([0]))
+
+        shnum, bucket = writers.items()[0]
+        # This test is going to hammer your filesystem if it doesn't make a sparse file for this.  :-(
+        bucket.remote_write(2**32, "ab")
+        bucket.remote_close()
+
+        readers = ss.remote_get_buckets("allocate")
+        reader = readers[shnum]
+        self.failUnlessEqual(reader.remote_read(2**32, 2), "ab")
+    test_large_share.skip = "This test can spuriously fail if you have less than 4 GiB free on your filesystem, and if your filesystem doesn't support efficient sparse files then it is very expensive (Mac OS X is the only system I know of in the desktop/server area that doesn't support efficient sparse files)."
+
     def test_dont_overfill_dirs(self):
         """
         This test asserts that if you add a second share whose storage index
@@ -354,8 +377,13 @@ class Server(unittest.TestCase):
         self.failUnlessEqual(already, set())
         self.failUnlessEqual(set(writers.keys()), set([0,1,2]))
 
-    def test_sizelimits(self):
-        ss = self.create("test_sizelimits", 5000)
+    def test_reserved_space(self):
+        ss = self.create("test_reserved_space", reserved_space=10000,
+                         klass=FakeDiskStorageServer)
+        # the FakeDiskStorageServer doesn't do real statvfs() calls
+        ss.DISKAVAIL = 15000
+        # 15k available, 10k reserved, leaves 5k for shares
+
         # a newly created and filled share incurs this much overhead, beyond
         # the size we request.
         OVERHEAD = 3*4
@@ -391,6 +419,11 @@ class Server(unittest.TestCase):
         self.failUnlessEqual(len(ss._active_writers), 0)
 
         allocated = 1001 + OVERHEAD + LEASE_SIZE
+
+        # we have to manually increase DISKAVAIL, since we're not doing real
+        # disk measurements
+        ss.DISKAVAIL -= allocated
+
         # now there should be ALLOCATED=1001+12+72=1085 bytes allocated, and
         # 5000-1085=3915 free, therefore we can fit 39 100byte shares
         already3,writers3 = self.allocate(ss,"vid3", range(100), 100, canary)
@@ -403,19 +436,6 @@ class Server(unittest.TestCase):
         ss.disownServiceParent()
         del ss
 
-        # creating a new StorageServer in the same directory should see the
-        # same usage.
-
-        # metadata that goes into the share file is counted upon share close,
-        # as well as at startup. metadata that goes into other files will not
-        # be counted until the next startup, so if we were creating any
-        # extra-file metadata, the allocation would be more than 'allocated'
-        # and this test would need to be changed.
-        ss = self.create("test_sizelimits", 5000)
-        already4,writers4 = self.allocate(ss, "vid4", range(100), 100, canary)
-        self.failUnlessEqual(len(writers4), 39)
-        self.failUnlessEqual(len(ss._active_writers), 39)
-
     def test_seek(self):
         basedir = self.workdir("test_seek_behavior")
         fileutil.make_dirs(basedir)
@@ -474,6 +494,17 @@ class Server(unittest.TestCase):
         self.failUnlessEqual(len(leases), 2)
         self.failUnlessEqual(set([l.renew_secret for l in leases]), set([rs1, rs2]))
 
+        # and a third lease, using add-lease
+        rs2a,cs2a = (hashutil.tagged_hash("blah", "%d" % self._lease_secret.next()),
+                     hashutil.tagged_hash("blah", "%d" % self._lease_secret.next()))
+        ss.remote_add_lease("si1", rs2a, cs2a)
+        leases = list(ss.get_leases("si1"))
+        self.failUnlessEqual(len(leases), 3)
+        self.failUnlessEqual(set([l.renew_secret for l in leases]), set([rs1, rs2, rs2a]))
+
+        # add-lease on a missing storage index is an error
+        self.failUnlessRaises(IndexError, ss.remote_add_lease, "si18", "", "")
+
         # check that si0 is readable
         readers = ss.remote_get_buckets("si0")
         self.failUnlessEqual(len(readers), 5)
@@ -499,7 +530,7 @@ class Server(unittest.TestCase):
         self.failUnlessRaises(IndexError, ss.remote_renew_lease, "si0", rs0)
 
 
-        # cancel the first lease on si1, leaving the second in place
+        # cancel the first lease on si1, leaving the second and third in place
         ss.remote_cancel_lease("si1", cs1)
         readers = ss.remote_get_buckets("si1")
         self.failUnlessEqual(len(readers), 5)
@@ -507,16 +538,18 @@ class Server(unittest.TestCase):
         self.failUnlessRaises(IndexError, ss.remote_renew_lease, "si1", rs1)
 
         leases = list(ss.get_leases("si1"))
-        self.failUnlessEqual(len(leases), 1)
-        self.failUnlessEqual(set([l.renew_secret for l in leases]), set([rs2]))
+        self.failUnlessEqual(len(leases), 2)
+        self.failUnlessEqual(set([l.renew_secret for l in leases]), set([rs2, rs2a]))
 
         ss.remote_renew_lease("si1", rs2)
-        # cancelling the second should make it go away
+        # cancelling the second and third should make it go away
         ss.remote_cancel_lease("si1", cs2)
+        ss.remote_cancel_lease("si1", cs2a)
         readers = ss.remote_get_buckets("si1")
         self.failUnlessEqual(len(readers), 0)
         self.failUnlessRaises(IndexError, ss.remote_renew_lease, "si1", rs1)
         self.failUnlessRaises(IndexError, ss.remote_renew_lease, "si1", rs2)
+        self.failUnlessRaises(IndexError, ss.remote_renew_lease, "si1", rs2a)
 
         leases = list(ss.get_leases("si1"))
         self.failUnlessEqual(len(leases), 0)
@@ -555,11 +588,19 @@ class Server(unittest.TestCase):
         ss.setNodeID("\x00" * 20)
         ss.setServiceParent(self.sparent)
 
-        canary = FakeCanary()
         already,writers = self.allocate(ss, "vid", [0,1,2], 75)
         self.failUnlessEqual(already, set())
         self.failUnlessEqual(writers, {})
 
+        stats = ss.get_stats()
+        self.failUnlessEqual(stats["storage_server.accepting_immutable_shares"],
+                             False)
+        if "storage_server.disk_avail" in stats:
+            # windows does not have os.statvfs, so it doesn't give us disk
+            # stats. But if there are stats, readonly_storage means
+            # disk_avail=0
+            self.failUnlessEqual(stats["storage_server.disk_avail"], 0)
+
     def test_discard(self):
         # discard is really only used for other tests, but we test it anyways
         workdir = self.workdir("test_discard")
@@ -581,6 +622,51 @@ class Server(unittest.TestCase):
         self.failUnlessEqual(set(b.keys()), set([0,1,2]))
         self.failUnlessEqual(b[0].remote_read(0, 25), "\x00" * 25)
 
+    def test_advise_corruption(self):
+        workdir = self.workdir("test_advise_corruption")
+        ss = StorageServer(workdir, discard_storage=True)
+        ss.setNodeID("\x00" * 20)
+        ss.setServiceParent(self.sparent)
+
+        si0_s = base32.b2a("si0")
+        ss.remote_advise_corrupt_share("immutable", "si0", 0,
+                                       "This share smells funny.\n")
+        reportdir = os.path.join(workdir, "corruption-advisories")
+        reports = os.listdir(reportdir)
+        self.failUnlessEqual(len(reports), 1)
+        report_si0 = reports[0]
+        self.failUnless(si0_s in report_si0, report_si0)
+        f = open(os.path.join(reportdir, report_si0), "r")
+        report = f.read()
+        f.close()
+        self.failUnless("type: immutable" in report)
+        self.failUnless(("storage_index: %s" % si0_s) in report)
+        self.failUnless("share_number: 0" in report)
+        self.failUnless("This share smells funny." in report)
+
+        # test the RIBucketWriter version too
+        si1_s = base32.b2a("si1")
+        already,writers = self.allocate(ss, "si1", [1], 75)
+        self.failUnlessEqual(already, set())
+        self.failUnlessEqual(set(writers.keys()), set([1]))
+        writers[1].remote_write(0, "data")
+        writers[1].remote_close()
+
+        b = ss.remote_get_buckets("si1")
+        self.failUnlessEqual(set(b.keys()), set([1]))
+        b[1].remote_advise_corrupt_share("This share tastes like dust.\n")
+
+        reports = os.listdir(reportdir)
+        self.failUnlessEqual(len(reports), 2)
+        report_si1 = [r for r in reports if si1_s in r][0]
+        f = open(os.path.join(reportdir, report_si1), "r")
+        report = f.read()
+        f.close()
+        self.failUnless("type: immutable" in report)
+        self.failUnless(("storage_index: %s" % si1_s) in report)
+        self.failUnless("share_number: 1" in report)
+        self.failUnless("This share tastes like dust." in report)
+
 
 
 class MutableServer(unittest.TestCase):
@@ -595,9 +681,9 @@ class MutableServer(unittest.TestCase):
         basedir = os.path.join("storage", "MutableServer", name)
         return basedir
 
-    def create(self, name, sizelimit=None):
+    def create(self, name):
         workdir = self.workdir(name)
-        ss = StorageServer(workdir, sizelimit)
+        ss = StorageServer(workdir)
         ss.setServiceParent(self.sparent)
         ss.setNodeID("\x00" * 20)
         return ss
@@ -954,7 +1040,7 @@ class MutableServer(unittest.TestCase):
             self.failUnlessEqual(a.expiration_time, b.expiration_time)
 
     def test_leases(self):
-        ss = self.create("test_leases", sizelimit=1000*1000)
+        ss = self.create("test_leases")
         def secrets(n):
             return ( self.write_enabler("we1"),
                      self.renew_secret("we1-%d" % n),
@@ -973,28 +1059,39 @@ class MutableServer(unittest.TestCase):
         f.write("you ought to be ignoring me\n")
         f.close()
 
+        s0 = MutableShareFile(os.path.join(bucket_dir, "0"))
+        self.failUnlessEqual(len(s0.debug_get_leases()), 1)
+
+        # add-lease on a missing storage index is an error
+        self.failUnlessRaises(IndexError, ss.remote_add_lease, "si18", "", "")
+
         # re-allocate the slots and use the same secrets, that should update
         # the lease
         write("si1", secrets(0), {0: ([], [(0,data)], None)}, [])
+        self.failUnlessEqual(len(s0.debug_get_leases()), 1)
 
         # renew it directly
         ss.remote_renew_lease("si1", secrets(0)[1])
+        self.failUnlessEqual(len(s0.debug_get_leases()), 1)
 
         # now allocate them with a bunch of different secrets, to trigger the
-        # extended lease code
+        # extended lease code. Use add_lease for one of them.
         write("si1", secrets(1), {0: ([], [(0,data)], None)}, [])
-        write("si1", secrets(2), {0: ([], [(0,data)], None)}, [])
+        self.failUnlessEqual(len(s0.debug_get_leases()), 2)
+        secrets2 = secrets(2)
+        ss.remote_add_lease("si1", secrets2[1], secrets2[2])
+        self.failUnlessEqual(len(s0.debug_get_leases()), 3)
         write("si1", secrets(3), {0: ([], [(0,data)], None)}, [])
         write("si1", secrets(4), {0: ([], [(0,data)], None)}, [])
         write("si1", secrets(5), {0: ([], [(0,data)], None)}, [])
 
+        self.failUnlessEqual(len(s0.debug_get_leases()), 6)
+
         # cancel one of them
         ss.remote_cancel_lease("si1", secrets(5)[2])
+        self.failUnlessEqual(len(s0.debug_get_leases()), 5)
 
-        s0 = MutableShareFile(os.path.join(bucket_dir, "0"))
         all_leases = s0.debug_get_leases()
-        self.failUnlessEqual(len(all_leases), 5)
-
         # and write enough data to expand the container, forcing the server
         # to move the leases
         write("si1", secrets(0),
@@ -1077,6 +1174,50 @@ class MutableServer(unittest.TestCase):
         self.failUnlessRaises(IndexError,
                               ss.remote_cancel_lease, "si2", "nonsecret")
 
+    def test_remove(self):
+        ss = self.create("test_remove")
+        self.allocate(ss, "si1", "we1", self._lease_secret.next(),
+                      set([0,1,2]), 100)
+        readv = ss.remote_slot_readv
+        writev = ss.remote_slot_testv_and_readv_and_writev
+        secrets = ( self.write_enabler("we1"),
+                    self.renew_secret("we1"),
+                    self.cancel_secret("we1") )
+        # delete sh0 by setting its size to zero
+        answer = writev("si1", secrets,
+                        {0: ([], [], 0)},
+                        [])
+        # the answer should mention all the shares that existed before the
+        # write
+        self.failUnlessEqual(answer, (True, {0:[],1:[],2:[]}) )
+        # but a new read should show only sh1 and sh2
+        self.failUnlessEqual(readv("si1", [], [(0,10)]),
+                             {1: [""], 2: [""]})
+
+        # delete sh1 by setting its size to zero
+        answer = writev("si1", secrets,
+                        {1: ([], [], 0)},
+                        [])
+        self.failUnlessEqual(answer, (True, {1:[],2:[]}) )
+        self.failUnlessEqual(readv("si1", [], [(0,10)]),
+                             {2: [""]})
+
+        # delete sh2 by setting its size to zero
+        answer = writev("si1", secrets,
+                        {2: ([], [], 0)},
+                        [])
+        self.failUnlessEqual(answer, (True, {2:[]}) )
+        self.failUnlessEqual(readv("si1", [], [(0,10)]),
+                             {})
+        # and the bucket directory should now be gone
+        si = base32.b2a("si1")
+        # note: this is a detail of the storage server implementation, and
+        # may change in the future
+        prefix = si[:2]
+        prefixdir = os.path.join(self.workdir("test_remove"), "shares", prefix)
+        bucketdir = os.path.join(prefixdir, si)
+        self.failUnless(os.path.exists(prefixdir))
+        self.failIf(os.path.exists(bucketdir))
 
 class Stats(unittest.TestCase):
 
@@ -1090,9 +1231,9 @@ class Stats(unittest.TestCase):
         basedir = os.path.join("storage", "Server", name)
         return basedir
 
-    def create(self, name, sizelimit=None):
+    def create(self, name):
         workdir = self.workdir(name)
-        ss = StorageServer(workdir, sizelimit)
+        ss = StorageServer(workdir)
         ss.setNodeID("\x00" * 20)
         ss.setServiceParent(self.sparent)
         return ss
