@@ -1,18 +1,20 @@
 
 import os
+from cStringIO import StringIO
 from twisted.trial import unittest
 from twisted.python.failure import Failure
 from twisted.python import log
 from twisted.internet import defer
-from cStringIO import StringIO
+from foolscap import eventual
 
-from allmydata import uri
-from allmydata.immutable import upload, encode
-from allmydata.interfaces import IFileURI, FileTooLargeError
+import allmydata # for __full_version__
+from allmydata import uri, monitor
+from allmydata.immutable import upload
+from allmydata.interfaces import IFileURI, FileTooLargeError, NotEnoughSharesError
 from allmydata.util.assertutil import precondition
 from allmydata.util.deferredutil import DeferredListShouldSucceed
-from allmydata.util.testutil import ShouldFailMixin
-from foolscap import eventual
+from common import SystemTestMixin
+from common_util import ShouldFailMixin
 
 MiB = 1024*1024
 
@@ -80,6 +82,17 @@ class FakeStorageServer:
         self.mode = mode
         self.allocated = []
         self.queries = 0
+        self.version = { "http://allmydata.org/tahoe/protocols/storage/v1" :
+                         { "maximum-immutable-share-size": 2**32 },
+                         "application-version": str(allmydata.__full_version__),
+                         }
+        if mode == "small":
+            self.version = { "http://allmydata.org/tahoe/protocols/storage/v1" :
+                             { "maximum-immutable-share-size": 10 },
+                             "application-version": str(allmydata.__full_version__),
+                             }
+
+
     def callRemote(self, methname, *args, **kwargs):
         def _call():
             meth = getattr(self, methname)
@@ -144,13 +157,23 @@ class FakeClient:
     def __init__(self, mode="good", num_servers=50):
         self.mode = mode
         self.num_servers = num_servers
+        if mode == "some_big_some_small":
+            self.peers = []
+            for fakeid in range(num_servers):
+                if fakeid % 2:
+                    self.peers.append( ("%20d" % fakeid,
+                                        FakeStorageServer("good")) )
+                else:
+                    self.peers.append( ("%20d" % fakeid,
+                                        FakeStorageServer("small")) )
+        else:
+            self.peers = [ ("%20d"%fakeid, FakeStorageServer(self.mode),)
+                           for fakeid in range(self.num_servers) ]
     def log(self, *args, **kwargs):
         pass
     def get_permuted_peers(self, storage_index, include_myself):
-        peers = [ ("%20d"%fakeid, FakeStorageServer(self.mode),)
-                  for fakeid in range(self.num_servers) ]
-        self.last_peers = [p[1] for p in peers]
-        return peers
+        self.last_peers = [p[1] for p in self.peers]
+        return self.peers
     def get_encoding_parameters(self):
         return self.DEFAULT_ENCODING_PARAMETERS
 
@@ -234,17 +257,17 @@ class GoodServer(unittest.TestCase, ShouldFailMixin):
         return DATA[:size]
 
     def test_too_large(self):
-        # we currently impose a sizelimit on uploaded files, because of
-        # limitations in the share format (see ticket #346 for details). The
-        # limit is set to ensure that no share is larger than 4GiB. Make sure
-        # that we reject files larger than that.
+        # we've removed the 4GiB share size limit (see ticket #346 for
+        # details), but still have an 8-byte field, so the limit is now
+        # 2**64, so make sure we reject files larger than that.
         k = 3; happy = 7; n = 10
         self.set_encoding_parameters(k, happy, n)
-        data1 = GiganticUploadable(k*4*1024*1024*1024)
+        big = k*(2**64)
+        data1 = GiganticUploadable(big)
         d = self.shouldFail(FileTooLargeError, "test_too_large-data1",
                             "This file is too large to be uploaded (data_size)",
                             self.u.upload, data1)
-        data2 = GiganticUploadable(k*4*1024*1024*1024-3)
+        data2 = GiganticUploadable(big-3)
         d.addCallback(lambda res:
                       self.shouldFail(FileTooLargeError,
                                       "test_too_large-data2",
@@ -347,7 +370,7 @@ class FullServer(unittest.TestCase):
         self.u.parent = self.node
 
     def _should_fail(self, f):
-        self.failUnless(isinstance(f, Failure) and f.check(encode.NotEnoughSharesError), f)
+        self.failUnless(isinstance(f, Failure) and f.check(NotEnoughSharesError), f)
 
     def test_data_large(self):
         data = DATA
@@ -485,6 +508,34 @@ class PeerSelection(unittest.TestCase):
         d.addCallback(_check)
         return d
 
+    def test_some_big_some_small(self):
+        # 10 shares, 20 servers, but half the servers don't support a
+        # share-size large enough for our file
+        self.node = FakeClient(mode="some_big_some_small", num_servers=20)
+        self.u = upload.Uploader()
+        self.u.running = True
+        self.u.parent = self.node
+
+        data = self.get_data(SIZE_LARGE)
+        self.set_encoding_parameters(3, 5, 10)
+        d = upload_data(self.u, data)
+        d.addCallback(extract_uri)
+        d.addCallback(self._check_large, SIZE_LARGE)
+        def _check(res):
+            # we should have put one share each on the big peers, and zero
+            # shares on the small peers
+            total_allocated = 0
+            for p in self.node.last_peers:
+                if p.mode == "good":
+                    self.failUnlessEqual(len(p.allocated), 1)
+                elif p.mode == "small":
+                    self.failUnlessEqual(len(p.allocated), 0)
+                total_allocated += len(p.allocated)
+            self.failUnlessEqual(total_allocated, 10)
+        d.addCallback(_check)
+        return d
+
+
 class StorageIndex(unittest.TestCase):
     def test_params_must_matter(self):
         DATA = "I am some data"
@@ -542,6 +593,31 @@ class StorageIndex(unittest.TestCase):
         d.addCallback(_done)
         return d
 
+class EncodingParameters(SystemTestMixin, unittest.TestCase):
+    def test_configure_parameters(self):
+        self.basedir = self.mktemp()
+        DATA = "data" * 100
+        u = upload.Data(DATA, convergence="")
+        d = self.set_up_nodes()
+        d.addCallback(lambda res: self.clients[0].upload(u))
+        d.addCallback(lambda ur: self.clients[0].create_node_from_uri(ur.uri))
+        m = monitor.Monitor()
+        d.addCallback(lambda fn: fn.check(m))
+        def _check(cr):
+            data = cr.get_data()
+            self.failUnlessEqual(data["count-shares-needed"], 7)
+            self.failUnlessEqual(data["count-shares-expected"], 12)
+        d.addCallback(_check)
+        return d
+
+    def _set_up_nodes_extra_config(self):
+        f = open(os.path.join(self.getdir("client0"), "tahoe.cfg"), "wt")
+        f.write("\n")
+        f.write("[client]\n")
+        f.write("shares.needed = 7\n")
+        f.write("shares.total = 12\n")
+        f.write("\n")
+        f.close()
 
 # TODO:
 #  upload with exactly 75 peers (shares_of_happiness)
