@@ -4,8 +4,8 @@ from zope.interface import implements
 from twisted.python.components import registerAdapter
 from allmydata import storage
 from allmydata.util import base32, hashutil
-from allmydata.interfaces import IURI, IDirnodeURI, IFileURI, IVerifierURI, \
-     IMutableFileURI, INewDirectoryURI, IReadonlyNewDirectoryURI
+from allmydata.interfaces import IURI, IDirnodeURI, IFileURI, IImmutableFileURI, \
+    IVerifierURI, IMutableFileURI, INewDirectoryURI, IReadonlyNewDirectoryURI
 
 # the URI shall be an ascii representation of the file. It shall contain
 # enough information to retrieve and validate the contents. It shall be
@@ -16,26 +16,34 @@ BASE32STR_256bits = '(%s{51}%s)' % (base32.BASE32CHAR, base32.BASE32CHAR_1bits)
 
 SEP='(?::|%3A)'
 NUMBER='([0-9]+)'
+NUMBER_IGNORE='(?:[0-9]+)'
 
 # URIs (soon to be renamed "caps") are always allowed to come with a leading
-# 'http://127.0.0.1:8123/uri/' that will be ignored.
-OPTIONALHTTPLEAD=r'(?:https?://(?:127.0.0.1|localhost):8123/uri/)?'
+# 'http://127.0.0.1:(8123|3456)/uri/' that will be ignored.
+OPTIONALHTTPLEAD=r'(?:https?://(?:[^:/]+)(?::%s)?/uri/)?' % NUMBER_IGNORE
 
 
 class _BaseURI:
     def __hash__(self):
-        return hash((self.__class__, self.to_string()))
-    def __cmp__(self, them):
-        if cmp(type(self), type(them)):
-            return cmp(type(self), type(them))
-        if cmp(self.__class__, them.__class__):
-            return cmp(self.__class__, them.__class__)
-        return cmp(self.to_string(), them.to_string())
+        return self.to_string().__hash__()
+    def __eq__(self, them):
+        if isinstance(them, _BaseURI):
+            return self.to_string() == them.to_string()
+        else:
+            return False
+    def __ne__(self, them):
+        if isinstance(them, _BaseURI):
+            return self.to_string() != them.to_string()
+        else:
+            return True
     def to_human_encoding(self):
-        return 'http://127.0.0.1:8123/uri/'+self.to_string()
+        return 'http://127.0.0.1:3456/uri/'+self.to_string()
+
+    def get_storage_index(self):
+        return self.storage_index
 
 class CHKFileURI(_BaseURI):
-    implements(IURI, IFileURI)
+    implements(IURI, IImmutableFileURI)
 
     STRING_RE=re.compile('^URI:CHK:'+BASE32STR_128bits+':'+
                          BASE32STR_256bits+':'+NUMBER+':'+NUMBER+':'+NUMBER+
@@ -92,7 +100,7 @@ class CHKFileURI(_BaseURI):
     def get_size(self):
         return self.size
 
-    def get_verifier(self):
+    def get_verify_cap(self):
         return CHKFileVerifierURI(storage_index=self.storage_index,
                                   uri_extension_hash=self.uri_extension_hash,
                                   needed_shares=self.needed_shares,
@@ -145,7 +153,7 @@ class CHKFileVerifierURI(_BaseURI):
 
 
 class LiteralFileURI(_BaseURI):
-    implements(IURI, IFileURI)
+    implements(IURI, IImmutableFileURI)
 
     STRING_RE=re.compile('^URI:LIT:'+base32.BASE32STR_anybytes+'$')
     HUMAN_RE=re.compile('^'+OPTIONALHTTPLEAD+'URI'+SEP+'LIT'+SEP+base32.BASE32STR_anybytes+'$')
@@ -175,8 +183,10 @@ class LiteralFileURI(_BaseURI):
         return False
     def get_readonly(self):
         return self
+    def get_storage_index(self):
+        return None
 
-    def get_verifier(self):
+    def get_verify_cap(self):
         # LIT files need no verification, all the data is present in the URI
         return None
 
@@ -222,6 +232,8 @@ class WriteableSSKFileURI(_BaseURI):
 
     def abbrev(self):
         return base32.b2a(self.writekey[:5])
+    def abbrev_si(self):
+        return base32.b2a(self.storage_index)[:5]
 
     def is_readonly(self):
         return False
@@ -229,7 +241,7 @@ class WriteableSSKFileURI(_BaseURI):
         return True
     def get_readonly(self):
         return ReadonlySSKFileURI(self.readkey, self.fingerprint)
-    def get_verifier(self):
+    def get_verify_cap(self):
         return SSKVerifierURI(self.storage_index, self.fingerprint)
 
 class ReadonlySSKFileURI(_BaseURI):
@@ -268,6 +280,8 @@ class ReadonlySSKFileURI(_BaseURI):
 
     def abbrev(self):
         return base32.b2a(self.readkey[:5])
+    def abbrev_si(self):
+        return base32.b2a(self.storage_index)[:5]
 
     def is_readonly(self):
         return True
@@ -275,7 +289,7 @@ class ReadonlySSKFileURI(_BaseURI):
         return True
     def get_readonly(self):
         return self
-    def get_verifier(self):
+    def get_verify_cap(self):
         return SSKVerifierURI(self.storage_index, self.fingerprint)
 
 class SSKVerifierURI(_BaseURI):
@@ -345,6 +359,8 @@ class _NewDirectoryBaseURI(_BaseURI):
 
     def abbrev(self):
         return self._filenode_uri.to_string().split(':')[2][:5]
+    def abbrev_si(self):
+        return base32.b2a(self._filenode_uri.storage_index)[:5]
 
     def get_filenode_uri(self):
         return self._filenode_uri
@@ -352,8 +368,11 @@ class _NewDirectoryBaseURI(_BaseURI):
     def is_mutable(self):
         return True
 
-    def get_verifier(self):
-        return NewDirectoryURIVerifier(self._filenode_uri.get_verifier())
+    def get_verify_cap(self):
+        return NewDirectoryURIVerifier(self._filenode_uri.get_verify_cap())
+
+    def get_storage_index(self):
+        return self._filenode_uri.get_storage_index()
 
 class NewDirectoryURI(_NewDirectoryBaseURI):
     implements(INewDirectoryURI)
