@@ -3,25 +3,33 @@ import simplejson
 import urllib
 import time
 
+from zope.interface import implements
 from twisted.internet import defer
+from twisted.internet.interfaces import IPushProducer
 from twisted.python.failure import Failure
 from twisted.web import http, html
-from nevow import url, rend, tags as T
+from nevow import url, rend, inevow, tags as T
 from nevow.inevow import IRequest
 
 from foolscap.eventual import fireEventually
 
 from allmydata.util import base32
-from allmydata.uri import from_string_verifier, from_string_dirnode, \
-     CHKFileVerifierURI
+from allmydata.uri import from_string_dirnode
 from allmydata.interfaces import IDirectoryNode, IFileNode, IMutableFileNode, \
-     ExistingChildError
-from allmydata.web.common import text_plain, WebError, IClient, \
-     boolean_of_arg, get_arg, should_create_intermediate_directories, \
+     ExistingChildError, NoSuchChildError
+from allmydata.monitor import Monitor, OperationCancelledError
+from allmydata import dirnode
+from allmydata.web.common import text_plain, WebError, \
+     IClient, IOpHandleTable, NeedOperationHandleError, \
+     boolean_of_arg, get_arg, get_root, \
+     should_create_intermediate_directories, \
      getxmlfile, RenderMixin
 from allmydata.web.filenode import ReplaceMeMixin, \
      FileNodeHandler, PlaceHolderNodeHandler
-from allmydata.web.checker_results import CheckerResults, DeepCheckResults
+from allmydata.web.check_results import CheckResults, \
+     CheckAndRepairResults, DeepCheckResults, DeepCheckAndRepairResults
+from allmydata.web.info import MoreInfo
+from allmydata.web.operations import ReloadMixin
 
 class BlockingFileError(Exception):
     # TODO: catch and transform
@@ -31,10 +39,10 @@ class BlockingFileError(Exception):
 def make_handler_for(node, parentnode=None, name=None):
     if parentnode:
         assert IDirectoryNode.providedBy(parentnode)
-    if IFileNode.providedBy(node):
-        return FileNodeHandler(node, parentnode, name)
     if IMutableFileNode.providedBy(node):
         return FileNodeHandler(node, parentnode, name)
+    if IFileNode.providedBy(node):
+        return FileNodeHandler(node, parentnode, name)
     if IDirectoryNode.providedBy(node):
         return DirectoryNodeHandler(node, parentnode, name)
     raise WebError("Cannot provide handler for '%s'" % node)
@@ -67,7 +75,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
         t = get_arg(req, "t", "").strip()
         if isinstance(node_or_failure, Failure):
             f = node_or_failure
-            f.trap(KeyError)
+            f.trap(NoSuchChildError)
             # No child by this name. What should we do about it?
             if DEBUG: print "no child", name
             if DEBUG: print "postpath", req.postpath
@@ -130,16 +138,12 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
 
         if t == "json":
             return DirectoryJSONMetadata(ctx, self.node)
+        if t == "info":
+            return MoreInfo(self.node)
         if t == "uri":
             return DirectoryURI(ctx, self.node)
         if t == "readonly-uri":
             return DirectoryReadonlyURI(ctx, self.node)
-        if t == "manifest":
-            return Manifest(self.node)
-        if t == "deep-size":
-            return DeepSize(ctx, self.node)
-        if t == "deep-stats":
-            return DeepStats(ctx, self.node)
         if t == 'rename-form':
             return RenameForm(self.node)
 
@@ -158,7 +162,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
                 # they're trying to set_uri and that name is already occupied
                 # (by us).
                 raise ExistingChildError()
-            d = self.parentnode.replace_me_with_a_childcap(ctx, replace)
+            d = self.replace_me_with_a_childcap(ctx, replace)
             # TODO: results
             return d
 
@@ -167,6 +171,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
     def render_POST(self, ctx):
         req = IRequest(ctx)
         t = get_arg(req, "t", "").strip()
+
         if t == "mkdir":
             d = self._POST_mkdir(req)
         elif t == "mkdir-p":
@@ -182,8 +187,16 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
             d = self._POST_rename(req)
         elif t == "check":
             d = self._POST_check(req)
-        elif t == "deep-check":
-            d = self._POST_deep_check(req)
+        elif t == "start-deep-check":
+            d = self._POST_start_deep_check(ctx)
+        elif t == "start-manifest":
+            d = self._POST_start_manifest(ctx)
+        elif t == "start-deep-size":
+            d = self._POST_start_deep_size(ctx)
+        elif t == "start-deep-stats":
+            d = self._POST_start_deep_stats(ctx)
+        elif t == "stream-manifest":
+            d = self._POST_stream_manifest(ctx)
         elif t == "set_children":
             # TODO: docs
             d = self._POST_set_children(req)
@@ -227,7 +240,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
             return defer.succeed(node)
         d = node.get(path[0])
         def _maybe_create(f):
-            f.trap(KeyError)
+            f.trap(NoSuchChildError)
             return node.create_empty_directory(path[0])
         d.addErrback(_maybe_create)
         d.addCallback(self._get_or_create_directories, path[1:])
@@ -260,7 +273,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
         def _maybe_got_node(node_or_failure):
             if isinstance(node_or_failure, Failure):
                 f = node_or_failure
-                f.trap(KeyError)
+                f.trap(NoSuchChildError)
                 # create a placeholder which will see POST t=upload
                 return PlaceHolderNodeHandler(self.node, name)
             else:
@@ -320,10 +333,14 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
             assert isinstance(to_name, unicode)
         if not from_name or not to_name:
             raise WebError("rename requires from_name and to_name")
-        for k,v in [ ('from_name', from_name), ('to_name', to_name) ]:
-            if v and "/" in v:
-                raise WebError("%s= may not contain a slash" % k,
-                               http.BAD_REQUEST)
+        if from_name == to_name:
+            return defer.succeed("redundant rename")
+
+        # allow from_name to contain slashes, so they can fix names that were
+        # accidentally created with them. But disallow them in to_name, to
+        # discourage the practice.
+        if "/" in to_name:
+            raise WebError("to_name= may not contain a slash", http.BAD_REQUEST)
 
         replace = boolean_of_arg(get_arg(req, "replace", "true"))
         d = self.node.move_child_to(from_name, self.node, to_name, replace)
@@ -332,14 +349,71 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
 
     def _POST_check(self, req):
         # check this directory
-        d = self.node.check()
-        d.addCallback(lambda res: CheckerResults(res))
+        verify = boolean_of_arg(get_arg(req, "verify", "false"))
+        repair = boolean_of_arg(get_arg(req, "repair", "false"))
+        if repair:
+            d = self.node.check_and_repair(Monitor(), verify)
+            d.addCallback(lambda res: CheckAndRepairResults(res))
+        else:
+            d = self.node.check(Monitor(), verify)
+            d.addCallback(lambda res: CheckResults(res))
         return d
 
-    def _POST_deep_check(self, req):
+    def _start_operation(self, monitor, renderer, ctx):
+        table = IOpHandleTable(ctx)
+        table.add_monitor(ctx, monitor, renderer)
+        return table.redirect_to(ctx)
+
+    def _POST_start_deep_check(self, ctx):
         # check this directory and everything reachable from it
-        d = self.node.deep_check()
-        d.addCallback(lambda res: DeepCheckResults(res))
+        if not get_arg(ctx, "ophandle"):
+            raise NeedOperationHandleError("slow operation requires ophandle=")
+        verify = boolean_of_arg(get_arg(ctx, "verify", "false"))
+        repair = boolean_of_arg(get_arg(ctx, "repair", "false"))
+        if repair:
+            monitor = self.node.start_deep_check_and_repair(verify)
+            renderer = DeepCheckAndRepairResults(monitor)
+        else:
+            monitor = self.node.start_deep_check(verify)
+            renderer = DeepCheckResults(monitor)
+        return self._start_operation(monitor, renderer, ctx)
+
+    def _POST_start_manifest(self, ctx):
+        if not get_arg(ctx, "ophandle"):
+            raise NeedOperationHandleError("slow operation requires ophandle=")
+        monitor = self.node.build_manifest()
+        renderer = ManifestResults(monitor)
+        return self._start_operation(monitor, renderer, ctx)
+
+    def _POST_start_deep_size(self, ctx):
+        if not get_arg(ctx, "ophandle"):
+            raise NeedOperationHandleError("slow operation requires ophandle=")
+        monitor = self.node.start_deep_stats()
+        renderer = DeepSizeResults(monitor)
+        return self._start_operation(monitor, renderer, ctx)
+
+    def _POST_start_deep_stats(self, ctx):
+        if not get_arg(ctx, "ophandle"):
+            raise NeedOperationHandleError("slow operation requires ophandle=")
+        monitor = self.node.start_deep_stats()
+        renderer = DeepStatsResults(monitor)
+        return self._start_operation(monitor, renderer, ctx)
+
+    def _POST_stream_manifest(self, ctx):
+        walker = ManifestStreamer(ctx, self.node)
+        monitor = self.node.deep_traverse(walker)
+        walker.setMonitor(monitor)
+        # register to hear stopProducing. The walker ignores pauseProducing.
+        IRequest(ctx).registerProducer(walker, True)
+        d = monitor.when_done()
+        def _done(res):
+            IRequest(ctx).unregisterProducer()
+            return res
+        d.addBoth(_done)
+        def _cancelled(f):
+            f.trap(OperationCancelledError)
+            return "Operation Cancelled"
+        d.addErrback(_cancelled)
         return d
 
     def _POST_set_children(self, req):
@@ -354,6 +428,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
             raise
         cs = []
         for name, (file_or_dir, mddict) in children.iteritems():
+            name = unicode(name) # simplejson-2.0.1 returns str *or* unicode
             cap = str(mddict.get('rw_uri') or mddict.get('ro_uri'))
             cs.append((name, cap, mddict.get('metadata')))
         d = self.node.set_children(cs, replace)
@@ -363,9 +438,7 @@ class DirectoryNodeHandler(RenderMixin, rend.Page, ReplaceMeMixin):
 
 def abbreviated_dirnode(dirnode):
     u = from_string_dirnode(dirnode.get_uri())
-    si = u.get_filenode_uri().storage_index
-    si_s = base32.b2a(si)
-    return si_s[:6]
+    return u.abbrev_si()
 
 class DirectoryAsHTML(rend.Page):
     # The remainder of this class is to render the directory into
@@ -389,17 +462,18 @@ class DirectoryAsHTML(rend.Page):
             header.append(" (readonly)")
         return ctx.tag[header]
 
-    def get_root(self, ctx):
-        req = IRequest(ctx)
-        # the addSlash=True gives us one extra (empty) segment
-        depth = len(req.prepath) + len(req.postpath) - 1
-        link = "/".join([".."] * depth)
-        return link
-
     def render_welcome(self, ctx, data):
-        link = self.get_root(ctx)
+        link = get_root(ctx)
         return T.div[T.a(href=link)["Return to Welcome page"]]
 
+    def render_show_readonly(self, ctx, data):
+        if self.node.is_readonly():
+            return ""
+        rocap = self.node.get_readonly_uri()
+        root = get_root(ctx)
+        uri_link = "%s/uri/%s/" % (root, urllib.quote(rocap))
+        return ctx.tag[T.a(href=uri_link)["Read-Only Version"]]
+
     def data_children(self, ctx, data):
         d = self.node.list()
         d.addCallback(lambda dict: sorted(dict.items()))
@@ -428,8 +502,9 @@ class DirectoryAsHTML(rend.Page):
         name, (target, metadata) = data
         name = name.encode("utf-8")
         assert not isinstance(name, unicode)
+        nameurl = urllib.quote(name, safe="") # encode any slashes too
 
-        root = self.get_root(ctx)
+        root = get_root(ctx)
         here = "%s/uri/%s/" % (root, urllib.quote(self.node.get_uri()))
         if self.node.is_readonly():
             delete = "-"
@@ -454,20 +529,6 @@ class DirectoryAsHTML(rend.Page):
 
         ctx.fillSlots("delete", delete)
         ctx.fillSlots("rename", rename)
-        if IDirectoryNode.providedBy(target):
-            check_url = "%s/uri/%s/" % (root, urllib.quote(target.get_uri()))
-            check_done_url = "../../uri/%s/" % urllib.quote(self.node.get_uri())
-        else:
-            check_url = "%s/uri/%s" % (root, urllib.quote(target.get_uri()))
-            check_done_url = "../uri/%s/" % urllib.quote(self.node.get_uri())
-        check = T.form(action=check_url, method="post")[
-            T.input(type='hidden', name='t', value='check'),
-            T.input(type='hidden', name='return_to', value=check_done_url),
-            T.input(type='submit', value='check', name="check"),
-            ]
-        ctx.fillSlots("overwrite",
-                      self.build_overwrite_form(ctx, name, target))
-        ctx.fillSlots("check", check)
 
         times = []
         TIME_FORMAT = "%H:%M:%S %d-%b-%Y"
@@ -493,7 +554,7 @@ class DirectoryAsHTML(rend.Page):
             # to prevent javascript in displayed .html files from stealing a
             # secret directory URI from the URL, send the browser to a URI-based
             # page that doesn't know about the directory at all
-            dlurl = "%s/file/%s/@@named=/%s" % (root, quoted_uri, urllib.quote(name))
+            dlurl = "%s/file/%s/@@named=/%s" % (root, quoted_uri, nameurl)
 
             ctx.fillSlots("filename",
                           T.a(href=dlurl)[html.escape(name)])
@@ -501,11 +562,10 @@ class DirectoryAsHTML(rend.Page):
 
             ctx.fillSlots("size", "?")
 
-            text_plain_url = "%s/file/%s/@@named=/foo.txt" % (root, quoted_uri)
-            text_plain_tag = T.a(href=text_plain_url)["text/plain"]
+            info_link = "%s/uri/%s?t=info" % (root, quoted_uri)
 
         elif IFileNode.providedBy(target):
-            dlurl = "%s/file/%s/@@named=/%s" % (root, quoted_uri, urllib.quote(name))
+            dlurl = "%s/file/%s/@@named=/%s" % (root, quoted_uri, nameurl)
 
             ctx.fillSlots("filename",
                           T.a(href=dlurl)[html.escape(name)])
@@ -513,9 +573,7 @@ class DirectoryAsHTML(rend.Page):
 
             ctx.fillSlots("size", target.get_size())
 
-            text_plain_url = "%s/file/%s/@@named=/foo.txt" % (root, quoted_uri)
-            text_plain_tag = T.a(href=text_plain_url)["text/plain"]
-
+            info_link = "%s/uri/%s?t=info" % (root, quoted_uri)
 
         elif IDirectoryNode.providedBy(target):
             # directory
@@ -528,38 +586,14 @@ class DirectoryAsHTML(rend.Page):
                 dirtype = "DIR"
             ctx.fillSlots("type", dirtype)
             ctx.fillSlots("size", "-")
-            text_plain_tag = None
+            info_link = "%s/uri/%s/?t=info" % (root, quoted_uri)
 
-        childdata = [T.a(href="%s?t=json" % name)["JSON"], ", ",
-                     T.a(href="%s?t=uri" % name)["URI"], ", ",
-                     T.a(href="%s?t=readonly-uri" % name)["readonly-URI"],
-                     ]
-        if text_plain_tag:
-            childdata.extend([", ", text_plain_tag])
-
-        ctx.fillSlots("data", childdata)
-
-        results = "--"
-        # TODO: include a link to see more results, including timestamps
-        # TODO: use a sparkline
-        ctx.fillSlots("checker_results", results)
+        ctx.fillSlots("info", T.a(href=info_link)["More Info"])
 
         return ctx.tag
 
     def render_forms(self, ctx, data):
         forms = []
-        deep_check = T.form(action=".", method="post",
-                            enctype="multipart/form-data")[
-            T.fieldset[
-            T.input(type="hidden", name="t", value="deep-check"),
-            T.input(type="hidden", name="return_to", value="."),
-            T.legend(class_="freeform-form-label")["Run a deep-check operation (EXPENSIVE)"],
-            T.input(type="submit", value="Deep-Check"),
-            " ",
-            "Verify every bit? (EVEN MORE EXPENSIVE):",
-            T.input(type="checkbox", name="verify"),
-            ]]
-        forms.append(T.div(class_="freeform-form")[deep_check])
 
         if self.node.is_readonly():
             forms.append(T.div["No upload forms: directory is read-only"])
@@ -609,26 +643,6 @@ class DirectoryAsHTML(rend.Page):
         forms.append(T.div(class_="freeform-form")[mount])
         return forms
 
-    def build_overwrite_form(self, ctx, name, target):
-        if IMutableFileNode.providedBy(target) and not target.is_readonly():
-            root = self.get_root(ctx)
-            action = "%s/uri/%s" % (root, urllib.quote(target.get_uri()))
-            done_url = "../uri/%s/" % urllib.quote(self.node.get_uri())
-            overwrite = T.form(action=action, method="post",
-                               enctype="multipart/form-data")[
-                T.fieldset[
-                T.input(type="hidden", name="t", value="upload"),
-                T.input(type='hidden', name='when_done', value=done_url),
-                T.legend(class_="freeform-form-label")["Overwrite"],
-                "Choose new file: ",
-                T.input(type="file", name="file", class_="freeform-input-file"),
-                " ",
-                T.input(type="submit", value="Overwrite")
-                ]]
-            return [T.div(class_="freeform-form")[overwrite],]
-        else:
-            return []
-
     def render_results(self, ctx, data):
         req = IRequest(ctx)
         return get_arg(req, "results", "")
@@ -657,6 +671,9 @@ def DirectoryJSONMetadata(ctx, dirnode):
                 kiddata[1]["ro_uri"] = ro_uri
             if rw_uri:
                 kiddata[1]["rw_uri"] = rw_uri
+            verifycap = childnode.get_verify_cap()
+            if verifycap:
+                kiddata[1]['verify_uri'] = verifycap.to_string()
             kiddata[1]['mutable'] = childnode.is_mutable()
             kids[name] = kiddata
         if dirnode.is_readonly():
@@ -670,13 +687,18 @@ def DirectoryJSONMetadata(ctx, dirnode):
             contents['ro_uri'] = dro_uri
         if drw_uri:
             contents['rw_uri'] = drw_uri
+        verifycap = dirnode.get_verify_cap()
+        if verifycap:
+            contents['verify_uri'] = verifycap.to_string()
         contents['mutable'] = dirnode.is_mutable()
         data = ("dirnode", contents)
-        return simplejson.dumps(data, indent=1)
+        return simplejson.dumps(data, indent=1) + "\n"
     d.addCallback(_got)
     d.addCallback(text_plain, ctx)
     return d
 
+
+
 def DirectoryURI(ctx, dirnode):
     return text_plain(dirnode.get_uri(), ctx)
 
@@ -710,37 +732,174 @@ class RenameForm(rend.Page):
         return ctx.tag
 
 
-class Manifest(rend.Page):
+class ManifestResults(rend.Page, ReloadMixin):
     docFactory = getxmlfile("manifest.xhtml")
 
+    def __init__(self, monitor):
+        self.monitor = monitor
+
+    def renderHTTP(self, ctx):
+        output = get_arg(inevow.IRequest(ctx), "output", "html").lower()
+        if output == "text":
+            return self.text(ctx)
+        if output == "json":
+            return self.json(ctx)
+        return rend.Page.renderHTTP(self, ctx)
+
+    def slashify_path(self, path):
+        if not path:
+            return ""
+        return "/".join([p.encode("utf-8") for p in path])
+
+    def text(self, ctx):
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+        lines = []
+        is_finished = self.monitor.is_finished()
+        lines.append("finished: " + {True: "yes", False: "no"}[is_finished])
+        for (path, cap) in self.monitor.get_status()["manifest"]:
+            lines.append(self.slashify_path(path) + " " + cap)
+        return "\n".join(lines) + "\n"
+
+    def json(self, ctx):
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+        m = self.monitor
+        s = m.get_status()
+
+        status = { "stats": s["stats"],
+                   "finished": m.is_finished(),
+                   "origin": base32.b2a(m.origin_si),
+                   }
+        if m.is_finished():
+            # don't return manifest/verifycaps/SIs unless the operation is
+            # done, to save on CPU/memory (both here and in the HTTP client
+            # who has to unpack the JSON). Tests show that the ManifestWalker
+            # needs about 1092 bytes per item, the JSON we generate here
+            # requires about 503 bytes per item, and some internal overhead
+            # (perhaps transport-layer buffers in twisted.web?) requires an
+            # additional 1047 bytes per item.
+            status.update({ "manifest": s["manifest"],
+                            "verifycaps": [i for i in s["verifycaps"]],
+                            "storage-index": [i for i in s["storage-index"]],
+                            })
+            # simplejson doesn't know how to serialize a set. We use a
+            # generator that walks the set rather than list(setofthing) to
+            # save a small amount of memory (4B*len) and a moderate amount of
+            # CPU.
+        return simplejson.dumps(status, indent=1)
+
+    def _si_abbrev(self):
+        return base32.b2a(self.monitor.origin_si)[:6]
+
     def render_title(self, ctx):
-        return T.title["Manifest of SI=%s" % abbreviated_dirnode(self.original)]
+        return T.title["Manifest of SI=%s" % self._si_abbrev()]
 
     def render_header(self, ctx):
-        return T.p["Manifest of SI=%s" % abbreviated_dirnode(self.original)]
+        return T.p["Manifest of SI=%s" % self._si_abbrev()]
 
     def data_items(self, ctx, data):
-        return self.original.build_manifest()
+        return self.monitor.get_status()["manifest"]
 
-    def render_row(self, ctx, refresh_cap):
-        ctx.fillSlots("refresh_capability", refresh_cap)
+    def render_row(self, ctx, (path, cap)):
+        ctx.fillSlots("path", self.slashify_path(path))
+        root = get_root(ctx)
+        # TODO: we need a clean consistent way to get the type of a cap string
+        if cap.startswith("URI:CHK") or cap.startswith("URI:SSK"):
+            nameurl = urllib.quote(path[-1].encode("utf-8"))
+            uri_link = "%s/file/%s/@@named=/%s" % (root, urllib.quote(cap),
+                                                   nameurl)
+        else:
+            uri_link = "%s/uri/%s" % (root, urllib.quote(cap))
+        ctx.fillSlots("cap", T.a(href=uri_link)[cap])
         return ctx.tag
 
-def DeepSize(ctx, dirnode):
-    d = dirnode.build_manifest()
-    def _measure_size(manifest):
-        total = 0
-        for verifiercap in manifest:
-            u = from_string_verifier(verifiercap)
-            if isinstance(u, CHKFileVerifierURI):
-                total += u.size
-        return str(total)
-    d.addCallback(_measure_size)
-    d.addCallback(text_plain, ctx)
-    return d
+class DeepSizeResults(rend.Page):
+    def __init__(self, monitor):
+        self.monitor = monitor
 
-def DeepStats(ctx, dirnode):
-    d = dirnode.deep_stats()
-    d.addCallback(simplejson.dumps, indent=1)
-    d.addCallback(text_plain, ctx)
-    return d
+    def renderHTTP(self, ctx):
+        output = get_arg(inevow.IRequest(ctx), "output", "html").lower()
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+        if output == "json":
+            return self.json(ctx)
+        # plain text
+        is_finished = self.monitor.is_finished()
+        output = "finished: " + {True: "yes", False: "no"}[is_finished] + "\n"
+        if is_finished:
+            stats = self.monitor.get_status()
+            total = (stats.get("size-immutable-files", 0)
+                     + stats.get("size-mutable-files", 0)
+                     + stats.get("size-directories", 0))
+            output += "size: %d\n" % total
+        return output
+
+    def json(self, ctx):
+        status = {"finished": self.monitor.is_finished(),
+                  "size": self.monitor.get_status(),
+                  }
+        return simplejson.dumps(status)
+
+class DeepStatsResults(rend.Page):
+    def __init__(self, monitor):
+        self.monitor = monitor
+
+    def renderHTTP(self, ctx):
+        # JSON only
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+        s = self.monitor.get_status().copy()
+        s["finished"] = self.monitor.is_finished()
+        return simplejson.dumps(s, indent=1)
+
+class ManifestStreamer(dirnode.DeepStats):
+    implements(IPushProducer)
+
+    def __init__(self, ctx, origin):
+        dirnode.DeepStats.__init__(self, origin)
+        self.req = IRequest(ctx)
+
+    def setMonitor(self, monitor):
+        self.monitor = monitor
+    def pauseProducing(self):
+        pass
+    def resumeProducing(self):
+        pass
+    def stopProducing(self):
+        self.monitor.cancel()
+
+    def add_node(self, node, path):
+        dirnode.DeepStats.add_node(self, node, path)
+        d = {"path": path,
+             "cap": node.get_uri()}
+
+        if IDirectoryNode.providedBy(node):
+            d["type"] = "directory"
+        else:
+            d["type"] = "file"
+
+        v = node.get_verify_cap()
+        if v:
+            v = v.to_string()
+        d["verifycap"] = v
+
+        r = node.get_repair_cap()
+        if r:
+            r = r.to_string()
+        d["repaircap"] = r
+
+        si = node.get_storage_index()
+        if si:
+            si = base32.b2a(si)
+        d["storage-index"] = si
+
+        j = simplejson.dumps(d, ensure_ascii=True)
+        assert "\n" not in j
+        self.req.write(j+"\n")
+
+    def finish(self):
+        stats = dirnode.DeepStats.get_results(self)
+        d = {"type": "stats",
+             "stats": stats,
+             }
+        j = simplejson.dumps(d, ensure_ascii=True)
+        assert "\n" not in j
+        self.req.write(j+"\n")
+        return ""
