@@ -1,4 +1,3 @@
-
 import time
 
 from twisted.internet import address
@@ -12,12 +11,11 @@ from formless import webform
 import allmydata # to display import path
 from allmydata import get_package_versions_string
 from allmydata import provisioning
-from allmydata.util import idlib
+from allmydata.util import idlib, log
 from allmydata.interfaces import IFileNode
-from allmydata.web import filenode, directory, unlinked, status
-from allmydata.web.common import abbreviate_size, IClient, getxmlfile, \
-     WebError, get_arg, RenderMixin
-
+from allmydata.web import filenode, directory, unlinked, status, operations
+from allmydata.web.common import abbreviate_size, IClient, \
+     getxmlfile, WebError, get_arg, RenderMixin
 
 
 class URIHandler(RenderMixin, rend.Page):
@@ -104,11 +102,25 @@ class FileHandler(rend.Page):
         raise WebError("/file must be followed by a file-cap and a name",
                        http.NOT_FOUND)
 
+class IncidentReporter(RenderMixin, rend.Page):
+    def render_POST(self, ctx):
+        req = IRequest(ctx)
+        log.msg(format="User reports incident through web page: %(details)s",
+                details=get_arg(req, "details", ""),
+                level=log.WEIRD, umid="LkD9Pw")
+        req.setHeader("content-type", "text/plain")
+        return "Thank you for your report!"
+
+
 class Root(rend.Page):
 
     addSlash = True
     docFactory = getxmlfile("welcome.xhtml")
 
+    def __init__(self, original=None):
+        rend.Page.__init__(self, original)
+        self.child_operations = operations.OphandleTable()
+
     child_uri = URIHandler()
     child_cap = URIHandler()
     child_file = FileHandler()
@@ -122,6 +134,8 @@ class Root(rend.Page):
     child_helper_status = status.HelperStatus()
     child_statistics = status.Statistics()
 
+    child_report_incident = IncidentReporter()
+
     def data_version(self, ctx, data):
         return get_package_versions_string()
     def data_import_path(self, ctx, data):
@@ -138,10 +152,8 @@ class Root(rend.Page):
             ss = client.getServiceNamed("storage")
             allocated_s = abbreviate_size(ss.allocated_size())
             allocated = "about %s allocated" % allocated_s
-            sizelimit = "no size limit"
-            if ss.sizelimit is not None:
-                sizelimit = "size limit is %s" % abbreviate_size(ss.sizelimit)
-            ul[T.li["Storage Server: %s, %s" % (allocated, sizelimit)]]
+            reserved = "%s reserved" % abbreviate_size(ss.reserved_space)
+            ul[T.li["Storage Server: %s, %s" % (allocated, reserved)]]
         except KeyError:
             ul[T.li["Not running storage server"]]
 
@@ -279,3 +291,15 @@ class Root(rend.Page):
             ]]
         return T.div[form]
 
+    def render_incident_button(self, ctx, data):
+        # this button triggers a foolscap-logging "incident"
+        form = T.form(action="report_incident", method="post",
+                      enctype="multipart/form-data")[
+            T.fieldset[
+            T.legend(class_="freeform-form-label")["Report an Incident"],
+            T.input(type="hidden", name="t", value="report-incident"),
+            "What went wrong?: ",
+            T.input(type="text", name="details"), " ",
+            T.input(type="submit", value="Report!"),
+            ]]
+        return T.div[form]
