@@ -46,8 +46,9 @@ class UploadResultsRendererMixin(RateAndTimeMixin):
             if sharemap is None:
                 return "None"
             l = T.ul()
-            for shnum in sorted(sharemap.keys()):
-                l[T.li["%d -> %s" % (shnum, sharemap[shnum])]]
+            for shnum, peerids in sorted(sharemap.items()):
+                peerids = ', '.join([idlib.shortnodeid_b2a(i) for i in peerids])
+                l[T.li["%d -> placed on [%s]" % (shnum, peerids)]]
             return l
         d.addCallback(_render)
         return d
@@ -772,6 +773,41 @@ class Status(rend.Page):
     docFactory = getxmlfile("status.xhtml")
     addSlash = True
 
+    def renderHTTP(self, ctx):
+        t = get_arg(inevow.IRequest(ctx), "t")
+        if t == "json":
+            return self.json(ctx)
+        return rend.Page.renderHTTP(self, ctx)
+
+    def json(self, ctx):
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+        client = IClient(ctx)
+        data = {}
+        data["active"] = active = []
+        for s in self.data_active_operations(ctx, None):
+            si_s = base32.b2a_or_none(s.get_storage_index())
+            size = s.get_size()
+            status = s.get_status()
+            if IUploadStatus.providedBy(s):
+                h,c,e = s.get_progress()
+                active.append({"type": "upload",
+                               "storage-index-string": si_s,
+                               "total-size": size,
+                               "status": status,
+                               "progress-hash": h,
+                               "progress-ciphertext": c,
+                               "progress-encode-push": e,
+                               })
+            elif IDownloadStatus.providedBy(s):
+                active.append({"type": "download",
+                               "storage-index-string": si_s,
+                               "total-size": size,
+                               "status": status,
+                               "progress": s.get_progress(),
+                               })
+
+        return simplejson.dumps(data, indent=1) + "\n"
+
     def _get_all_statuses(self, client):
         return itertools.chain(client.list_all_upload_statuses(),
                                client.list_all_download_statuses(),
@@ -892,13 +928,14 @@ class HelperStatus(rend.Page):
         return rend.Page.renderHTTP(self, ctx)
 
     def render_JSON(self, ctx):
+        inevow.IRequest(ctx).setHeader("content-type", "text/plain")
         try:
             h = IClient(ctx).getServiceNamed("helper")
         except KeyError:
-            return simplejson.dumps({})
+            return simplejson.dumps({}) + "\n"
 
         stats = h.get_stats()
-        return simplejson.dumps(stats, indent=1)
+        return simplejson.dumps(stats, indent=1) + "\n"
 
     def render_active_uploads(self, ctx, data):
         return data["chk_upload_helper.active_uploads"]
@@ -937,7 +974,8 @@ class Statistics(rend.Page):
             stats = provider.get_stats()
         t = get_arg(inevow.IRequest(ctx), "t")
         if t == "json":
-            return simplejson.dumps(stats, indent=1)
+            inevow.IRequest(ctx).setHeader("content-type", "text/plain")
+            return simplejson.dumps(stats, indent=1) + "\n"
         # is there a better way to provide 'data' to all rendering methods?
         self.original = stats
         return rend.Page.renderHTTP(self, ctx)
